/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.Writable;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public abstract class AbstractVertex<V extends Writable, E extends Writable, M extends Writable> {
  
  public static enum VertexType {
    MASTER, REPLICA, GHOST
  }
  
  protected VertexType vertexType;
  protected V vertexID;
  
  protected M value;
  //TODO: we suppose the message value used by neighbors are the same, maybe we could design a meta-data
  protected M message;

  //the value and message set by last iteration, and used by current iteration
  protected M lastValue; 
  protected M lastMessage;
  public boolean updated = false;

  // out-edges in local peer
  protected List<AbstractRefEdge<V, E, M>> outEdges = new ArrayList<AbstractRefEdge<V, E, M>>() ;
  protected boolean awaked = false;

  private static final Log LOG = LogFactory.getLog(AbstractVertex.class);

  protected M getNewInstanceOfM() {
    return null;
  }
  
  protected V getNewInstanceOfV() {
    return null;
  }
  
  protected E getNewInstanceOfE() {
    return null;
  }

  public void setVertexType(VertexType type) {
    this.vertexType = type;
  }
  
  public VertexType getVertexType() {
    return this.vertexType;
  }
  
  public boolean isMaster() {
    return this.vertexType == VertexType.MASTER;
  }
  
  public V getVertexID() {
    return vertexID;
  }
  
  public void setVertexID(V vertexID) {
    this.vertexID = vertexID;
  }
  
  public M getLastValue() {
    return this.lastValue;
  }
  
  public M getNewValue() {
    return this.value;
  }
  
  public void setLastValue(Writable value) {
    this.lastValue = (M) value;
  }

  public void setNewValue(Writable value) {
    this.value = (M) value;
    this.updated = true;
  }

  public M getLastMessage() {
    return this.lastMessage;
  }

  public M getNewMessage() {
    return this.message;
  }
  
  public void setLastMessage(Writable message) {
    this.lastMessage = (M) message;
  }

  public void setNewMessage(Writable message) {
    this.message = (M) message;
    this.updated = true;
  }

  public void addOutEdge(AbstractRefEdge<V, E, M> edge) {
    this.outEdges.add(edge);
  }
  
  public List<AbstractRefEdge<V, E, M>> getOutEdges() {
    return this.outEdges;
  }

  @SuppressWarnings("unchecked")
  public long awake(List<Writable> list) {
    int num = 0;

    for (Writable id : list) {
      for (AbstractRefEdge<V, E, M> e : this.outEdges) {
        if (e.getVertexID().equals((V) id)) {
          AbstractVertex<V, E, M> neighbor = e.getVertex();
          neighbor.awaked = true;
          e.getPair().activate();
          num++;
          break;
        }
      }
    }
    return num;
  }
  
  /**
   * (BSP-g) awake neighbors
   */  
  public long awakeNeighbors() {
    for (AbstractRefEdge<V, E, M> e: this.outEdges) {
      AbstractVertex<V, E, M> neighbor = e.getVertex();
      neighbor.awaked = true;
      e.getPair().activate();
    }
    return this.outEdges.size();
  }

    /**
   * write information of the vertex
   */
  
  public void write(DataOutput out) throws java.io.IOException{
    //LOG.debug("int the write func");
    if (this.vertexType == null) {
      LOG.debug("wtf");
      this.vertexType = VertexType.MASTER;
    }
    int vType = 0;
    switch (this.vertexType) {
      case MASTER:
        vType = 1;
        break;
      case REPLICA:
        vType = 2;
        break;
      case GHOST:
        vType = 3;
        break;
    }
    try {
      out.writeInt(vType);

      if (vertexID == null)
        LOG.debug("null ID");
      else
        vertexID.write(out);

      out.writeBoolean(this.updated);
      if (this.updated) {
	if (this.value != null) {
	  out.writeBoolean(true);
	  this.value.write(out);
	} else {
	  out.writeBoolean(false);
	}
	/*
	  if (this.lastValue != null) {
	  out.writeBoolean(true);
	  this.lastValue.write(out);
	  } else {
	  out.writeBoolean(false);
	  }
	*/
	if (this.message != null) {
	  out.writeBoolean(true);
	  this.message.write(out);
	} else {
	  out.writeBoolean(false);
	}
	/*
	  if (this.lastMessage != null) {
	  out.writeBoolean(true);
	  this.lastMessage.write(out);
	  } else {
	  out.writeBoolean(false);
	  }
	*/
      } else {
	if (this.lastValue != null) {
	  out.writeBoolean(true);
	  this.lastValue.write(out);
	} else {
	  out.writeBoolean(false);
	}

	if (this.lastMessage != null) {
	  out.writeBoolean(true);
	  this.lastMessage.write(out);
	} else {
	  out.writeBoolean(false);
	}
      }
      //value.write(out);
      //message.write(out);
      //lastValue.write(out);
      //lastMessage.write(out);
      out.writeBoolean(awaked);
    }
    catch (IOException e) {
      throw e;
    }
  }
  
  public void readFields(DataInput in) throws java.io.IOException {
    //LOG.debug("read fields called");
    try {
      int vType = in.readInt();
      switch (vType) {
        case 1:
          this.vertexType = VertexType.MASTER;
          break;
        case 2:
          this.vertexType = VertexType.REPLICA;
          break;
        case 3:
          this.vertexType = VertexType.GHOST;
          break;
      }
      if (vertexID == null)
	vertexID = this.getNewInstanceOfV();
      vertexID.readFields(in);

      this.updated = in.readBoolean();
      if (this.updated) {
	if (in.readBoolean()) {
	  if (this.value == null)
	    this.value = this.getNewInstanceOfM();
	  this.value.readFields(in);
	} else {
	  this.value = null;
	}
	/*
	  if (in.readBoolean()) {
	  if (this.lastValue == null)
          this.lastValue = this.getNewInstanceOfM();
	  this.lastValue.readFields(in);
	  } else {
	  this.lastValue = null;
	  }
	*/
	if (in.readBoolean()) {
	  if (this.message == null)
	    this.message = this.getNewInstanceOfM();
	  this.message.readFields(in);
	} else {
	  this.message = null;
	}
	/*
	  if (in.readBoolean()) {
	  if (this.lastMessage == null)
          this.lastMessage = this.getNewInstanceOfM();
	  this.lastMessage.readFields(in);
	  } else {
	  this.lastMessage = null;
	  }
	*/
      } else {
	if (in.readBoolean()) {
          if (this.lastValue == null)
	    this.lastValue = this.getNewInstanceOfM();
          this.lastValue.readFields(in);
	} else {
          this.lastValue = null;
	}

	if (in.readBoolean()) {
          if (this.lastMessage == null)
	    this.lastMessage = this.getNewInstanceOfM();
          this.lastMessage.readFields(in);
	} else {
          this.lastMessage = null;
	}
	
	this.value = this.lastValue;
	this.message = this.lastMessage;
      }
      awaked = in.readBoolean();
    }
    catch (IOException e) {
      throw e;
    }
  }
}
