/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;

public class LyraSyncMessage implements Writable {
  private static final Log LOG = LogFactory.getLog(LyraSyncMessage.class);

  public static Class<? extends Writable> VERTEX_ID_CLASS;
  public static Class<? extends Writable> VERTEX_VALUE_CLASS;
  
  private IntWritable localID;
  private Writable vertexValue;
  private Writable message;
  private boolean activated;
  private boolean activateAll;
  private List<Writable> activateList;
  
  public LyraSyncMessage() {
    
  }
  
  public LyraSyncMessage(IntWritable localID, Writable vertexValue, 
      Writable msg, boolean activated, boolean all, List<Writable> list) {
    this.localID = localID;
    this.vertexValue = vertexValue;
    this.message = msg;
    this.activated = activated;
    this.activateAll = all;
    this.activateList = list;
  }
  
  @Override
  public void write(DataOutput out) throws IOException {
    localID.write(out);
    vertexValue.write(out);
    byte stateBitMap = 0;
    if (activated)
      stateBitMap |= 1;
    
    if (message != null)
      stateBitMap |= 4;
    
    if (activated) {
      if (activateAll)
        stateBitMap |= 2;
      out.writeByte(stateBitMap);
      
      if (activateAll == false) {
        out.writeInt(activateList.size());
        for (Writable destinationVertexId : activateList) {
          destinationVertexId.write(out);
        }
      }
    } else {
      out.writeByte(stateBitMap);
    }
    if (message != null)
      message.write(out);
  }
  
  @Override
  public void readFields(DataInput in) throws IOException {
    localID = new IntWritable();
    localID.readFields(in);
    vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
    vertexValue.readFields(in);

    byte stateBitMap = in.readByte();
    if ((stateBitMap & 1) == 1){
      activated = true;
      
      if ((stateBitMap & 2) == 2) {
        activateAll = true;
        activateList = null;
      } else {
        activateAll = false;
        int size = in.readInt();
        activateList = new ArrayList<Writable>(size);
        for (int i = 0; i < size; i++) {
          Writable activatedVertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
          activatedVertexId.readFields(in);
          activateList.add(activatedVertexId);
        }
      }      
    } else {
      activated = false;
      activateAll = false;
      activateList = null;
    }
    if ((stateBitMap & 4) == 4){
      message = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      message.readFields(in);
    } else {
      message = null;
    }
  }
  
  public Writable getVertexValue() {
    return vertexValue;
  }
  
  public IntWritable getLocalID() {
    return localID;
  }
  
  public Writable getMessage() {
    return message;
  }
  
  public boolean getActivated() {
    return activated;
  }
  
  public boolean getActivateAll() {
    return activateAll;
  }
  
  public List<Writable> getActivateList() {
    return activateList;
  }
}
