/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp.message;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configurable;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.ipc.RPC;
import org.apache.hadoop.ipc.RPC.Server;
import org.apache.hama.Constants;
import org.apache.hama.bsp.AbstractGlobalState;
import org.apache.hama.bsp.AbstractVertex;
import org.apache.hama.bsp.LyraMessageBundle;
import org.apache.hama.bsp.LyraSyncMessage;
import org.apache.hama.bsp.LyraTestMessage;
import org.apache.hama.bsp.LyraTestMessageBundle;
import org.apache.hama.ipc.HamaRPCProtocolVersion;
import org.apache.hama.util.LRUCache;

public class LyraMessageManagerImpl<M extends Writable> implements
    LyraMessageManager<M>, Configurable {

  protected static final Log LOG = LogFactory
      .getLog(LyraMessageManagerImpl.class);
  
  private Configuration conf;
  private Server server = null;

  private AbstractVertex<?, ?, ?>[] replicas;
  private AbstractGlobalState globalStat;
  private AtomicInteger numberMessages = new AtomicInteger(0);
  
  private final HashMap<String, InetSocketAddress> peerSocketCache = new HashMap<String, InetSocketAddress>();
  private LRUCache<InetSocketAddress, LyraMessageManager<M>> peersLRUCache = null;
  
  private int maxCachedConnections = 100;
  
  private int portOffset = 0;
  private int maxTaskPerGroom = 0;

  private double[] testSet;
  
  @Override
  public void init(Configuration conf) {
    init(conf, 1, 0, 1);
  }
  
  @Override
  public void init(Configuration conf, int numberThreads, int threadID, int numberServers) {
    this.conf = conf;
    this.maxCachedConnections = conf.getInt(MAX_CACHED_CONNECTIONS_KEY, 100);
    // avoid port conflict
    this.portOffset = (threadID % numberServers) * conf.getInt(Constants.MAX_TASKS_PER_GROOM, 3);
    
    LOG.info("Initialize lyra messenger on " + threadID + "/" + numberThreads + "thread");
    if (threadID < numberServers) {
      String lyraMsgHost = conf.get(Constants.PEER_HOST,
          Constants.DEFAULT_PEER_HOST);
      // NOTE: set PEER_MSG_PORT in GroomServer.main()
      int basePort = conf
          .getInt(Constants.PEER_MSG_PORT, 
              Constants.DEFAULT_PEER_PORT + Constants.DEFAULT_PEER_MSG_PORT_SPAN);
      int lyraMsgPort = basePort + portOffset;
      InetSocketAddress lyraMsgAddress = new InetSocketAddress(lyraMsgHost, lyraMsgPort);
      LOG.info("Start lyra msg-server at " + lyraMsgAddress);
      
      startRPCServer(conf, lyraMsgAddress);
    }
    
    peersLRUCache = new LRUCache<InetSocketAddress, LyraMessageManager<M>>(
        maxCachedConnections) {
      @Override
      protected final boolean removeEldestEntry(
          Map.Entry<InetSocketAddress, LyraMessageManager<M>> eldest) {
        if (size() > this.capacity) {
          LyraMessageManager<M> proxy = eldest.getValue();
          RPC.stopProxy(proxy);
          return true;
        }
        return false;
      }
    };
  }
  
  private final void startRPCServer(Configuration conf,
      InetSocketAddress msgAddress) {
    try {
      this.server = RPC.getServer(this, msgAddress.getHostName(),
          msgAddress.getPort(), conf);
      server.start();
    } catch (IOException e) {
      LOG.error("Fail to start lyra msg-server at " + msgAddress, e);
      throw new RuntimeException("lyra msg-server could not be launched!");
    }
  }

  private InetSocketAddress getLyraMsgAddress(String peerName) {
    String[] peerAddrParts = peerName.split(":");
    if (peerAddrParts.length != 2) {
      throw new ArrayIndexOutOfBoundsException(
          "Peername must consist of exactly ONE \":\"! Given peername was: "
              + peerName);
    }
    int msgPort = Integer.valueOf(peerAddrParts[1]) 
        + Constants.DEFAULT_PEER_MSG_PORT_SPAN + portOffset;
    return new InetSocketAddress(peerAddrParts[0], msgPort);
  }

  @SuppressWarnings("unchecked")
  protected final LyraMessageManager<M> getBSPPeerConnection(
      InetSocketAddress addr) throws IOException {
    LyraMessageManager<M> bspPeerConnection;
    if (!peersLRUCache.containsKey(addr)) {
      bspPeerConnection = (LyraMessageManager<M>) RPC.getProxy(
          LyraMessageManager.class, HamaRPCProtocolVersion.versionID, addr,
          this.conf);
      peersLRUCache.put(addr, bspPeerConnection);
    } else {
      bspPeerConnection = peersLRUCache.get(addr);
    }
    return bspPeerConnection;
  }
  
  @Override
  public void setConf(Configuration conf) {
    this.conf = conf;
  }

  @Override
  public Configuration getConf() {
    return conf;
  }
  
  @Override
  public void close() {
    if (server != null) {
      server.stop();
    }
  }

  @Override
  public void register(AbstractVertex<?, ?, ?>[] replicas, AbstractGlobalState stat) {
    this.replicas = replicas;
    this.globalStat = stat;
  }
  
  public void register(double[] testSet) {
    this.testSet = testSet;
  }
  
  @Override
  public void prepare() {
    this.numberMessages.set(0);
  }
  
  @Override
  public int size() {
    return numberMessages.get();
  }
  
  @Override
  public void transfer(String peerName, LyraMessageBundle bundle)
      throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target lyra msg-server.
    if (peerSocketCache.containsKey(peerName)) {
      addr = peerSocketCache.get(peerName);
    } else {
      addr = getLyraMsgAddress(peerName);
      peerSocketCache.put(peerName, addr);
    }
    
    LyraMessageManager<M> bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      bspPeerConnection.put(bundle);
    }
  }
  
  @Override
  public void transfer(String peerName, LyraTestMessageBundle bundle)
      throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target lyra msg-server.
    if (peerSocketCache.containsKey(peerName)) {
      addr = peerSocketCache.get(peerName);
    } else {
      addr = getLyraMsgAddress(peerName);
      peerSocketCache.put(peerName, addr);
    }
    
    LyraMessageManager<M> bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      bspPeerConnection.put(bundle);
    }
  }
  
  @Override
  public void transfer(String peerName, MapWritable stat)
      throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target peer.
    if (peerSocketCache.containsKey(peerName)) {
      addr = peerSocketCache.get(peerName);
    } else {
      addr = getLyraMsgAddress(peerName);
      peerSocketCache.put(peerName, addr);
    }
    
    LyraMessageManager<M> bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      bspPeerConnection.put(stat);
    }
  }
  
  @Override
  public void transfer(String peerName, DoubleWritable value)
      throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target peer.
    if (peerSocketCache.containsKey(peerName)) {
      addr = peerSocketCache.get(peerName);
    } else {
      addr = getLyraMsgAddress(peerName);
      peerSocketCache.put(peerName, addr);
    }
    
    LyraMessageManager<M> bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      bspPeerConnection.put(value);
    }
  }
  
  @Override
  public void transfer(String peerName)
      throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target peer.
    if (peerSocketCache.containsKey(peerName)) {
      addr = peerSocketCache.get(peerName);
    } else {
      addr = getLyraMsgAddress(peerName);
      peerSocketCache.put(peerName, addr);
    }
    
    LyraMessageManager<M> bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      bspPeerConnection.put();
    }
  }

  @Override
  public long getProtocolVersion(String protocol, long clientVersion)
      throws IOException {
    return versionID;
  }
  
  @Override
  public final void put(MapWritable stat) throws IOException {
    synchronized (globalStat) {
      this.globalStat.getNextState().putAll(stat);
    }
    numberMessages.incrementAndGet();
  }

  @Override
  public void put(LyraMessageBundle bundle) throws IOException {
    // thread safe ^.^
    for (Writable msg : bundle.getMessages()) {
      LyraSyncMessage sync = (LyraSyncMessage) msg;
      
      AbstractVertex<?, ?, ?> replica = replicas[sync.getLocalID().get()];
      
      replica.setNewValue(sync.getVertexValue());
      replica.setNewMessage(sync.getMessage());
      boolean activated = sync.getActivated();
      
      //local activate 
      if (activated) {
        boolean activateAll = sync.getActivateAll();
        if (activateAll)
          replica.awakeNeighbors();
        else
          replica.awake(sync.getActivateList());
      }
    }
    numberMessages.addAndGet(bundle.getMessages().size());
  }
  
  @Override
  public void put(LyraTestMessageBundle bundle) throws IOException {
    // thread safe ^.^
    for (Writable msg : bundle.getMessages()) {
      LyraTestMessage test = (LyraTestMessage) msg;
      testSet[(int)test.getID().get()] = test.getValue().get();
    }
    numberMessages.addAndGet(bundle.getMessages().size());
  }
  
  public final void put(DoubleWritable value) throws IOException {
    //numberMessages.incrementAndGet();
  }
  
  public final void put() throws IOException {
    numberMessages.incrementAndGet();
  }
}
