/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hama.examples;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;
import org.apache.commons.math3.linear.CholeskyDecomposition;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hama.HamaConfiguration;
import org.apache.hama.bsp.DxReadable;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.HashPartitioner2;
import org.apache.hama.bsp.MetisPartitioner;
import org.apache.hama.bsp.TextInputFormat;
import org.apache.hama.bsp.TextOutputFormat;
import org.apache.hama.graph.Edge;
import org.apache.hama.graph.GraphJob;
import org.apache.hama.graph.RefEdge;
import org.apache.hama.graph.Vertex;
import org.apache.hama.graph.VertexInputReader;

public class DxALS {
  private static final Log LOG = LogFactory.getLog(DxALS.class);
  public static final double LAMBDA = 0.1;
  public static final int NLATENT = 15;
  public static double LOCAL_CONVERGENCE_ERROR = 0.0001;
  public static final long MAX_UPDATES = Long.MAX_VALUE;
  public static long NUSERS = 10000;

  // the edge data type
  public static class DxALSEdge implements Writable {

    enum EdgeRoleEnum {
      TRAIN, VALIDATE, PREDICT
    };

    private float edgeObs;
    private EdgeRoleEnum edgeRole;

    // default constructor
    public DxALSEdge() {

    }

    // this constructor will override the default one, need to add one manually
    public DxALSEdge(EdgeRoleEnum role, float obs) {
      edgeRole = role;
      edgeObs = obs;
    }

    @Override
    public void write(DataOutput out) throws IOException {
      out.writeFloat(edgeObs);
      if (edgeRole != null){
        out.writeBoolean(true);
        new Text(edgeRole.toString()).write(out);
      } else {
        out.writeBoolean(false);
      }
    }

    @Override
    public void readFields(DataInput in) throws IOException {
      edgeObs = in.readFloat();
      if (in.readBoolean()){
        Text tmp = new Text();
        tmp.readFields(in);
        edgeRole = EdgeRoleEnum.valueOf(tmp.toString());
      } else {
        edgeRole = null;
      }
    }

  }

  // since Hama requires vertexValue and Message must be the same type (M),
  // and the message must identify where it is from, so the Message(vertexValue)
  // type now includes vertexid(msgVid)
  // this requires that setValue should avoid reset the msgVid attribute
  public static class DxALSMessage implements Writable, DxReadable {

    private Text msgVid;
    private float msgResidual;
    private int msgNupdates;
    private RealVector msgFactor;

    public DxALSMessage() {
      msgVid = new Text();
      msgFactor = new ArrayRealVector(NLATENT);
    }

    public DxALSMessage(Text id, float res, int updates, RealVector vec) {
      msgVid = id;
      msgResidual = res;
      msgNupdates = updates;
      msgFactor = vec;
    }

    public DxALSMessage(DxALSMessage msg) {
      msgVid = msg.getVid();
      msgResidual = msg.getResidual();
      msgNupdates = msg.getNupdates();
      msgFactor = msg.getFactor();
    }

    public Text getVid() {
      return msgVid;
    }

    public float getResidual() {
      return msgResidual;
    }

    public int getNupdates() {
      return msgNupdates;
    }

    /*
     * public RealVector getFactor(){ return msgFactor; }
     */

    @Override
    public void dxWriteBytes(DataOutput out) throws IOException {
      double[] factorArr = msgFactor.toArray();
      Assert.assertEquals(NLATENT, factorArr.length);
      // LOG.info("[DX-DxALS] factorArr of this vertex is " + factorArr[0] + " "
      // + factorArr[2] + " " + factorArr[4]);
      // FIXME: the final output is still skewed, the last one with a cursor and
      // is not complete
      for (int i = 0; i < factorArr.length; i++) {
        out.write(new Double(factorArr[i]).toString().getBytes("UTF-8"));
        out.write(" ".getBytes("UTF-8"));
      }
    }

    @Override
    public void write(DataOutput out) throws IOException {
      msgVid.write(out);
      out.writeFloat(msgResidual);
      out.writeInt(msgNupdates);
      double[] factorArr = msgFactor.toArray();
      Assert.assertEquals(NLATENT, factorArr.length);
      for (int i = 0; i < factorArr.length; i++)
        out.writeDouble(factorArr[i]);
    }

    @Override
    public void readFields(DataInput in) throws IOException {
      msgVid.readFields(in);
      msgResidual = in.readFloat();
      msgNupdates = in.readInt();
      double[] factorArr = new double[NLATENT];
      for (int i = 0; i < NLATENT; i++)
        factorArr[i] = in.readDouble();
      // FIXME: haven't looked into whether array matrix/vector is fast
      msgFactor = new ArrayRealVector(factorArr);
    }

    public RealMatrix getXtX() {
      Array2DRowRealMatrix XtX = new Array2DRowRealMatrix(NLATENT, NLATENT);

      for (int i = 0; i < NLATENT; i++)
        for (int j = 0; j < NLATENT; j++)
          XtX.setEntry(i, j, msgFactor.getEntry(i) * msgFactor.getEntry(j));
      return XtX;
    }

    public RealVector getXy(float obs) {
      RealVector Xy = msgFactor.mapMultiply(obs);
      return Xy;
    }

    public RealVector getFactor() {
      return msgFactor;
    }

  }

  public static class DxALSVertex extends Vertex<Text, DxALSEdge, DxALSMessage> {


    protected DxALSMessage getNewInstanceOfM() {
      return new DxALSMessage();
    }

    protected Text getNewInstanceOfV() {
      return new Text();
    }

    protected DxALSEdge getNewInstanceOfE() {
      return new DxALSEdge();
    }

    
    public void setup(Configuration conf) {
      String val = conf.get("hama.graph.max.convergence.error");
      if (val != null) {
        LOCAL_CONVERGENCE_ERROR = Double.parseDouble(val);
      }
      val = conf.get("als.nusers");
      if(val != null){
    	  NUSERS = Long.parseLong(val);
      }
    }

    @Override
    public void compute(Iterator<DxALSMessage> messages) throws IOException {
      DxALSMessage newValue;
      if (this.getSuperstepCount() == 0) {
        RealVector fact = new ArrayRealVector(NLATENT);
        newValue = new DxALSMessage(this.getVertexID(), 0, 0, fact);

        Random rdm = new java.util.Random();
        String seedStr = "";
        //if (this.getVertexID().toString().startsWith("m"))
        if(Long.parseLong(this.getVertexID().toString()) > NUSERS)	//movie vertices [NUSERS+1, ...)
          seedStr += (this.getNumVertices() + 1);
        //else if (this.getVertexID().toString().startsWith("u"))
        else if(Long.parseLong(this.getVertexID().toString()) <= NUSERS)	//movie vertices [NUSERS+1, ...)
          seedStr += (this.getNumVertices() + 2);
        seedStr += this.getVertexID().toString().substring(1);
        rdm.setSeed(Long.parseLong(seedStr));

        for (int i = 0; i < NLATENT; i++)
          fact.setEntry(i, rdm.nextGaussian());
        this.setValue(newValue);

        // set movie vertices to sleep and no message will invoke them, so that,
        // in the first round, only user vertices will be active
        if(Long.parseLong(this.getVertexID().toString()) > NUSERS){
          this.activateNeighbors(new DxALSMessage(newValue));
        }

        this.voteToHalt();
        return;
      }

      DxALSMessage value = this.getValue();
      // a new message object, new value is stored in this object, should later
      // use this object to setValue
      newValue = new DxALSMessage(value);

      RealMatrix XtXSum = new Array2DRowRealMatrix(NLATENT, NLATENT);
      RealVector XySum = new ArrayRealVector(NLATENT);

      RealVector oldFactor = value.msgFactor;

      DxALSMessage msg;
      Text otherVid;
      Edge<Text, DxALSEdge> edge;
      // HashMap<Text, NeighborValue> neighborMap = new HashMap<Text,
      // NeighborValue>();

      // changed from orig
      for (RefEdge<Text, DxALSEdge, DxALSMessage> e : this.getInEdges()) {
        // get the ref vertex at the other end
        Vertex<Text, DxALSEdge, DxALSMessage> v = e.getVertex();
        // get the message(stored in local) it should send in last ss
        msg = v.getMessage();
        float obs = e.getValue().edgeObs;

        if (e.isAwake() && msg != null) {
          XtXSum = XtXSum.add(msg.getXtX());
          XySum = XySum.add(msg.getXy(obs));
        }
      }
      // FIXME! the length will never be null because you new a size=NLATENT
      // factor
      if (XySum.toArray().length == 0) {
        newValue.msgResidual = 0;
        newValue.msgNupdates++;
        this.setValue(newValue);
        this.voteToHalt();
        return;
      }
      for (int i = 0; i < XtXSum.getRowDimension(); i++)
        XtXSum.addToEntry(i, i, LAMBDA);

      CholeskyDecomposition cd = new CholeskyDecomposition(XtXSum);
      newValue.msgFactor = cd.getSolver().solve(XySum);

      // initially convert this to int, and it's always 0
      newValue.msgResidual = (float) (newValue.msgFactor.subtract(oldFactor)
          .getL1Norm() / (XtXSum.getRowDimension()));
      newValue.msgVid = this.getVertexID();

      ++newValue.msgNupdates;
      this.setValue(newValue);

      // only in-edges have replica information, out-edges have only local
      // vertices (only id, not the entire vertex info)
      // so send messages should also iterate over in-edges
      for (RefEdge<Text, DxALSEdge, DxALSMessage> e : this.getInEdges()) {
        // get the edge information
        DxALSEdge edgeValue = e.getValue();
        if (edgeValue.edgeRole == DxALSEdge.EdgeRoleEnum.TRAIN) {
          if (e.isAwake()) {
            // get the vertex at the other end of the edge
            Vertex neighborVertex = e.getVertex();
            // get the message the other vertex send in last superstep
            DxALSMessage neighborMsg = (DxALSMessage) (neighborVertex
                .getMessage());
            RealVector otherFactor = neighborMsg.msgFactor;
            int otherUpdates = neighborMsg.msgNupdates;
            double predict = newValue.msgFactor.dotProduct(otherFactor);
            double error = Math.abs(edgeValue.edgeObs - predict);
            double priority = error * newValue.msgResidual;
            if (priority > LOCAL_CONVERGENCE_ERROR
                && otherUpdates < MAX_UPDATES) {
              this.activate((Text) (neighborVertex.getVertexID()), newValue);
            }
          }
        }
      }

      this.voteToHalt();
    }

  }

  public static class DxALSTextReader2 extends
      VertexInputReader<LongWritable, Text, Text, DxALSEdge, DxALSMessage> {

    /**
     * The text file essentially should look like: <br/>
     * FROM_VERTEX_ID\tTO_VERTEX_ID<br/>
     * E.G:<br/>
     * 0\t11342<br/>
     * 0\t824020<br/>
     * etc.
     */
    String lastVertexId = null;
    // List<String> adjacents = new ArrayList<String>();
    List<Edge<Text, DxALSEdge>> adjacents = new ArrayList<Edge<Text, DxALSEdge>>();

    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, DxALSEdge, DxALSMessage> vertex) {
      // the tail vertex of input
      if (key == null || value == null) {
        assert (lastVertexId != null);
        vertex.setVertexID(new Text(lastVertexId));
        for (Edge<Text, DxALSEdge> adjacent : adjacents) {
          vertex.addEdge(adjacent);
        }
        adjacents.clear();
        lastVertexId = null;
        return true;
      }

      // normal parse
      String line = value.toString();
      String[] lineSplit = line.split(" ");
      if (!line.startsWith("#")) {
        if (lastVertexId == null) {
          lastVertexId = lineSplit[0];
        }
        if (lastVertexId.equals(lineSplit[0])) {
          adjacents.add(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
              new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
                  .parseFloat(lineSplit[2]))));
        } else {
          vertex.setVertexID(new Text(lastVertexId));
          for (Edge<Text, DxALSEdge> adjacent : adjacents) {
            vertex.addEdge(adjacent);
          }
          adjacents.clear();
          lastVertexId = lineSplit[0];
          adjacents.add(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
              new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
                  .parseFloat(lineSplit[2]))));
          return true;
        }
      }
      return false;
    }
  }

  public static class DxALSTextReader extends
      VertexInputReader<LongWritable, Text, Text, DxALSEdge, DxALSMessage> {

    String lastVertexId = null;

    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, DxALSEdge, DxALSMessage> vertex) throws Exception {
      return true;
    }

    public boolean dxParseVertex(LongWritable key, Text value,
        Vertex<Text, DxALSEdge, DxALSMessage> vertex,
        Vertex<Text, DxALSEdge, DxALSMessage> vertexNext) throws Exception {
      String line = value.toString();
      // line.replaceAll("[ ]+", " ");
      String[] lineSplit = line.split(" ");

      double[] factArr = new double[NLATENT];

      Assert.assertNotNull(vertexNext.getEdges());
      Assert.assertNotNull(vertex.getEdges());
      // LOG.info("#####vertexNext size is now " +
      // vertexNext.getEdges().size());

      if (!line.startsWith("#")) {
        if (lastVertexId == null) {
          lastVertexId = lineSplit[0];
          vertex.setVertexID(new Text(lastVertexId));
        }
        if (vertexNext.getEdges().size() != 0) {
          Assert.assertNotNull(vertexNext.getVertexID());
          // LOG.info("vertexNext is not null, it's " + vertexNext.getVertexID()
          // + ":" + vertexNext.getEdges().toString());
          vertex.setVertexID(vertexNext.getVertexID());

          // TODO, only add one edge is enough for the current data input
          vertex.addEdge(vertexNext.getEdges().iterator().next());
          vertexNext.setVertexID(null);
          vertexNext.getEdges().clear();
        }
        if (lastVertexId.equals(lineSplit[0])) {
          vertex.addEdge(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
              new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
                  .parseFloat(lineSplit[2]))));
          // LOG.info("vertex id " + vertex.getVertexID() + "edges: " +
          // vertex.getEdges().toString());
        } else {
          lastVertexId = lineSplit[0];
          vertexNext.setVertexID(new Text(lastVertexId));
          vertexNext.addEdge(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
              new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
                  .parseFloat(lineSplit[2]))));
          return true;
        }
      }

      return false;
    }

  }

  public static void main(String[] args) throws IOException,
      InterruptedException, ClassNotFoundException {
    HamaConfiguration conf = new HamaConfiguration(new Configuration());
    GraphJob alsJob = createJob(args, conf);
    long startTime = System.currentTimeMillis();
    if (alsJob.waitForCompletion(true)) {
      System.out.println("Job Finished in "
          + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds");
    }

  } // end of main()

  public static GraphJob createJob(String[] args, HamaConfiguration conf)
      throws IOException {
    GraphJob alsJob = new GraphJob(conf, PageRank.class);
    alsJob.setJobName("ALS");

    alsJob.setVertexClass(DxALSVertex.class);
    alsJob.setInputPath(new Path(args[0]));
    alsJob.setOutputPath(new Path(args[1]));

    // set the defaults
    alsJob.setMaxIteration(30);
    alsJob.setNumBspTask(3);
    alsJob.set("hama.graph.selective.activate", "true");
    alsJob.set("hama.graph.repair", "true");
    alsJob.set("hama.graph.mun.threads", "1");
    alsJob.set("ft.need.edge.info", "true");
    alsJob.setPartitioner(HashPartitioner.class);
    alsJob.setPartitioner(HashPartitioner2.class);

    if (args.length >= 3)
      alsJob.setNumBspTask(Integer.parseInt(args[2]));
    if (args.length >= 4)
      alsJob.setMaxIteration(Integer.parseInt(args[3]));
    if (args.length >= 5){
        if (args[4].equals("1"))
            alsJob.setPartitioner(MetisPartitioner.class);
    }
    if (args.length >= 6)
      alsJob.set("hama.graph.max.convergence.error", args[5]);
    if (args.length >= 7)
    	alsJob.set("als.nusers", args[6]);
    if (args.length >= 8)
      alsJob.set("hama.graph.min.rate", args[7]);

    alsJob.setVertexIDClass(Text.class);
    alsJob.setVertexValueClass(DxALSMessage.class);
    alsJob.setEdgeValueClass(DxALSEdge.class);

    alsJob.setInputKeyClass(LongWritable.class); // input key is generated by
                                                 // the system
    alsJob.setInputValueClass(Text.class); // all your actual input is in value
    alsJob.setInputFormat(TextInputFormat.class);
    alsJob.setVertexInputReaderClass(DxALSTextReader2.class);

    alsJob.setOutputFormat(TextOutputFormat.class);
    alsJob.setOutputKeyClass(Text.class);
    alsJob.setOutputValueClass(DxALSMessage.class);
    return alsJob;
  }
}
