/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.examples;

import java.io.IOException;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hama.HamaConfiguration;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.MetisPartitioner;
import org.apache.hama.bsp.TextInputFormat;
import org.apache.hama.bsp.TextOutputFormat;
import org.apache.hama.graph.Aggregator;
import org.apache.hama.graph.AverageAggregator;
import org.apache.hama.graph.Edge;
import org.apache.hama.graph.GraphJob;
import org.apache.hama.graph.RefEdge;
import org.apache.hama.graph.Vertex;
import org.apache.hama.graph.VertexInputReader;

/**
 * Real pagerank with dangling node contribution.
 */
public class PageRank {

  public static class PageRankVertex extends
      Vertex<Text, NullWritable, DoubleWritable> {

    static double DAMPING_FACTOR = 0.85;
    static double MAXIMUM_CONVERGENCE_ERROR = 0.000000001d; // e.g. 0.1% of the
                                                           // initial value

    int numEdges;

    protected Text getNewInstanceOfV() {
      return new Text("");
    }
    
    protected DoubleWritable getNewInstanceOfM() {
      return new DoubleWritable(0);
    }

    
    @Override
    public void setup(Configuration conf) {
      String val = conf.get("hama.pagerank.alpha");
      if (val != null) {
        DAMPING_FACTOR = Double.parseDouble(val);
      }
      val = conf.get("hama.graph.max.convergence.error");
      if (val != null) {
        MAXIMUM_CONVERGENCE_ERROR = Double.parseDouble(val);
      }
      numEdges = this.getEdges().size();
    }

    @Override
    public void compute(Iterator<DoubleWritable> messages) throws IOException {
      //(BSP-g) message is null!
      if (this.getSuperstepCount() == 0) {
        // initialize this vertex to 1 / count of global vertices in this graph
        double value = 1.0 / this.getNumVertices();
        this.setValue(new DoubleWritable(value));
        
        this.activateNeighbors(new DoubleWritable(value / numEdges));
      } else if (this.getSuperstepCount() >= 1) {
        double lastValue = this.getValue().get();

        double sum = 0;
        for (RefEdge<Text, NullWritable, DoubleWritable> e : this.getInEdges()) {
          Vertex<Text, NullWritable, DoubleWritable> v = e.getVertex();
          DoubleWritable msg = v.getMessage();
          sum += msg.get();
        }
        double value = (1.0d - DAMPING_FACTOR) / this.getNumVertices() + (DAMPING_FACTOR * sum);
        
        this.setValue(new DoubleWritable(value));
        
        double localError = Math.abs(value - lastValue);
        if (localError > MAXIMUM_CONVERGENCE_ERROR) {
          this.activateNeighbors(new DoubleWritable(value / numEdges));
        }
        voteToHalt();
      }
    }    
  }
  
  public static class WikiTextReader extends
      VertexInputReader<LongWritable, Text, Text, NullWritable, DoubleWritable> {

    /**
     * The text file essentially should look like: <br/>
     * VERTEX_ID: link1 link2 link3 ...(n-space separated VERTEX_IDs)<br/>
     */
    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, NullWritable, DoubleWritable> vertex) throws Exception {
      String[] split = value.toString().split(" ");
      for (int i = 0; i < split.length; i++) {
        if (i == 0) {
          String id = split[0];
          vertex.setVertexID(new Text(id.substring(0, id.length() - 1)));
        } else {
          vertex
              .addEdge(new Edge<Text, NullWritable>(new Text(split[i]), null));
        }
      }
      return true;
    }

  }

  public static class PagerankTextReader extends
      VertexInputReader<LongWritable, Text, Text, NullWritable, DoubleWritable> {

    /**
     * The text file essentially should look like: <br/>
     * VERTEX_ID\t(n-tab separated VERTEX_IDs)<br/>
     * E.G:<br/>
     * 1\t2\t3\t4<br/>
     * 2\t3\t1<br/>
     * etc.
     */
    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, NullWritable, DoubleWritable> vertex) throws Exception {
      String[] split = value.toString().split("\t");
      for (int i = 0; i < split.length; i++) {
        if (i == 0) {
          vertex.setVertexID(new Text(split[i]));
        } else {
          vertex
              .addEdge(new Edge<Text, NullWritable>(new Text(split[i]), null));
        }
      }
      return true;
    }

  }

  public static class PagerankTextReader2 extends
      VertexInputReader<LongWritable, Text, Text, NullWritable, DoubleWritable> {

    /**
     * The text file essentially should look like: <br/>
     * FROM_VERTEX_ID\tTO_VERTEX_ID<br/>
     * E.G:<br/>
     * 0\t11342<br/>
     * 0\t824020<br/>
     * etc.
     */
    String lastVertexId = null;
    List<String> adjacents = new ArrayList<String>();

    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, NullWritable, DoubleWritable> vertex) {
      // the tail vertex of input 
      if (key == null || value == null) {
        assert(lastVertexId != null);
        vertex.setVertexID(new Text(lastVertexId));
        for (String adjacent : adjacents) {
          vertex.addEdge(new Edge<Text, NullWritable>(new Text(adjacent),
              null));
        }
        adjacents.clear();
        lastVertexId = null;
        return true;
      }

      // normal parse
      String line = value.toString();
      String[] lineSplit = line.split("\t");
      if (!line.startsWith("#")) {
        if (lastVertexId == null) {
          lastVertexId = lineSplit[0];
        }
        if (lastVertexId.equals(lineSplit[0])) {
          adjacents.add(lineSplit[1]);
        } else {
          vertex.setVertexID(new Text(lastVertexId));
          for (String adjacent : adjacents) {
            vertex.addEdge(new Edge<Text, NullWritable>(new Text(adjacent),
                null));
          }
          adjacents.clear();
          lastVertexId = lineSplit[0];
          adjacents.add(lineSplit[1]);
          return true;
        }
      }
      return false;
    }
  }
  
  
  public static class PagerankTextReader3 extends
      VertexInputReader<LongWritable, Text, Text, NullWritable, DoubleWritable> {

    /**
     * The text file essentially should look like: <br/>
     * VERTEX_ID : (n-space separated VERTEX_IDs)<br/>
     * E.G:<br/>
     * 1 : 23 4
     * 5 : 
     * 6 : 7
     * etc.
     */
    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, NullWritable, DoubleWritable> vertex) throws Exception {
      if (value == null || key == null){
        return false;
      }
      String[] split = value.toString().split(" ");
      for (int i = 0; i < split.length; i++) {
        if (i == 0) {
          vertex.setVertexID(new Text(split[i]));
        } else if (i > 1) {
          vertex.addEdge(new Edge<Text, NullWritable>(new Text(split[i]), null));
        }
      }
      return true;
    }

  }
  
  
  private static void printUsage() {
    System.out
        .println("Usage: <input> <output> [number of tasks] " +
                  "[max iterations (default 30)] " +
                  "[graph partitioner (defaul HashPartitioner, 0 for hash, 1 for metis, 2 for fennel)] " +
                  "[epsilon (convergence error, default 0.001)] " +
                  "[min rate of convergence (default 0.0)]" + 
                  "[mumber of threads (default 1)]" + 
                  "[damping factor (default 0.85)]" +
                  "[hama.graph.vertex.num for fennel]" +
                  "[hama.graph.edge.num for fennel]");
    System.exit(-1);
  }

  public static void main(String[] args) throws IOException,
      InterruptedException, ClassNotFoundException {
    if (args.length < 2)
      printUsage();

    HamaConfiguration conf = new HamaConfiguration(new Configuration());
    GraphJob pageJob = createJob(args, conf);

    long startTime = System.currentTimeMillis();
    if (pageJob.waitForCompletion(true)) {
      System.out.println("Job Finished in "
          + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds");
    }
  }

  @SuppressWarnings("unchecked")
  public static GraphJob createJob(String[] args, HamaConfiguration conf)
      throws IOException {
    GraphJob pageJob = new GraphJob(conf, PageRank.class);
    pageJob.setJobName("Pagerank");

    pageJob.setVertexClass(PageRankVertex.class);
    pageJob.setInputPath(new Path(args[0]));
    pageJob.setOutputPath(new Path(args[1]));

    // set the defaults
    pageJob.setMaxIteration(30);
    pageJob.set("hama.pagerank.alpha", "0.85");
    pageJob.set("hama.graph.repair", "true");
    pageJob.set("hama.graph.max.convergence.error", "0.0");
    pageJob.set("hama.graph.mun.threads", "1");
    pageJob.set("ft.need.edge.info", "false");
    pageJob.setPartitioner(HashPartitioner.class);
    
    if (args.length >= 10)
      pageJob.set("hama.graph.edge.num", args[9]);
    if (args.length >= 9)
      pageJob.set("hama.graph.vertex.num", args[8]);
    if (args.length >= 8)
      pageJob.set("hama.pagerank.alpha", args[7]);
    if (args.length >= 7)
      pageJob.set("hama.graph.min.rate", args[6]);
    if (args.length >= 6)
      pageJob.set("hama.graph.max.convergence.error", args[5]);
    if (args.length >= 5) {
      if (args[4].equals("1"))
        pageJob.setPartitioner(MetisPartitioner.class);
      if (args[4].equals("2")){
        pageJob.set("bsp.input.partitioner.fennel", "true");
      }
    }
    if (args.length >= 4)
      pageJob.setMaxIteration(Integer.parseInt(args[3]));
    if (args.length >= 3)
      pageJob.setNumBspTask(Integer.parseInt(args[2]));

    // error, dangling node probability sum
    //pageJob.setAggregatorClass(AverageAggregator.class,
    //    DanglingNodeAggregator.class);

    pageJob.setVertexIDClass(Text.class);
    pageJob.setVertexValueClass(DoubleWritable.class);
    pageJob.setEdgeValueClass(NullWritable.class);

    pageJob.setInputKeyClass(LongWritable.class);
    pageJob.setInputValueClass(Text.class);
    pageJob.setInputFormat(TextInputFormat.class);
    pageJob.setVertexInputReaderClass(PagerankTextReader3.class);
    //pageJob.setVertexInputReaderClass(WikiTextReader.class);
    pageJob.setOutputFormat(TextOutputFormat.class);
    pageJob.setOutputKeyClass(Text.class);
    pageJob.setOutputValueClass(DoubleWritable.class);
    
    //TODO: we need try to combine message?
    return pageJob;
  }

  public static class DanglingNodeAggregator implements
      Aggregator<DoubleWritable, Vertex<Text, NullWritable, DoubleWritable>> {

    double danglingNodeSum;

    @Override
    public void aggregate(Vertex<Text, NullWritable, DoubleWritable> vertex,
        DoubleWritable value) {
      if (vertex != null) {
        if (vertex.getEdges().size() == 0) {
          danglingNodeSum += value.get();
        }
      } else {
        danglingNodeSum += value.get();
      }
    }

    @Override
    public DoubleWritable getValue() {
      return new DoubleWritable(danglingNodeSum);
    }

  }
}
