/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.examples;

import java.io.IOException;
import java.util.Iterator;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hama.HamaConfiguration;
import org.apache.hama.bsp.Combiner;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.MetisPartitioner;
import org.apache.hama.bsp.TextInputFormat;
import org.apache.hama.bsp.TextOutputFormat;
import org.apache.hama.graph.Edge;
import org.apache.hama.graph.GraphJob;
import org.apache.hama.graph.RefEdge;
import org.apache.hama.graph.Vertex;
import org.apache.hama.graph.VertexInputReader;

public class SSSP {

  public static final String START_VERTEX = "shortest.paths.start.vertex.name";

  public static class ShortestPathVertex extends
      Vertex<Text, DoubleWritable, DoubleWritable> {


    public Text getNewInstanceOfV() {
      return new Text();
    }

    public DoubleWritable getNewInstanceOfE() {
      return new DoubleWritable(0);
    }

    public DoubleWritable getNewInstanceOfM() {
      return new DoubleWritable(0);
    }

    public ShortestPathVertex() {
      this.setValue(new DoubleWritable(Double.MAX_VALUE));
    }

    public boolean isStartVertex() {
      Text startVertex = new Text(getConf().get(START_VERTEX));
      return (this.getVertexID().equals(startVertex)) ? true : false;
    }

    public void compute(Iterator<DoubleWritable> messages) throws IOException {
      if (this.getSuperstepCount() == 0){
        this.setValue(new DoubleWritable(Double.MAX_VALUE));
        if (isStartVertex()) {
          this.setValue(new DoubleWritable(0));
          this.activateNeighbors(new DoubleWritable(0));
        }
      } else {
        double minDist = Double.MAX_VALUE;
        for (RefEdge<Text, DoubleWritable, DoubleWritable> e : this
            .getInEdges()) {
          if (e.isAwake()) {
            double distance = e.getValue().get()
                + e.getVertex().getMessage().get();
            if (minDist > distance)
              minDist = distance;
          }
        }
        if (minDist < this.getValue().get()) {
          this.setValue(new DoubleWritable(minDist));
          this.activateNeighbors(new DoubleWritable(minDist));
        }
      }
      voteToHalt();
      
      
    }

    
    public void compute_old(Iterator<DoubleWritable> messages) throws IOException {
      double minDist = isStartVertex() ? 0 : Double.MAX_VALUE;

      while (messages.hasNext()) {
        DoubleWritable msg = messages.next();
        if (msg.get() < minDist) {
          minDist = msg.get();
        }
      }

      if (minDist < this.getValue().get()) {
        this.setValue(new DoubleWritable(minDist));
        for (Edge<Text, DoubleWritable> e : this.getEdges()) {
          sendMessage(e, new DoubleWritable(minDist + e.getValue().get()));
        }
      } else {
        voteToHalt();
      }
    }
  }

  public static class MinDoubleCombiner extends Combiner<DoubleWritable> {

    @Override
    public DoubleWritable combine(Iterable<DoubleWritable> messages) {
      double minDist = Double.MAX_VALUE;

      Iterator<DoubleWritable> it = messages.iterator();
      while (it.hasNext()) {
        double msgValue = it.next().get();
        if (minDist > msgValue)
          minDist = msgValue;
      }

      return new DoubleWritable(minDist);
    }
  }

  public static class SSSPTextReader extends
      VertexInputReader<LongWritable, Text, Text, DoubleWritable, DoubleWritable> {

    /**
     * The text file essentially should look like: <br/>
     * VERTEX_ID\t(n-tab separated VERTEX_ID:VERTEX_VALUE pairs)<br/>
     * E.G:<br/>
     * 1\t2:25\t3:32\t4:21<br/>
     * 2\t3:222\t1:922<br/>
     * etc.
     */
    @Override
    public boolean parseVertex(LongWritable key, Text value,
        Vertex<Text, DoubleWritable, DoubleWritable> vertex) throws Exception {
      if (key == null && value == null){
        return false;
      }
      String[] split = value.toString().split("\t");
      //System.out.println(split.length + "   " + value.toString());
      for (int i = 0; i < split.length; i++) {
        if (i == 0) {
          vertex.setVertexID(new Text(split[i]));
        } else {
          String[] split2 = split[i].split(":");
          vertex.addEdge(new Edge<Text, DoubleWritable>(new Text(split2[0]),
              new DoubleWritable(Double.parseDouble(split2[1]))));
        }
      }
      return true;
    }

  }

  private static void printUsage() {
    System.out.println("Usage: <startnode> <input> <output> [tasks]");
    System.exit(-1);
  }

  public static void main(String[] args) throws IOException,
      InterruptedException, ClassNotFoundException {
    if (args.length < 3)
      printUsage();

    // Graph job configuration
    HamaConfiguration conf = new HamaConfiguration();
    GraphJob ssspJob = new GraphJob(conf, SSSP.class);
    // Set the job name
    ssspJob.setJobName("Single Source Shortest Path");
    // Iterate until all the nodes have been reached.
    ssspJob.setMaxIteration(Integer.MAX_VALUE);
    ssspJob.set("hama.graph.selective.activate", "true");

    ssspJob.setPartitioner(HashPartitioner.class);
    
    conf.set(START_VERTEX, args[0]);
    ssspJob.setInputPath(new Path(args[1]));
    ssspJob.setOutputPath(new Path(args[2]));

    if (args.length >= 4)
      ssspJob.setNumBspTask(Integer.parseInt(args[3]));
    if (args.length >= 5)
      ssspJob.setMaxIteration(Integer.parseInt(args[4]));
    if (args.length >= 6) {
      if (args[5].equals("1"))
        ssspJob.setPartitioner(MetisPartitioner.class);
    }
    
    ssspJob.setVertexClass(ShortestPathVertex.class);
    //ssspJob.setCombinerClass(MinDoubleCombiner.class);
    ssspJob.setInputFormat(TextInputFormat.class);
    ssspJob.setInputKeyClass(LongWritable.class);
    ssspJob.setInputValueClass(Text.class);

    ssspJob.setOutputFormat(TextOutputFormat.class);
    ssspJob.setVertexInputReaderClass(SSSPTextReader.class);
    ssspJob.setOutputKeyClass(Text.class);
    ssspJob.setOutputValueClass(DoubleWritable.class);

    ssspJob.setVertexIDClass(Text.class);
    ssspJob.setVertexValueClass(DoubleWritable.class);
    ssspJob.setEdgeValueClass(DoubleWritable.class);

    long startTime = System.currentTimeMillis();
    if (ssspJob.waitForCompletion(true)) {
      System.out.println("Job Finished in "
          + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds");
    }
  }
}
