/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;

/**
 * A message that is either MapWritable (for meta communication purposes) or a
 * real message (vertex ID and value). It can be extended by adding flags, for
 * example for a graph repair call.
 */
public final class GraphJobMessage implements Writable {
  private static final Log LOG = LogFactory.getLog(GraphJobMessage.class);
  
  public static final int MAP_FLAG = 0x01;
  public static final int REPAIR_FLAG = 0x04;
  public static final int PARTITION_FLAG = 0x08;
  public static final int SIZE_FLAG = 0x10;

  public static final int REPLICA_FLAG = 0x20; // (BSP-g) build replica vertex in setup
  public static final int REGISTER_FLAG = 0x40; // (BSP-g) register replicas to master
  public static final int TEST_FLAG = 0x80;

  public static final int RECOVER_REPLICA_FLAG = 0x100;
  
  // statically defined because it is process-wide information, therefore in caps
  // considered as a constant
  public static Class<?> VERTEX_CLASS;
  public static Class<? extends Writable> VERTEX_ID_CLASS;
  public static Class<? extends Writable> VERTEX_VALUE_CLASS;
  public static Class<? extends Writable> EDGE_VALUE_CLASS;

  // default flag to -1 "unknown"
  private int flag = -1;
  private MapWritable map;
  private Writable vertexId;
  private Writable vertexValue;
  private Vertex<?, ?, ?> vertex;
  private IntWritable numberVertices;
  private IntWritable numberReplicas;

  private IntWritable peerIndex;
  private IntWritable localID;
  
  private LongWritable id;
  private DoubleWritable value;
  
  private List<?> edges;
  
  public GraphJobMessage() {
  }

  public GraphJobMessage(MapWritable map) {
    this.flag = MAP_FLAG;
    this.map = map;
  }

  public GraphJobMessage(Writable vertexId) {
    this.flag = REPAIR_FLAG;
    this.vertexId = vertexId;
  }
  
  public GraphJobMessage(Vertex<?, ?, ?> vertex) {
    this.flag = PARTITION_FLAG;
    this.vertex = vertex;
  }

  public GraphJobMessage(IntWritable numberVertices, IntWritable numberReplicas) {
    this.flag = SIZE_FLAG;
    this.numberVertices = numberVertices;
    this.numberReplicas = numberReplicas;
  }

  /**
   * build in-edges of destination vertex, and create replica if necessary
   * @param vertexId
   * @param vertexValue
   * @param destinationVertexId
   * @param edgeValue
   * @param peerName peer's name of master vertex 
   */
  public GraphJobMessage(Writable vertexId, Writable vertexValue, List<?>edges, IntWritable peerIndex) {
    this.flag = REPLICA_FLAG;
    this.vertexId = vertexId;
    this.vertexValue = vertexValue;
    this.edges = edges;
    this.peerIndex = peerIndex;
  }

  public GraphJobMessage(IntWritable peerIndex, 
      Writable vertexId, IntWritable localID) {
    this.flag = REGISTER_FLAG;
    this.peerIndex = peerIndex;
    this.vertexId = vertexId;
    this.localID = localID;
  }

  public GraphJobMessage(Vertex<?, ?, ?> vertex, List<?> edges, IntWritable peerIndex) {
    this.flag = RECOVER_REPLICA_FLAG;
    this.vertex = vertex;
    this.edges = edges;
    this.peerIndex = peerIndex;
  }
  
  public GraphJobMessage(LongWritable id, DoubleWritable value) {
    this.flag = TEST_FLAG;
    this.id = id;
    this.value = value;
  }
  
  @Override
  public void write(DataOutput out) throws IOException {
    out.writeInt(this.flag);
    if (isMapMessage()) {
      map.write(out);
    } else if (isPartitioningMessage()) {
      vertex.getVertexID().write(out);
      if (vertex.getNewValue() != null) {
        out.writeBoolean(true);
        vertex.getNewValue().write(out);
      } else {
        out.writeBoolean(false);
      }
      List<?> outEdges = vertex.getEdges();
      out.writeInt(outEdges.size());
      for (Object e : outEdges) {
        Edge<?, ?> edge = (Edge<?, ?>) e;
        edge.getDestinationVertexID().write(out);
        if (edge.getValue() != null) {
          out.writeBoolean(true);
          edge.getValue().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    } else if (isSizeMessage()) {
      numberVertices.write(out);
      numberReplicas.write(out);
    } else if (isRepairMessage()) {
      vertexId.write(out);
    } else if (isReplicaMessage()) {
      vertexId.write(out);
      if (vertexValue == null) {
        vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      }
      vertexValue.write(out);
      peerIndex.write(out);
      
      if (edges == null)
        out.writeBoolean(false);
      else {
        out.writeBoolean(true);
        out.writeInt(edges.size());
        for (Object e : edges) {
          Edge<?, ?> edge = (Edge<?, ?>) e;
          edge.getDestinationVertexID().write(out);
          if (edge.getValue() != null) {
            out.writeBoolean(true);
            edge.getValue().write(out);
          } else {
            out.writeBoolean(false);
          }
        }
      }
    } else if (isRecoverReplicaMessage()) {
      this.vertex.write(out);
      peerIndex.write(out);
      
      if (edges == null)
        out.writeBoolean(false);
      else {
        out.writeBoolean(true);
        out.writeInt(edges.size());
        for (Object e : edges) {
          Edge<?, ?> edge = (Edge<?, ?>) e;
          edge.getDestinationVertexID().write(out);
          if (edge.getValue() != null) {
            out.writeBoolean(true);
            edge.getValue().write(out);
          } else {
            out.writeBoolean(false);
          }
        }
      }
    } else if (isRegisterMessage()) {
      peerIndex.write(out);
      vertexId.write(out);
      localID.write(out);
    } else if (isTestMessage()) {
      id.write(out);
      value.write(out);
    } else {
      LOG.error("Unsupported message type (flag: " + this.flag + "");
      throw new UnsupportedOperationException("Unsupported message type (flag: " + this.flag + "");
    }
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    flag = in.readInt();
    if (isMapMessage()) {
      map = new MapWritable();
      map.readFields(in);
    } else if (isPartitioningMessage()) {
      Vertex<Writable, Writable, Writable> vertex = GraphJobRunner
          .newVertexInstance(VERTEX_CLASS, null);
      Writable vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      vertex.setVertexID(vertexId);
      if (in.readBoolean()) {
        Writable vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS,
            null);
        vertexValue.readFields(in);
        vertex.setNewValue(vertexValue); // updated, then commit to lastValue
      }
      int size = in.readInt();
      vertex.setEdges(new ArrayList<Edge<Writable, Writable>>(size));
      for (int i = 0; i < size; i++) {
        Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS,
            null);
        edgeVertexID.readFields(in);
        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
          edgeValue.readFields(in);
        }
        vertex.getEdges().add(
            new Edge<Writable, Writable>(edgeVertexID, edgeValue));
      }
      this.vertex = vertex;
    } else if (isSizeMessage()) {
      numberVertices = new IntWritable();
      numberVertices.readFields(in);
      numberReplicas = new IntWritable();
      numberReplicas.readFields(in);
    } else if (isRepairMessage()) {
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
    } else if (isReplicaMessage()) {
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      vertexValue.readFields(in);
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
      
      if (in.readBoolean()) {
        int size = in.readInt();
        List<Edge<Writable, Writable>> tmpEdges = new ArrayList<Edge<Writable, Writable>>(size);
        for (int i = 0; i < size; i++) {
          Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS,
                                                              null);
          edgeVertexID.readFields(in);
          Writable edgeValue = null;
          if (in.readBoolean()) {
            edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
            edgeValue.readFields(in);
          }
          tmpEdges.add(new Edge<Writable, Writable>(edgeVertexID, edgeValue));
        }
      
        this.edges = tmpEdges;
      } else {
        this.edges = null;
      }
    } else if (isRecoverReplicaMessage()) {
      Vertex<Writable, Writable, Writable> vertex = GraphJobRunner.newVertexInstance(VERTEX_CLASS, null);
      Writable vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertex.setVertexID(vertexId);
      vertex.readFields(in);
      this.vertex = vertex;
      
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
      
      if (in.readBoolean()) {
        int size = in.readInt();
        List<Edge<Writable, Writable>> tmpEdges = new ArrayList<Edge<Writable, Writable>>(size);
        for (int i = 0; i < size; i++) {
          Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS,
                                                              null);
          edgeVertexID.readFields(in);
          Writable edgeValue = null;
          if (in.readBoolean()) {
            edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
            edgeValue.readFields(in);
          }
          tmpEdges.add(new Edge<Writable, Writable>(edgeVertexID, edgeValue));
        }
      
        this.edges = tmpEdges;
      } else {
        this.edges = null;
      }
    } else if (isRegisterMessage()) {
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      localID = new IntWritable();
      localID.readFields(in);      
    } else if (isTestMessage()) {
      id = new LongWritable();
      id.readFields(in);
      value = new DoubleWritable();
      value.readFields(in);      
    } else {
      LOG.error("Unsupported message type (flag: " + this.flag + "");
      throw new UnsupportedOperationException("Unsupported message type (flag: " + this.flag + "");
    }
  }

  public MapWritable getMap() {
    return map;
  }

  public Writable getVertexId() {
    return vertexId;
  }

  public Writable getVertexValue() {
    return vertexValue;
  }
  
  public Vertex<?, ?, ?> getVertex() {
    return vertex;
  }

  public IntWritable getVerticesSize() {
    return numberVertices;
  }
  
  public IntWritable getReplicasSize() {
    return numberReplicas;
  }
  
  public IntWritable getPeerIndex() { // (BSP-g)
    return peerIndex;
  }

  
  public List<?> getEdges() {
    return edges;
  }

  public IntWritable getLocalID() {
    return localID;
  }
  
  public LongWritable getID() {
    return id;
  }
  
  public DoubleWritable getValue() {
    return value;
  }
  
  public boolean isMapMessage() {
    return flag == MAP_FLAG;
  }

  public boolean isRepairMessage() {
    return flag == REPAIR_FLAG;
  }

  public boolean isPartitioningMessage() {
    return flag == PARTITION_FLAG;
  }

  public boolean isSizeMessage() {
    return flag == SIZE_FLAG;
  }
  
  public boolean isReplicaMessage() {
    return flag == REPLICA_FLAG;
  }
  
  public boolean isRegisterMessage() {
    return flag == REGISTER_FLAG;
  }
  
  public boolean isRecoverReplicaMessage() {
    return flag == RECOVER_REPLICA_FLAG;
  }
  
  public boolean isTestMessage() {
    return flag == TEST_FLAG;
  }
  
  @Override
  public String toString() {
    return "GraphJobMessage [flag=" + flag + ", map=" + map 
        + ", vertexId=" + vertexId + ", vertexValue=" + vertexValue 
        + ", vertex=" + vertex + ", #vertices=" + numberVertices + ", #replicas=" + numberReplicas
        + ", peerIndex=" + peerIndex
        + "]";
  }

}
