/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hama.bsp.AbstractVertex;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.LyraMessageBundle;
import org.apache.hama.bsp.LyraSyncMessage;
import org.apache.hama.bsp.Partitioner;
import org.apache.hama.graph.Edge;
import org.junit.Assert;

public abstract class Vertex<V extends Writable, E extends Writable, M extends Writable>
    extends AbstractVertex<V, E, M> /*implements VertexInterface<V, E, M>*/ {
  public static final Log LOG = LogFactory.getLog(Vertex.class);

  public static boolean needEdgeState = false;
  
  GraphJobRunner<?, ?, ?> runner;
  
  private List<Edge<V, E>> edges;
  private List<Edge<V, E>> edgesBack;
  private Map<V, Integer> neighborsPeerIndexMap;
  private boolean selectiveActivate = false;
  
  // TODO: double state in in/out edges
  // in-edges, only used by master vertex
  private List<RefEdge<V, E, M>> inEdges = new ArrayList<RefEdge<V, E, M>>();
  // out-edges in local
  //private List<RefEdge<V, E, M>> outEdges = new ArrayList<RefEdge<V, E, M>>() ;

  // used by master
  private int numberReplicas = 0;
  private int[] replicaIDs; // use peerIndex as array index, -1L means no replica in that peer
  
  // used by replica
  private int masterIndex = -1;
  private int localID = -1;
  
  //activate state (out-edges)
  private boolean activated = false; // activate vertex or not?
  private boolean activateAllNeighbors = false; // if true, activatedVertices should be empty
  //private List<Writable> activatedVertices = new ArrayList<Writable>(); // the list of vertices to activate
  //private List<Writable> activatedVertices = null; // the list of vertices to activate
  private List<Writable>[] activateLists = null;

  private boolean votedToHalt = false;

  private int lastCheckPointIter = 0;

  private boolean statesChanged = true;
  
  private Set<V> awakedSet = null;
  
  public Configuration getConf() {
    return runner.getPeer().getConfiguration();
  }

  //@Override
  public void setup(Configuration conf) {
    Assert.assertTrue(isMaster());
  }
  
  /**
   * The user-defined function
   */
  public void compute(Iterator<M> messages) throws IOException {}


  /**
   * @return the destination peer name of the destination of the given directed
   *         edge.
   */
  public String getDestinationPeerName(Edge<V, E> edge) {
    return getDestinationPeerName(edge.getDestinationVertexID());
  }

  /**
   * @return the destination peer name of the given vertex id, determined by the
   *         partitioner.
   */
  public String getDestinationPeerName(V vertexId) {
    return runner.getPeer().getPeerName(
        getPartitioner().getPartition(vertexId, value,
            runner.getPeer().getNumPeers()));
  }

  //@Override
  public void sendMessageToNeighbors(M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public void sendMessage(V destinationVertexID, M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public void sendMessage(Edge<V, E> e, M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public long getSuperstepCount() {
    return runner.getNumberIterations();
  }

  // edges
  public void setEdges(List<Edge<V, E>> list) {
    this.edges = list;
  }

  public void addEdge(Edge<V, E> edge) {
    if (edges == null)
      this.edges = new ArrayList<Edge<V, E>>();
    this.edges.add(edge);
  }

  //@Override
  public List<Edge<V, E>> getEdges() {
    return edges;
  }
  
  public void clearEdges() {
    this.edges.clear();
    this.edges = null;
  }
  
  public void addNeighborMap(V vertexID, int peerIndex) {
    neighborsPeerIndexMap.put(vertexID, new Integer(peerIndex));
  }
  
  public int getNeighorPeerIndex(V vertexID) {
    return neighborsPeerIndexMap.get(vertexID).intValue();
  }
  
  @SuppressWarnings("unchecked")
  public void enableSelectiveActivate() {
    selectiveActivate = true;
    int numPeers = getPeer().getNumPeers();
    activateLists = new List[numPeers];
    for (int i = 0; i < numPeers; i++)
      activateLists[i] = new ArrayList<Writable>();
    neighborsPeerIndexMap = new HashMap<V, Integer>();
  }

  //@Override
  public M getValue() { // only used by APP
    return lastValue;
  }

  //@Override
  public void setValue(M value) { // only used by APP
    this.value = value;
    this.updated = true;
  }
  
  //@Override
  public M getMessage() { // only used by APP
    return lastMessage;
  }

  public int getMaxIteration() {
    return runner.getMaxIteration();
  }

  /**
   * Get the last aggregated value of the defined aggregator, null if nothing
   * was configured or not returned a result. You have to supply an index, the
   * index is defined by the order you set the aggregator classes in
   * {@link GraphJob#setAggregatorClass(Class...)}. Index is starting at zero,
   * so if you have a single aggregator you can retrieve it via
   * {@link #getLastAggregatedValue}(0).
   */
  @SuppressWarnings("unchecked")
  public M getLastAggregatedValue(int index) {
    return (M) runner.getLastAggregatedValue(index);
  }

  /**
   * Get the number of aggregated vertices in the last superstep. Or null if no
   * aggregator is available.You have to supply an index, the index is defined
   * by the order you set the aggregator classes in
   * {@link GraphJob#setAggregatorClass(Class...)}. Index is starting at zero,
   * so if you have a single aggregator you can retrieve it via
   * {@link #getNumLastAggregatedVertices}(0).
   */
  public IntWritable getNumLastAggregatedVertices(int index) {
    return runner.getNumLastAggregatedVertices(index);
  }

  public int getNumPeers() {
    return runner.getPeer().getNumPeers();
  }

  /**
   * Gives access to the BSP primitives and additional features by a peer.
   */
  public BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> getPeer() {
    return runner.getPeer();
  }

  /**
   * @return the configured partitioner instance to message vertices.
   */
  @SuppressWarnings("unchecked")
  public Partitioner<V, M> getPartitioner() {
    return (Partitioner<V, M>) runner.getPartitioner();
  }

  //@Override
  public long getNumVertices() {
    return runner.getNumberVertices();
  }

  //@Override
  public void voteToHalt() {
    this.votedToHalt = true;
  }

  void setActive() {
    this.votedToHalt = false;
  }

  public boolean isHalted() {
    return votedToHalt;
  }
  
  public boolean isConverged() {
    return !activated;
  }

  /**
   * (BSP-g) default implementation
   */
  //@Override
  public boolean merge(Vertex<V, E, M> vertex) {
    Assert.assertEquals("can't merge vertex:" + this.getVertexID() 
        + " with vertex:" + vertex.getVertexID(), this, vertex);
    for (Edge<V, E> e: vertex.getEdges())
        this.addEdge(e);
    return true;
  }

  /**
   * (BSP-g) add peer index with the replica of vertex
   */
  public void addReplicaLocalID(int peerIndex, int localID) {
    if (this.numberReplicas == 0) { // init replica set
      this.replicaIDs = new int[getPeer().getNumPeers()];
      for (int i = 0; i < this.replicaIDs.length; i++)
        this.replicaIDs[i] = -1;
    }

    this.replicaIDs[peerIndex] = localID;
    this.numberReplicas++;
  }
  
  public void setMasterPeerIndex(int peerIndex) {
    this.masterIndex = peerIndex;
  }

  public int getMasterPeerIndex() {
    return this.masterIndex;
  }
  
  public void setLocalID(int id) {
    this.localID = id;
  }

  public int getLocalID() {
    return this.localID;
  }

  public long remoteSync(LyraMessageBundle[] bundles) throws IOException {
    if ((numberReplicas > 0)) {
      for (int i=0; i < this.replicaIDs.length; i++) {
        if (this.replicaIDs[i] != -1) {
          bundles[i].addMessage(new LyraSyncMessage(
              new IntWritable(this.replicaIDs[i]), 
              this.value, this.message, this.activated, this.activateAllNeighbors, this.activateLists[i]));
        }
      }
      return numberReplicas;
    }
    return 0;
  }
  
  /**
   * (BSP-g) sync vertex with replicas
   * @return #local-message
   */
  public long localSync() throws IOException {
    if (this.activated) { // sync with activate
      if (this.activateAllNeighbors) {
        return this.awakeNeighbors();
      } else {
        if (!selectiveActivate)
          throw new UnsupportedOperationException("forget to enable 'hama.graph.selective.activate'");
        return this.awake(this.activateLists[getPeer().getPeerIndex()]);
      }
    }
    return 0;
  }
  
  /**
   * (BSP-g) be called before compute and after message parsing
   */
  //@Override
  public boolean commit() {
    //reset vertex state
    if (this.updated) {
      this.lastValue = this.value;
      this.lastMessage = this.message;
      this.updated = false;
    }
    
    //reset awake state
    if (this.awaked) {
      this.votedToHalt = false;
      for (RefEdge<V, E, M> e: this.inEdges) {
        e.commit();
      }
      this.awaked = false;
    }
    
    //reset activate state
    if (this.activated) {
      if (selectiveActivate && !activateAllNeighbors) {
        for (List<Writable> list : activateLists)
          list.clear();
      }
      activateAllNeighbors = false;
      activated = false;
    }
    
    return !this.votedToHalt;
  }
  
  //@Override
  public void activate(Edge<V, E> e, M msg) throws IOException {
    activate(e.getDestinationVertexID(), msg);
  }
  
  //@Override
  public void activate(V destinationVertexID, M msg) throws IOException {
    if (!selectiveActivate)
      throw new UnsupportedOperationException("forget to enable 'hama.graph.selective.activate'");
    
    this.activated = true;
    int peerIndex = getNeighorPeerIndex(destinationVertexID);
    this.activateLists[peerIndex].add(destinationVertexID);
    
    this.message = msg;
    this.updated = true;
  }
  
  //@Override
  public void activateNeighbors(M msg) throws IOException {
    this.activated = true;
    this.activateAllNeighbors = true;
    
    this.message = msg;
    this.updated = true;
  }
  
  /**
   * (BSP-g) set in-edges
   */
  public void setInEdges(List<RefEdge<V, E, M>> list) {
    this.inEdges = list;
  }

  /**
   * (BSP-g) add an in-edge
   */
  public void addInEdge(RefEdge<V, E, M> edge) {
    this.inEdges.add(edge);
  }
  
  /**
   * (BSP-g) get the in-edges
   * @return a list of in-edges of this vertex
   */
  public List<RefEdge<V, E, M>> getInEdges() {
    return this.inEdges;
  }
  
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((vertexID == null) ? 0 : vertexID.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Vertex<?, ?, ?> other = (Vertex<?, ?, ?>) obj;
    if (vertexID == null) {
      if (other.vertexID != null)
        return false;
    } else if (!vertexID.equals(other.vertexID))
      return false;
    return true;
  }

  public int getLastCheckPointIter() {
    return this.lastCheckPointIter;
  }

  public void setLastCheckPointIter(int lastIter) {
    this.lastCheckPointIter = lastIter;
  }

  public boolean isAwaked() {
    return this.awaked;
  }

  public boolean isChanged() {
    return this.statesChanged;
  }

  public void setChanged(boolean changed) {
    this.statesChanged = changed;
  }

  public void writeEdges(DataOutput out) throws java.io.IOException {
    this.vertexID.write(out);
    
    List<?> outEdges = this.edgesBack;
    out.writeInt(outEdges.size());
    for (Object e : outEdges) {
      Edge<?, ?> edge = (Edge<?, ?>) e;
      edge.getDestinationVertexID().write(out);
      if (edge.getValue() != null) {
        out.writeBoolean(true);
        edge.getValue().write(out);
      } else {
        out.writeBoolean(false);
      }
    }

    out.writeInt(this.numberReplicas);
    if (this.replicaIDs != null) {
      out.writeBoolean(true);
      out.writeInt(this.replicaIDs.length);
      for (int i = 0; i < this.replicaIDs.length; i++) {
        out.writeInt(this.replicaIDs[i]);
      }
    } else {
      out.writeBoolean(false);
    }
  }

  public void readEdges(DataInput in) throws IOException {
    this.vertexID = this.getNewInstanceOfV();
    this.vertexID.readFields(in);

    int size = in.readInt();
    this.setEdges(new ArrayList<Edge<V, E>>(size));
    for (int i = 0; i < size; i++) {
      V edgeVertexID = this.getNewInstanceOfV();
      edgeVertexID.readFields(in);
      E edgeValue = null;
      if (in.readBoolean()) {
        edgeValue = this.getNewInstanceOfE();
        edgeValue.readFields(in);
      }
      this.getEdges().add(
			  new Edge<V, E>(edgeVertexID, edgeValue));
    }

    this.numberReplicas = in.readInt();
    if (in.readBoolean()) {
      size = in.readInt();
      this.replicaIDs = new int[size];
      for (int i = 0; i < size; i++) {
        this.replicaIDs[i] = in.readInt();
      }
    } else {
      this.replicaIDs = null;
    }
  }

  public void storeOrigEdges() {
    this.edgesBack = this.edges;
  }

  public void write(DataOutput out) throws java.io.IOException {
    super.write(out);
    
    
    out.writeBoolean(activated);
    out.writeBoolean(activateAllNeighbors);
    out.writeBoolean(votedToHalt);
    
    
    boolean needActivateList = needEdgeState;
    if (needActivateList) {
      out.writeBoolean(selectiveActivate);
      if (selectiveActivate) {
	if (activateLists == null) {
	  out.writeInt(-1);
	  LOG.debug("null list");
	} else {
	  out.writeInt(activateLists.length);
	  for (int i = 0; i < activateLists.length; i++) {
	    if (activateLists[i] != null) {
	      out.writeInt(activateLists[i].size());
	      for (int j = 0; j < activateLists[i].size(); j++) {
		activateLists[i].get(j).write(out);
		if (activateLists[i].get(j) == null)
		  LOG.debug("null vertex?");
	      }
	    } else {
	      out.writeInt(-1);
	    }
	  }
	}
      }
    }
    
    
    //write awake info about in edges
    boolean needEdgeInfo = needEdgeState;
    List<RefEdge<V, E, M>> edges = null;
    if (needEdgeInfo)
      edges = this.getInEdges();
    if (edges != null) {
      out.writeBoolean(true);
      int numOfAwake = 0;
      for (RefEdge<V, E, M> e : edges) {
        if (e.isAwake())
          numOfAwake++;
      }
      out.writeInt(numOfAwake);
      for (RefEdge<V, E, M> e : edges) {
        if (e.isAwake()) {
          e.getVertexID().write(out);
        }
      }
    } else {
      out.writeBoolean(false);
    }
   
    
  }
  
  @SuppressWarnings("unchecked")
  public void readFields(DataInput in) throws java.io.IOException {
    super.readFields(in);
    
    activated = in.readBoolean();
    activateAllNeighbors = in.readBoolean();
    votedToHalt = in.readBoolean();
    
    boolean needActivateList = needEdgeState;
    
    if (needActivateList) {
      selectiveActivate = in.readBoolean();
      if (selectiveActivate) {
	int listLen = in.readInt();
	if (listLen == -1)
	  activateLists = null;
	else {
	  activateLists = new List[listLen];
	  for (int i = 0; i < listLen; i++)
	    activateLists[i] = new ArrayList<Writable>();
	  for (int i = 0; i < listLen; i++) {
	    int len = in.readInt();
	    if (len == -1)
	      activateLists[i] = null;
	    else {
	      for (int j = 0; j < len; j++) {
		activateLists[i].add(j, this.getNewInstanceOfV());
		activateLists[i].get(j).readFields(in);
	      }
	    }
	  }
	}
      }
    }
    
    if (in.readBoolean()) {
      awakedSet = new HashSet<V>();
      int num = in.readInt();
      for (int i = 0; i < num; i++) {
        V tmpID = this.getNewInstanceOfV();
        tmpID.readFields(in);
        awakedSet.add(tmpID);
      }
    }
    
  }
  
  public void reAwakeInEdges() {
    boolean needEdgeInfo = needEdgeState;
    if (!needEdgeInfo)
      return;
    if (this.getInEdges() != null) {
      for (RefEdge<V, E, M> e : this.getInEdges()) {
        if (this.awakedSet.contains(e.getVertexID())) {
          e.activate();
          e.commit();
        }
      }
    } else {
      LOG.debug(this.getVertexID() + " empty in edge");
    }
  }
  
  @Override
  public String toString() {
    return getVertexID() + (isMaster() ? "(Master)" : "(Replica)")
        + " value" + (this.value != null ? "= " + this.value : " = NULL") 
        + " lastValue=" + (this.lastValue != null ? " (" + this.lastValue + ")" : " (NULL)")
        + " message=" + (this.message != null ? " (" + this.message + ")" : " (NULL)")
        + " lastMessage" + (this.lastMessage != null ? " (" + this.lastMessage + ")" : " (NULL)")
        + " awaked = " + (this.awaked)
        + " votedToHalt = " + (this.votedToHalt)
        + " updated = " + (this.updated)
        + " activated = " + this.activated
        + " // " + edges + " in{" + inEdges + "} lout{" + outEdges + "}";
  }

}
