/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.Writable;

public abstract class AbstractVertex<V extends Writable, E extends Writable, M extends Writable> {
  
  public static enum VertexType {
    MASTER, REPLICA, MIRROR
  }
  
  public VertexType vertexType;
  public V vertexID;
  

  public boolean votedToHalt = false; // modified by vertex itself
  
  public boolean awaked = false; // modify by others

  public M value = null;
  
  public M msg = null;
  
  //activate state (out-edges)
  public boolean activate = false; // activate vertex or not?
  public boolean activateAllNeighbors = false; // if true, activatedVertices should be empty

  public List<V>[] activateLists = null;
  
  /************************************************************/
  
  public boolean nVotedToHalt = false; // modified by vertex itself
  
  public boolean nAwaked = false; // modify by others

  public M nValue = null;
  
  public M nMsg = null;
  
  //activate state (out-edges)
  public boolean nActivate = false; // activate vertex or not?
  public boolean nActivateAllNeighbors = false; // if true, activatedVertices should be empty

  public List<V>[] nActivateLists = null;

  
  
  // out-edges in local peer
  protected List<AbstractRefEdge<V, E, M>> outEdges = new ArrayList<AbstractRefEdge<V, E, M>>() ;
  
  protected abstract M getNewInstanceOfM();
 

  public void setVertexType(VertexType type) {
    this.vertexType = type;
  }
  
  public VertexType getVertexType() {
    return this.vertexType;
  }
  
  public boolean isMaster() {
    return this.vertexType == VertexType.MASTER;
  }
  
  public V getVertexID() {
    return vertexID;
  }
  
  public void setVertexID(V vertexID) {
    this.vertexID = vertexID;
  }
  
  @SuppressWarnings("unchecked")
  public List<V>[] getnActivateLists(int n) {
    if (nActivateLists == null){
      nActivateLists = new List[n];
      for (int i = 0; i < n; i++)
        nActivateLists[i] = new ArrayList<V>();
    }
    return nActivateLists;
  }

  public M getLastValue() {
    return this.value;
  }
  
  public M getNewValue() {
    return this.nValue;
  }
  
  public void setLastValue(M value) {
    this.value = (M) value;
  }

  public void setNewValue(M value) {
    this.nValue = value;
  }

  public M getLastMessage() {
    return this.msg;
  }

  public M getNewMessage() {
    return this.nMsg;
  }
  
  public void setNewMsg(M msg) {
    this.nMsg = msg;
  }
  
  public void setLastMessage(M message) {
    this.msg = message;
  }


  public void setNextState(LyraSyncMessage<V, M> msg) {
    throw new UnsupportedOperationException("This method should be overide");
  }

  public void addOutEdge(AbstractRefEdge<V, E, M> edge) {
    this.outEdges.add(edge);
  }

  public List<AbstractRefEdge<V, E, M>> getOutEdges() {
    return this.outEdges;
  }
  
  public void setOutEdges(List<AbstractRefEdge<V, E, M>> outEdges) {
    this.outEdges = outEdges;
  }


  public boolean hasOutEdge(AbstractRefEdge<V, E, M> edge) {
    return this.outEdges.contains(edge);
  }

  /****************************************************************************************************************/
  /*
   * These two methods are called during the commit point, so the state is the newest state
   */
  public long awake(List<V> list) {
    int num = 0;
    for (V id : list) {
      for (AbstractRefEdge<V, E, M> e : this.outEdges) {
        if (e.getVertexID().equals((V) id)) {
          AbstractVertex<V, E, M> neighbor = e.getVertex();
          neighbor.nAwaked = true;
          e.getPair().activate();
          num++;
          break;
        }
      }
    }
    return num;
  }
  
  /**
   * (BSP-g) awake neighbors
   */  
  public long awakeNeighbors() {
    for (AbstractRefEdge<V, E, M> e: this.outEdges) {
      AbstractVertex<V, E, M> neighbor = e.getVertex();
      neighbor.nAwaked = true;
      e.getPair().activate();
    }
    return this.outEdges.size();
  }
  
  
  /**
   * (BSP-g) sync vertex with replicas
   * since called after commitState, the state is the newest state
   * @return #local-message
   */
  public long localActivateVertices() throws IOException {
    throw new UnsupportedOperationException("This method should be overide");
  }
  
  /****************************************************************************************************************/
  public abstract int getLocalID();
  
  public abstract void writeRecoverMasterMsg(DataOutput out) throws IOException;
  
  public abstract void readRecoverMasterMsg(DataInput in) throws IOException ;
  
  public abstract void writeRecoverReplicaMsg(DataOutput out,  int peerIndex, boolean isMirror) throws IOException;
  
  public abstract void readRecoverReplicaMsg(DataInput in,  int peerIndex, boolean isMirror) throws IOException;
  
  public abstract void writeMirrorInfo(DataOutput out) throws IOException;
  
  public abstract void readMirrorInfo(DataInput in) throws IOException;
}
