/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;

public class AbstractVertexRuntimeState<V extends Writable, M extends Writable> implements Writable{

  public static Class<? extends Writable> VERTEX_ID_CLASS;
  public static Class<? extends Writable> VERTEX_VALUE_CLASS;
  
  public long iteration = -1;
  
  public boolean votedToHalt = false; // modified by vertex itself
  
  public boolean awaked = false; // modify by others


  public M value = null;
  
  public M msg = null;
  
  //activate state (out-edges)
  public boolean activated = false; // activate vertex or not?
  public boolean activateAllNeighbors = false; // if true, activatedVertices should be empty
  //private List<Writable> activatedVertices = new ArrayList<Writable>(); // the list of vertices to activate
  //private List<Writable> activatedVertices = null; // the list of vertices to activate
  public List<V>[] activateLists = null;
  
  public AbstractVertexRuntimeState(M v, M m, long iteration){
    this.value = v;
    this.msg = m;
    this.iteration = iteration;
  }
  
  public AbstractVertexRuntimeState<V, M> reInit(M v, M m, long iteration, byte lastActive){
    this.value = v;
    this.msg = m;
    this.iteration = iteration;
   
    activated = false;
    activateAllNeighbors = false;
    activateLists = null;
    
    votedToHalt = false;
    awaked = false;
    
    return this;
  }
  
  public AbstractVertexRuntimeState(){}
  
  public AbstractVertexRuntimeState(LyraSyncMessage<V, M> syncMsg) {
    
    votedToHalt = syncMsg.votedToHalt;
    
    awaked = syncMsg.awaked;
    
    value = syncMsg.value;
    
    msg = syncMsg.msg;
    
    activated = syncMsg.activated; 
    
    activateAllNeighbors = syncMsg.activateAllNeighbors; 

    activateLists = syncMsg.activateLists;
  }

  @SuppressWarnings("unchecked")
  public List<V>[] getActivateLists(int n){
    if (activateLists == null){
      activateLists = new List[n];
      for (int i = 0; i < n; i++)
        activateLists[i] = new ArrayList<V>();
    }
    return activateLists;
  }
  
  
  @Override
  public void write(DataOutput out) throws IOException {
    //out.writeLong(iteration);

    out.writeBoolean(votedToHalt);
    out.writeBoolean(awaked);
    value.write(out);
    msg.write(out);
    if (activated == false){
      out.writeBoolean(false);
    } else {
      out.writeBoolean(true);
      if (activateAllNeighbors) {
        out.writeBoolean(true);
      } else {
        out.writeBoolean(false);
        out.writeInt(activateLists.length);
        for (int i = 0; i < activateLists.length; i++){
          int size = activateLists[i].size();
          out.writeInt(size);
          for (Writable destinationVertexId : activateLists[i]) {
            destinationVertexId.write(out);
          }
        }
      }
    }
  }
  
  @SuppressWarnings("unchecked")
  @Override
  public void readFields(DataInput in) throws IOException {
    //iteration = in.readLong();

    votedToHalt = in.readBoolean();
    awaked = in.readBoolean();
    value = (M) ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
    value.readFields(in);
    msg = (M) ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
    msg.readFields(in);
    if (in.readBoolean() == false){
      activated = false;
    } else {
      activated = true;
      if (in.readBoolean()) {
        activateAllNeighbors = true;
        activateLists = null;
      } else {
        activateAllNeighbors = false;
        int length = in.readInt();
        activateLists = new List[length];
        for (int j = 0; j < length; j++){
          int size = in.readInt();
          activateLists[j] = new ArrayList<V>(size);
          for (int i = 0; i < size; i++) {
            V activatedVertexId = (V) ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
            activatedVertexId.readFields(in);
            activateLists[j].add(activatedVertexId);
          }
        }
      }
    }
  }
  
  
  
  
  public String toString(){
    return "iteration=" + iteration + "  votedToHalt=" + votedToHalt + " awaked=" + awaked + " value=" + value + " msg=" + msg + " activated=" + activated + " activateAllNeighbors=" + activateAllNeighbors; 
  }
 
}