/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.LocalFileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.BytesWritable;
import org.apache.hadoop.io.DataInputBuffer;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.Constants;
import org.apache.hama.bsp.Counters.Counter;
import org.apache.hama.bsp.ft.FaultTolerantPeerService;
import org.apache.hama.bsp.message.LyraMessageManager;
import org.apache.hama.bsp.message.LyraMessageManagerFactory;
import org.apache.hama.bsp.message.LyraRecoverSender;
import org.apache.hama.bsp.message.LyraRecoverSenderFactory;
import org.apache.hama.bsp.message.MessageManager;
import org.apache.hama.bsp.message.MessageManagerFactory;
import org.apache.hama.bsp.message.queue.MessageQueue;
import org.apache.hama.bsp.sync.PeerSyncClient;
import org.apache.hama.bsp.sync.SyncException;
import org.apache.hama.bsp.sync.SyncServiceFactory;
import org.apache.hama.ipc.BSPPeerProtocol;
import org.apache.hama.util.KeyValuePair;

/**
 * This class represents a BSP peer.
 */
public final class BSPPeerImpl<K1, V1, K2, V2, M extends Writable> implements
    BSPPeer<K1, V1, K2, V2, M> {

  private static final Log LOG = LogFactory.getLog(BSPPeerImpl.class);

  public static enum PeerCounter {
    COMPRESSED_MESSAGES, SUPERSTEP_SUM, TASK_INPUT_RECORDS, TASK_OUTPUT_RECORDS, IO_BYTES_READ, MESSAGE_BYTES_TRANSFERED, MESSAGE_BYTES_RECEIVED, TOTAL_MESSAGES_SENT, TOTAL_MESSAGES_RECEIVED, COMPRESSED_BYTES_SENT, COMPRESSED_BYTES_RECEIVED, TIME_IN_SYNC_MS
  }

  private final Configuration conf;
  private final FileSystem fs;
  private BSPJob bspJob;

  private TaskStatus currentTaskStatus;

  private TaskAttemptID taskId;
  private BSPPeerProtocol umbilical;

  private String[] allPeers;

  // SYNC
  private PeerSyncClient syncClient;
  private MessageManager<M> messenger;
  private LyraMessageManager lyraMessenger;
  private LyraMessageManager[] lyraMessengerMT;
  private LyraMessageManager[] lyraRecoverRecievers;
  private LyraRecoverSender[] lyraRecoverSenders;
  private int numberLyraMessageManagers = 0;

  // IO
  private int partition;
  private String splitClass;
  private BytesWritable split;
  private OutputCollector<K2, V2> collector;
  private RecordReader<K1, V1> in;
  private RecordWriter<K2, V2> outWriter;
  private final KeyValuePair<K1, V1> cachedPair = new KeyValuePair<K1, V1>();

  private InetSocketAddress peerAddress;

  private Counters counters;
  private Combiner<M> combiner;

  private FaultTolerantPeerService<M> faultToleranceService;

  private long splitSize = 0L;

  // stat
  private long onceSYN = 0, onceSND = 0;
  
  /**
   * Protected default constructor for LocalBSPRunner.
   */
  protected BSPPeerImpl() {
    conf = null;
    fs = null;
  }

  /**
   * For unit test.
   * 
   * @param conf is the configuration file.
   * @param dfs is the Hadoop FileSystem.
   */
  protected BSPPeerImpl(final Configuration conf, FileSystem dfs) {
    this.conf = conf;
    this.fs = dfs;
  }

  /**
   * For unit test.
   * 
   * @param conf is the configuration file.
   * @param dfs is the Hadoop FileSystem.
   * @param counters is the counters from outside.
   */
  public BSPPeerImpl(final Configuration conf, FileSystem dfs, Counters counters) {
    this(conf, dfs);
    this.counters = counters;
  }

  public BSPPeerImpl(BSPJob job, Configuration conf, TaskAttemptID taskId,
      BSPPeerProtocol umbilical, int partition, String splitClass,
      BytesWritable split, Counters counters) throws Exception {
    this(job, conf, taskId, umbilical, partition, splitClass, split, counters,
        -1, TaskStatus.State.RUNNING);
  }

  /**
   * BSPPeer Constructor.
   * 
   * BSPPeer acts on behalf of clients performing bsp() tasks.
   * 
   * @param conf is the configuration file containing bsp peer host, port, etc.
   * @param umbilical is the bsp protocol used to contact its parent process.
   * @param taskId is the id that current process holds.
   * @throws Exception
   */
  @SuppressWarnings("unchecked")
  public BSPPeerImpl(BSPJob job, Configuration conf, TaskAttemptID taskId,
      BSPPeerProtocol umbilical, int partition, String splitClass,
      BytesWritable split, Counters counters, long superstep,
      TaskStatus.State state) throws Exception {
    this.conf = conf;
    this.taskId = taskId;
    this.umbilical = umbilical;
    this.bspJob = job;
    // IO
    this.partition = partition;
    this.splitClass = splitClass;
    this.split = split;
    this.counters = counters;

    this.fs = FileSystem.get(conf);

    String bindAddress = conf.get(Constants.PEER_HOST,
        Constants.DEFAULT_PEER_HOST);
    int bindPort = conf
        .getInt(Constants.PEER_PORT, Constants.DEFAULT_PEER_PORT);
    peerAddress = new InetSocketAddress(bindAddress, bindPort);
    LOG.info("PeerAddress: " + peerAddress + " port: " + bindPort);
    
    initializeIO();


    TaskStatus.Phase phase = TaskStatus.Phase.STARTING;
    String stateString = "running";
    if (state == TaskStatus.State.RECOVERING) {
      phase = TaskStatus.Phase.RECOVERING;
      stateString = "recovering";
    }

    setCurrentTaskStatus(new TaskStatus(taskId.getJobID(), taskId, 1.0f, state,
        stateString, peerAddress.getHostName(), phase, counters));

    initializeSyncService(superstep, state);
    initilizeMessaging();
    if (LOG.isDebugEnabled()) {
      LOG.debug("Initialized Messaging service.");
    }
    
    final String combinerName = conf.get("bsp.combiner.class");
    if (combinerName != null) {
      combiner = (Combiner<M>) ReflectionUtils.newInstance(
          conf.getClassByName(combinerName), conf);
    }
    /*
    if (conf.getBoolean(Constants.FAULT_TOLERANCE_FLAG, false)) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Fault tolerance enabled.");
      }
      if (superstep > 0)
        conf.setInt("attempt.superstep", (int) superstep);
      Class<?> ftClass = conf.getClass(Constants.FAULT_TOLERANCE_CLASS,
          AsyncRcvdMsgCheckpointImpl.class, BSPFaultTolerantService.class);
      if (ftClass != null) {
        if (superstep > 0) {
          counters.incrCounter(PeerCounter.SUPERSTEP_SUM, superstep);
        }

        this.faultToleranceService = ((BSPFaultTolerantService<M>) ReflectionUtils
            .newInstance(ftClass, null)).constructPeerFaultTolerance(job, this,
            syncClient, peerAddress, this.taskId, superstep, conf, messenger);
        TaskStatus.State newState = this.faultToleranceService
            .onPeerInitialized(state);

        if (state == TaskStatus.State.RECOVERING) {
          if (newState == TaskStatus.State.RUNNING) {
            phase = TaskStatus.Phase.STARTING;
            stateString = "running";
            state = newState;
          }

          setCurrentTaskStatus(new TaskStatus(taskId.getJobID(), taskId, 1.0f,
              state, stateString, peerAddress.getHostName(), phase, counters));
          if (LOG.isDebugEnabled()) {
            LOG.debug("State after FT service initialization - "
                + newState.toString());
          }

        }

        if (LOG.isDebugEnabled()) {
          LOG.debug("Initialized fault tolerance service");
        }
      }
    }
    */

    if (state != TaskStatus.State.RECOVERING) {
      doFirstSync(superstep);
    }
    
    if (LOG.isDebugEnabled()) {
      LOG.info(new StringBuffer("BSP Peer successfully initialized for ")
          .append(this.taskId.toString()).append(" ").append(superstep)
          .toString());
    }
  }
  
  /**
   * Transfers DistributedCache files into the local cache files. Also creates
   * symbolic links for URIs specified with a fragment if
   * DistributedCache.getSymlinks() is true.
   * 
   * @throws IOException If a DistributedCache file cannot be found.
   */
  public final void moveCacheFiles() throws IOException {
    StringBuilder files = new StringBuilder();
    boolean first = true;
    if (DistributedCache.getCacheFiles(conf) != null) {
      for (URI uri : DistributedCache.getCacheFiles(conf)) {
        if (uri != null) {
          if (!first) {
            files.append(",");
          }
          if (null != uri.getFragment() && DistributedCache.getSymlink(conf)) {

            FileUtil.symLink(uri.getPath(), uri.getFragment());
            files.append(uri.getFragment()).append(",");
          }
          FileSystem hdfs = FileSystem.get(conf);
          Path pathSrc = new Path(uri.getPath());
          if (hdfs.exists(pathSrc)) {
            LocalFileSystem local = FileSystem.getLocal(conf);
            Path pathDst = new Path(local.getWorkingDirectory(),
                pathSrc.getName());
            hdfs.copyToLocalFile(pathSrc, pathDst);
            files.append(pathDst.toUri().getPath());
          }
          first = false;
        }
      }
    }
    if (files.length() > 0) {
      DistributedCache.addLocalFiles(conf, files.toString());
    }
  }

  @SuppressWarnings("unchecked")
  public final void initInput() throws IOException {
    InputSplit inputSplit = null;
    // reinstantiate the split
    try {
      if (splitClass != null) {
        inputSplit = (InputSplit) ReflectionUtils.newInstance(
            getConfiguration().getClassByName(splitClass), getConfiguration());
      }
    } catch (ClassNotFoundException exp) {
      IOException wrap = new IOException("Split class " + splitClass
          + " not found");
      wrap.initCause(exp);
      throw wrap;
    }

    if (inputSplit != null) {
      DataInputBuffer splitBuffer = new DataInputBuffer();
      splitBuffer.reset(split.getBytes(), 0, split.getLength());
      inputSplit.readFields(splitBuffer);
      if (in != null) {
        in.close();
      }
      in = new TrackedRecordReader<K1, V1>(bspJob.getInputFormat()
          .getRecordReader(inputSplit, bspJob),
          getCounter(BSPPeerImpl.PeerCounter.TASK_INPUT_RECORDS),
          getCounter(BSPPeerImpl.PeerCounter.IO_BYTES_READ));
      this.splitSize = inputSplit.getLength();
    }
  }

  /**
   * @return the size of assigned split
   */
  @Override
  public long getSplitSize() {
    return splitSize;
  }

  /**
   * @return the position in the input stream.
   */
  @Override
  public long getPos() throws IOException {
    return in.getPos();
  }

  public final void initilizeMessaging() throws ClassNotFoundException {
    messenger = MessageManagerFactory.getMessageManager(conf);
    messenger.init(taskId, this, conf, peerAddress);
  }
  
  public final void initializeSyncService(long superstep, TaskStatus.State state)
      throws Exception {
    syncClient = SyncServiceFactory.getPeerSyncClient(conf);
    syncClient.init(conf, taskId.getJobID(), taskId);
    if (state == TaskStatus.State.RUNNING){
      syncClient.lyraRegister(taskId.getJobID(), taskId, peerAddress.getHostName(),
          peerAddress.getPort());
    } else if (state == TaskStatus.State.RECOVERING){
      syncClient.lyraDeregister(taskId.getJobID(), taskId);
      syncClient.lyraRegister(taskId.getJobID(), taskId, peerAddress.getHostName(),
          peerAddress.getPort());
    } else {
      throw new Exception("Error state during initializeSyncService ("+state+")");
    }
  }
  
  
  private void doFirstSync(long superstep) throws SyncException {
    if (superstep > 0)
      --superstep;
    syncClient.enterBarrier(taskId.getJobID(), taskId, superstep);
    syncClient.leaveBarrier(taskId.getJobID(), taskId, superstep);
  }

  @SuppressWarnings("unchecked")
  public final void initializeIO() throws Exception {

    initInput();

    String outdir = null;
    if (conf.get("bsp.output.dir") != null) {
      Path outputDir = new Path(conf.get("bsp.output.dir",
          "tmp-" + System.currentTimeMillis()), Task.getOutputName(partition));
      outdir = outputDir.makeQualified(fs).toString();
    }
    outWriter = bspJob.getOutputFormat().getRecordWriter(fs, bspJob, outdir);
    final RecordWriter<K2, V2> finalOut = outWriter;

    collector = new OutputCollector<K2, V2>() {
      @Override
      public void collect(K2 key, V2 value) throws IOException {
        finalOut.write(key, value);
      }
    };

    try {
      moveCacheFiles();
    } catch (Exception e) {
      LOG.error(e);
    }

  }

  @Override
  public final M getCurrentMessage() throws IOException {
    return messenger.getCurrentMessage();
  }

  @Override
  public final void send(String peerName, M msg) throws IOException {
    messenger.send(peerName, msg);
  }

  /*
   * (non-Javadoc)
   * @see org.apache.hama.bsp.BSPPeerInterface#sync()
   */
  @Override
  public final void sync() throws IOException, SyncException,
      InterruptedException {
    
    this.onceSND = System.currentTimeMillis();
    // normally all messages should been send now, finalizing the send phase
    messenger.finishSendPhase();
    Iterator<Entry<InetSocketAddress, MessageQueue<M>>> it = messenger
        .getMessageIterator();

    while (it.hasNext()) {
      Entry<InetSocketAddress, MessageQueue<M>> entry = it.next();
      final InetSocketAddress addr = entry.getKey();
      final Iterable<M> messages = entry.getValue();

      final BSPMessageBundle<M> bundle = combineMessages(messages);
      // remove this message during runtime to save a bit of memory
      it.remove();
      try {
        if (!bundle.isEmpty())
          messenger.transfer(addr, bundle);
      } catch (Exception e) {
        LOG.error("Error while sending messages", e);
      }
      //LOG.info("transfer msg (" + bundle.getMessages().size() 
      //      + ") to " + addr.getHostName() + ":"+ addr.getPort());
    }

    if (this.faultToleranceService != null) {
      try {
        this.faultToleranceService.beforeBarrier();
      } catch (Exception e) {
        throw new IOException(e);
      }
    }

    this.onceSYN = System.currentTimeMillis();
    this.onceSND = this.onceSYN - this.onceSND;
    
    // Peng: replace rpc sync with ft-zookerper sync
    //LOG.info("==============enter 1st barrier=========");
    this.lyraSync(Constants.BARRIER_NORMAL);
    //LOG.info("==============leave 1st barrier=========");

    
    if (this.faultToleranceService != null) {
      try {
        this.faultToleranceService.duringBarrier();
      } catch (Exception e) {
        throw new IOException(e);
      }
    }

    // Clear outgoing queues.
    messenger.clearOutgoingQueues();

    //LOG.info("==============enter 2nd barrier=========");
    this.lyraSync(Constants.BARRIER_NORMAL);
    //LOG.info("==============leave 2nd barrier=========");
    //barrier();
    //leaveBarrier();
    
    this.onceSYN = System.currentTimeMillis() - this.onceSYN;
    

    currentTaskStatus.setCounters(counters);

    if (this.faultToleranceService != null) {
      try {
        this.faultToleranceService.afterBarrier();
      } catch (Exception e) {
        throw new IOException(e);
      }
    }
  }
  
  @Override
  public final Iterator<Entry<InetSocketAddress, MessageQueue<M>>> 
      preSync() throws IOException {

    messenger.finishSendPhase();
    Iterator<Entry<InetSocketAddress, MessageQueue<M>>> it = messenger
        .getMessageIterator();

    return it;
  }
  
  @Override
  public final void transfer(
      Iterator<Entry<InetSocketAddress, MessageQueue<M>>> it) {

    while (true) {
      Entry<InetSocketAddress, MessageQueue<M>> entry = null;
      synchronized (it) {
        if (!it.hasNext())
          return;        
        entry = it.next();
        it.remove();
      }
      
      final InetSocketAddress addr = entry.getKey();
      final Iterable<M> messages = entry.getValue();
      final BSPMessageBundle<M> bundle = combineMessages(messages);
      
      try {
        messenger.transfer(addr, bundle);
      } catch (Exception e) {
        LOG.error("Error while sending messages", e);
      }
    }
  }
  

  /**
   * Lyra code
   */  
  @Override
  public final void initilizeLyraMessenger() throws ClassNotFoundException {
    lyraMessenger = LyraMessageManagerFactory.getMessageManager(conf);
    lyraMessenger.init(conf);
    lyraMessenger.setMyIdx(this.getPeerIndex());
    
  }
  
  @Override
  public final void initilizeLyraMessenger(int numberMessengers, int numberReceivers) 
        throws ClassNotFoundException {
    lyraMessengerMT = LyraMessageManagerFactory.getMessageManagers(conf, numberMessengers);
    for (int i = 0; i < numberMessengers; i++){
      lyraMessengerMT[i].init(conf, numberMessengers, i, numberReceivers);
      lyraMessengerMT[i].setMyIdx(this.getPeerIndex());
    }
    numberLyraMessageManagers = numberMessengers;
  }
  
  @Override
  public final void initilizeLyraRecoverMessenger(int rpcNumber) throws ClassNotFoundException {
    /**
     * after recovery, lyraRecoverRecievers except the first will be close
     */
    lyraRecoverRecievers = LyraMessageManagerFactory.getMessageManagers(conf, rpcNumber);
    for (int i = 0; i < rpcNumber; i++){
      lyraRecoverRecievers[i].init(conf, rpcNumber, i, rpcNumber);
      lyraRecoverRecievers[i].setMyIdx(this.getPeerIndex());
    }
  }
  
  @Override
  public final void initilizeLyraRecoverSender(int rpcNumber) throws ClassNotFoundException {
    lyraRecoverSenders = LyraRecoverSenderFactory.getRecoverSender(conf, rpcNumber);
    for (int i = 0; i < rpcNumber; i++){
      lyraRecoverSenders[i].init(conf, this.getPeerIndex()+i, rpcNumber); // id = id+i, which is tricky and is depend on the mod operation
    }
  }
  
  @Override
  public final void registerMessaging(AbstractVertex<?, ?, ?>[] replicas, 
      AbstractGlobalState stat) {
    lyraMessenger.register(replicas, stat);
  }
  
  @Override
  public final void registerMessaging(int index, AbstractVertex<?, ?, ?>[] replicas, 
      AbstractGlobalState stat) {
    lyraMessengerMT[index].register(replicas, stat);
  }
  
  @Override
  public final void prepare() {
    lyraMessenger.prepare();
  }
  
  @Override
  public final void prepare(int index) {
    lyraMessengerMT[index].prepare();
  }
  
  @Override
  public final boolean syncVertex(String peerName, LyraSyncBundle bundle) {
    try {
      lyraMessenger.syncVertex(peerName, bundle);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public final boolean syncVertex(int index, String peerName, LyraSyncBundle bundle) {
    try {
      lyraMessengerMT[index].syncVertex(peerName, bundle);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public final boolean syncStat(String peerName, MapWritable stat) {
    try {
      lyraMessenger.syncStat(peerName, stat);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public final boolean syncStat(int index, String peerName, MapWritable stat) {
    try {
      lyraMessengerMT[index].syncStat(peerName, stat);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public final boolean completeMirror(String peerName, LyraMessageBundle<LyraMirrorMessage> bundle) {
    try {
      lyraMessenger.completeMirror(peerName, bundle);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  
  @Override
  public boolean transferRecoveryMessage(int id, String peerName, LyraRecoverMessage msg){
    try {
      lyraRecoverSenders[id].transfer(peerName, msg);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public boolean clearRecoveryMessageBuffer(int id, String peerName){
    try {
      lyraRecoverSenders[id].clearTransferBuffer(peerName);
      return true;
    } catch (Exception e) {
      LOG.error("Error while sending messages", e);
    }
    return false;
  }
  
  @Override
  public List<LyraRecoverMessage> uglyGetMessages(){
    return lyraMessenger.getRecoverMessageBundle();
  }
  
  @SuppressWarnings("unchecked")
  @Override
  public List<LyraRecoverMessage>[] uglyCombineRecoverMessages(){
    int rpcServers = conf.getInt(Constants.RECOVER_RPC_NUM, 4);
    List<LyraRecoverMessage> finalMessages[] = new ArrayList[rpcServers];
    for (int i = 0; i < lyraRecoverRecievers.length; i++)
    {
      finalMessages[i] = lyraRecoverRecievers[i].getRecoverMessageBundle();
    }
    lyraMessenger = lyraRecoverRecievers[0];
    
    return finalMessages;
  }
  
  @Override
  public final void registerRecoverMessaging(AbstractVertex<?, ?, ?>[] replicas) {
    for (int i = 0; i < lyraRecoverRecievers.length; i++)
    {
      lyraRecoverRecievers[i].register(replicas, null);
    }
  }
  
  @Override
  public final int getNumCurrentLyraMessages() {
    return lyraMessenger.size();
  }

  @Override
  public final int getNumCurrentLyraMessages(int index) {
    return lyraMessengerMT[index].size();
  }

  
  @Override
  public final List<Integer> lyraSync(int state) throws IOException, 
        SyncException, InterruptedException {
    
    this.onceSYN = System.currentTimeMillis();
    List<Integer> result = barrier(state);
    this.onceSYN = System.currentTimeMillis() - this.onceSYN;
    
    incrementCounter(PeerCounter.TIME_IN_SYNC_MS, this.onceSYN);
    incrementCounter(PeerCounter.SUPERSTEP_SUM, 1L);
    
    currentTaskStatus.setCounters(counters);
    umbilical.statusUpdate(taskId, currentTaskStatus);
    
    if (result.size() > 0 && result.get(0) == Constants.NEWBIE_HANDLER){
      // new node enters, so we re-init the allPeers
      this.allPeers = null;
      this.initPeerNames();
    }
    
    return result;
  }
  
  @Override
  public final void lyraSAwakeOtherFromSync(int state) throws IOException, 
        SyncException, InterruptedException {
    
    this.onceSYN = System.currentTimeMillis();
    awakeOthersFromBarrier(state);
    this.onceSYN = System.currentTimeMillis() - this.onceSYN;
    
    incrementCounter(PeerCounter.TIME_IN_SYNC_MS, this.onceSYN);
    incrementCounter(PeerCounter.SUPERSTEP_SUM, 1L);
    
    currentTaskStatus.setCounters(counters);
    umbilical.statusUpdate(taskId, currentTaskStatus);
    
    // reinit is necessary, because others may also crash
    this.allPeers = null;
    this.initPeerNames();
  }

  private final BSPMessageBundle<M> combineMessages(Iterable<M> messages) {
    BSPMessageBundle<M> bundle = new BSPMessageBundle<M>();
    if (combiner != null) {
      bundle.addMessage(combiner.combine(messages));
    } else {
      for (M message : messages) {
        bundle.addMessage(message);
      }
    }
    return bundle;
  }

  protected final void enterBarrier() throws SyncException {
    syncClient.enterBarrier(taskId.getJobID(), taskId,
        currentTaskStatus.getSuperstepCount());
  }

  protected final void leaveBarrier() throws SyncException {
    syncClient.leaveBarrier(taskId.getJobID(), taskId,
        currentTaskStatus.getSuperstepCount());
  }
  
  protected final void awakeOthersFromBarrier(int state) throws SyncException {
    //replace RPCBarrier with zkBarrier
    long ss = syncClient.lyraAwakeOthersFromBarrier(taskId.getJobID(), taskId,
        currentTaskStatus.getSuperstepCount(), state);
    Counter ssCounter = currentTaskStatus.getCounters().findCounter(PeerCounter.SUPERSTEP_SUM);
    long delta = ss - ssCounter.getValue();     
    currentTaskStatus.getCounters().findCounter(PeerCounter.SUPERSTEP_SUM).increment(delta);
  }
  
  protected final List<Integer> barrier(int state) throws SyncException {
    //replace RPCBarrier with zkBarrier
    return syncClient.lyraBarrier(taskId.getJobID(), taskId,
        currentTaskStatus.getSuperstepCount(), state);
  }
  

  /**
   * Delete files from the local cache
   * 
   * @throws IOException If a DistributedCache file cannot be found.
   */
  public void deleteLocalFiles() throws IOException {
    if (DistributedCache.getLocalCacheFiles(conf) != null) {
      for (Path path : DistributedCache.getLocalCacheFiles(conf)) {
        if (path != null) {
          LocalFileSystem local = FileSystem.getLocal(conf);
          if (local.exists(path)) {
            local.delete(path, true); // recursive true
          }
        }
      }
    }
    DistributedCache.setLocalFiles(conf, "");
  }

  public void lyraRegisterInfo(String myInfo) throws SyncException {
    syncClient.lyraPersistentPut(taskId.getJobID(), taskId.getTaskID().getId(), myInfo);
  }
  
  public String lyraGetInfo() throws SyncException{
    return syncClient.lyraPersistentGet(taskId.getJobID(), taskId.getTaskID().getId()); 
  }
  
  public final void close() {
    // there are many catches, because we want to close always every component
    // even if the one before failed.
    if (in != null) {
      try {
        in.close();
      } catch (Exception e) {
        LOG.error(e);
      }
    }
    if (outWriter != null) {
      try {
        outWriter.close();
      } catch (Exception e) {
        LOG.error(e);
      }
    }
    this.clear();
    try {
      syncClient.close();
    } catch (Exception e) {
      LOG.error(e);
    }
    try {
      messenger.close();
    } catch (Exception e) {
      LOG.error(e);
    }
    
    try {
      if (numberLyraMessageManagers > 0)
        for (int i = 0; i < numberLyraMessageManagers; i++)
          lyraMessengerMT[i].close();
      else
        lyraMessenger.close();
    } catch (Exception e) {
      LOG.error(e);
    }
    try {
      if (lyraRecoverRecievers != null){
       for (int i = 1; i < lyraRecoverRecievers.length; i++)
         lyraRecoverRecievers[i].close();
      }
    } catch (Exception e) {
      LOG.error(e);
    }
    
    // Delete files from the local cache
    try {
      deleteLocalFiles();
    } catch (Exception e) {
      LOG.error(e);
    }
  }

  @Override
  public final void clear() {
    messenger.clearOutgoingQueues();
  }

  /**
   * @return the string as host:port of this Peer
   */
  @Override
  public final String getPeerName() {
    return peerAddress.getHostName() + ":" + peerAddress.getPort();
  }

  @Override
  public final String[] getAllPeerNames() {
    initPeerNames();
    return allPeers;
  }

  @Override
  public final String getPeerName(int index) {
    initPeerNames();
    return allPeers[index];
  }

  @Override
  public int getPeerIndex() {
    return this.taskId.getTaskID().getId();
  }

  @Override
  public final int getNumPeers() {
    initPeerNames();
    return allPeers.length;
  }

  private final void initPeerNames() {
    if (allPeers == null) {
      allPeers = syncClient.lyraGetAllPeerNames(taskId);
    }
  }

  /**
   * @return the number of messages
   */
  @Override
  public final int getNumCurrentMessages() {
    return messenger.getNumCurrentMessages();
  }
  
  /**
   * Sets the current status
   * 
   * @param currentTaskStatus the new task status to set
   */
  public final void setCurrentTaskStatus(TaskStatus currentTaskStatus) {
    this.currentTaskStatus = currentTaskStatus;
  }

  /**
   * update bsp task detialed phase to groomserver
   * @param phase
   * @throws InterruptedException 
   * @throws IOException 
   */
  public final void updateBspPhase(TaskStatus.Phase phase) throws IOException, InterruptedException{
         this.currentTaskStatus.setPhase(phase);
         this.umbilical.statusUpdate(taskId, currentTaskStatus);
  }
  
  public final TaskStatus.State getState(){
    return this.currentTaskStatus.getRunState();
  }
  
  public final void setState(TaskStatus.State state){
    this.currentTaskStatus.setRunState(state);
  }
  
  /**
   * @return the count of current super-step
   */
  @Override
  public final long getSuperstepCount() {
    return currentTaskStatus.getSuperstepCount();
  }

  /**
   * Gets the job configuration.
   * 
   * @return the conf
   */
  @Override
  public final Configuration getConfiguration() {
    return conf;
  }

  /*
   * IO STUFF
   */

  @Override
  public final void write(K2 key, V2 value) throws IOException {
    incrementCounter(PeerCounter.TASK_OUTPUT_RECORDS, 1);
    collector.collect(key, value);
  }

  @Override
  public final boolean readNext(K1 key, V1 value) throws IOException {
    return in.next(key, value);
  }

  @Override
  public final KeyValuePair<K1, V1> readNext() throws IOException {
    K1 k = in.createKey();
    V1 v = in.createValue();
    if (in.next(k, v)) {
      cachedPair.clear();
      cachedPair.setKey(k);
      cachedPair.setValue(v);
      return cachedPair;
    } else {
      return null;
    }
  }

  @Override
  public final void reopenInput() throws IOException {
    initInput();
  }

  @Override
  public final Counter getCounter(Enum<?> name) {
    return counters == null ? null : counters.findCounter(name);
  }

  @Override
  public final Counter getCounter(String group, String name) {
    Counters.Counter counter = null;
    if (counters != null) {
      counter = counters.findCounter(group, name);
    }
    return counter;
  }

  public Counters getCounters() {
    return counters;
  }

  @Override
  public final void incrementCounter(Enum<?> key, long amount) {
    if (counters != null) {
      counters.incrCounter(key, amount);
    }
  }

  @Override
  public final void incrementCounter(String group, String counter, long amount) {
    if (counters != null) {
      counters.incrCounter(group, counter, amount);
    }
  }

  @Override
  public TaskAttemptID getTaskId() {
    return taskId;
  }

}
