/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;

public class LyraRecoverMessage implements Writable {
  public static Class<?> VERTEX_CLASS;

  
  public static final byte RECOVER_MASTER_FLAG = 0x1;
  public static final byte RECOVER_REPLICA_FLAG = 0x2;
  public static final byte RECOVER_STAT_FLAG = 0x4;
  
  private byte flag;

  private AbstractVertex<?, ?, ?> vertex;
  private int targetPeer;
  private boolean isMirror;

  private MapWritable oldStat;

  private long iter;

  private long numVertices;
  
  public LyraRecoverMessage() {
    
  }

  public LyraRecoverMessage(AbstractVertex<?, ?, ?> vertex) {
    this.flag = RECOVER_MASTER_FLAG;
    this.vertex = vertex;
  }
  
  public LyraRecoverMessage(AbstractVertex<?, ?, ?> vertex, int targetPeer, boolean isMirror) {
    this.flag = RECOVER_REPLICA_FLAG;
    this.vertex = vertex;
    this.targetPeer = targetPeer;
    this.isMirror = isMirror;
  }

  public LyraRecoverMessage(MapWritable oldStat,long iter, long numberVertices){
    this.flag = RECOVER_STAT_FLAG;
    this.oldStat = oldStat;
    this.iter = iter;
    this.numVertices = numberVertices;
  }
  
  @Override
  public void write(DataOutput out) throws IOException {
    out.writeByte(this.flag);
    if (isRecoverMasterMessage()){
      vertex.writeRecoverMasterMsg(out);
    } else if (isRecoverReplicaMessage()){
      out.writeInt(targetPeer);
      out.writeBoolean(isMirror);
      vertex.writeRecoverReplicaMsg(out, targetPeer, isMirror);
    } else if (isRecoverStatMessage()){
      oldStat.write(out);
      out.writeLong(iter);
      out.writeLong(numVertices);
    }
  }


  @Override
  public void readFields(DataInput in) throws IOException {
    flag = in.readByte();
    if (isRecoverMasterMessage()){
      this.vertex = newVertexInstance(VERTEX_CLASS, null);
      vertex.readRecoverMasterMsg(in);
    } else if (isRecoverReplicaMessage()){
      this.targetPeer = in.readInt();
      this.isMirror = in.readBoolean();
      this.vertex = newVertexInstance(VERTEX_CLASS, null);
      vertex.readRecoverReplicaMsg(in, targetPeer, isMirror);
    } else if (isRecoverStatMessage()){
      this.oldStat = new MapWritable();
      this.oldStat.readFields(in);
      this.iter = in.readLong();
      this.numVertices = in.readLong();
    }
  }
  
  
  public boolean isRecoverMasterMessage() {
    return flag == RECOVER_MASTER_FLAG;
  }
  
  public boolean isRecoverReplicaMessage() {
    return flag == RECOVER_REPLICA_FLAG;
  }

  public boolean isRecoverStatMessage() {
    return flag == RECOVER_STAT_FLAG;
  }
  
  public AbstractVertex<?, ?, ?> getVertex() {
    return vertex;
  }
  
  
  public MapWritable getOldStat() {
    return oldStat;
  }

  public long getIter() {
    return iter;
  }

  public long getNumVertices() {
    return numVertices;
  }

  /**
   * @return a new vertex instance
   */
  public static <V extends Writable, E extends Writable, M extends Writable> AbstractVertex<V, E, M> newVertexInstance(
      Class<?> vertexClass, Configuration conf) {
    @SuppressWarnings("unchecked")
    AbstractVertex<V, E, M> vertex = (AbstractVertex<V, E, M>) ReflectionUtils.newInstance(
        vertexClass, conf);
    return vertex;
  }

}