/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;

public class LyraSyncMessage<V extends Writable, M extends Writable> implements Writable {
  public static Class<? extends Writable> VERTEX_ID_CLASS;
  public static Class<? extends Writable> VERTEX_VALUE_CLASS;
  
  private IntWritable localID;
  

  public boolean votedToHalt = false; // modified by vertex itself
  
  public boolean awaked = false; // modify by others

  public M value = null;
  
  public M msg = null;
  
  //activate state (out-edges)
  public boolean activated = false; // activate vertex or not?
  public boolean activateAllNeighbors = false; // if true, activatedVertices should be empty

  public List<V>[] activateLists = null;
  public List<V> activateList = null;
  
  public LyraSyncMessage() {
    
  }
  
  /**
   * create a SyncMessage to send to replica, here we will differentiate the mirrors and the ghost replicas
   * @param localID
   * @param state
   * @param replicaIndex if replicaIndex == -1, it means the receiver is a mirror (actually this is ugly, -_-|| ).
   *        if replicaIndex != -1, it means replicaIndex is will use an index to nActivateLists.
   */
  @SuppressWarnings("unchecked")
  public LyraSyncMessage(IntWritable localID, @SuppressWarnings("rawtypes") AbstractVertex state, int replicaIndex) {
    // here we need the next state of the vertex
    this.localID = localID;
    
    votedToHalt = state.nVotedToHalt;
    awaked = state.nAwaked;
    value = (M) state.nValue;
    msg = (M) state.nMsg;
    activated = state.nActivate;
    activateAllNeighbors = state.nActivateAllNeighbors;
    if (replicaIndex == -1){
      activateLists = (List<V>[])state.nActivateLists;
    } else {
      if (state.nActivateLists != null)
        activateList = state.nActivateLists[replicaIndex];
    }
  }
  
  @Override
  public void write(DataOutput out) throws IOException {
    localID.write(out);
    
    value.write(out);
    
    byte stateBitMap = 0;
    if (votedToHalt)
      stateBitMap |= 1;
    if (awaked)
      stateBitMap |= 2;
    if (activated)
      stateBitMap |= 4;
    if (msg != null)
      stateBitMap |= 16;

    if (activated) {
      if (activateAllNeighbors){
	stateBitMap |= 8;
      }else{
        if (activateLists != null)
          stateBitMap |= 32;
        if (activateList != null)
          stateBitMap |= 64;
      }
      
      out.writeByte(stateBitMap);
      
      if (activateAllNeighbors == false) {
        if (activateLists != null){
          out.writeInt(activateLists.length);
	  for (int i = 0; i < activateLists.length; i++) {
	    int size = activateLists[i].size();
	    out.writeInt(size);
	    for (Writable destinationVertexId : activateLists[i]) {
	      destinationVertexId.write(out);
	    }
	  }
        }
        if (activateList != null){
          out.writeInt(activateList.size());
          for (Writable destinationVertexId : activateList) {
            destinationVertexId.write(out);
          }
        }
      }
    } else {
      out.writeByte(stateBitMap);
    }
    
    if (msg != null) {
      msg.write(out);
    }
  }
  
  @SuppressWarnings("unchecked")
  @Override
  public void readFields(DataInput in) throws IOException {
    localID = new IntWritable();
    localID.readFields(in);
        
    value = (M) ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
    value.readFields(in);
    
    byte stateBitMap = in.readByte();
    if ((stateBitMap & 1) == 1)
      votedToHalt = true;
    else
      votedToHalt = false;
    
    if ((stateBitMap & 2) == 2)
      awaked = true;
    else
      awaked = false;
    
    if ((stateBitMap & 4) == 4)
      activated = true;
    else
      activated = false;

    if (activated) {
      if ((stateBitMap & 8) == 8)
	activateAllNeighbors = true;
      else
	activateAllNeighbors = false;
      
      if (activateAllNeighbors == false) {
        if ((stateBitMap & 32) == 32){     
          int length = in.readInt();
          activateLists = new List[length];
          for (int j = 0; j < length; j++) {
            int size = in.readInt();
            activateLists[j] = new ArrayList<V>(size);
            for (int i = 0; i < size; i++) {
              V activatedVertexId = (V) ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
              activatedVertexId.readFields(in);
              activateLists[j].add(activatedVertexId);
            }
          }
        }
        if ((stateBitMap & 64) == 64) {
          int size = in.readInt();
          activateList = new ArrayList<V>(size);
          for (int i = 0; i < size; i++) {
            V activatedVertexId = (V) ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
            activatedVertexId.readFields(in);
            activateList.add(activatedVertexId);
          }
        }
      }
    }
    
    if ((stateBitMap & 16) == 16){
      msg = (M) ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      msg.readFields(in);
    } else {
      msg = null;
    }
  }
  

  public IntWritable getLocalID() {
    return localID;
  }

  @SuppressWarnings("rawtypes")
  public AbstractVertexRuntimeState getState() {
    return new AbstractVertexRuntimeState<V, M>(this);
  }
  
  public String toString(){
    return "localID= " + localID + " " + "  votedToHalt=" + votedToHalt + "  awaked=" + awaked + " value=" + value + " msg=" + msg + " activated=" + activated + " activateAllNeighbors=" + activateAllNeighbors;
  }

}