/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp.message;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configurable;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.ipc.RPC;
import org.apache.hama.Constants;
import org.apache.hama.bsp.LyraRecoverMessage;
import org.apache.hama.bsp.LyraRecoverMessageBundle;
import org.apache.hama.ipc.HamaRPCProtocolVersion;
import org.apache.hama.util.LRUCache;

public class LyraRecoverSenderImpl implements Configurable, LyraRecoverSender{

  protected static final Log LOG = LogFactory
      .getLog(LyraRecoverSenderImpl.class);
  
  private Configuration conf;

  private final HashMap<String, InetSocketAddress> peerRecoverySocketCache = new HashMap<String, InetSocketAddress>();
  private LRUCache<InetSocketAddress, LyraMessageManager> peersLRUCache = null;
  
  private int maxCachedConnections = 100;
  
  private int portOffset = 0;
  private int myId = -1;
  private int handlers = -1;
  
  private int threshold = 1000;
  private ConcurrentMap<String, ArrayList<LyraRecoverMessage>> recoverSndMsgs = new ConcurrentHashMap<String, ArrayList<LyraRecoverMessage>>();
 
  
  @Override
  public void init(Configuration conf, int myId, int handler) {
    init(conf, 1, 0, 1);
    this.myId = myId;
    this.handlers = handler;
  }
  
  @SuppressWarnings("serial")
  private void init(Configuration conf, int numberThreads, int threadID, int numberServers) {
    this.conf = conf;
    this.maxCachedConnections = 100;
    // avoid port conflict
    this.portOffset = (threadID % numberServers) * conf.getInt(Constants.MAX_TASKS_PER_GROOM, 3);
    this.threshold = conf.getInt(Constants.RECOVER_MSG_THRESHOLD, 1000);
    
    peersLRUCache = new LRUCache<InetSocketAddress, LyraMessageManager>(
        maxCachedConnections) {
      @Override
      protected final boolean removeEldestEntry(
          Map.Entry<InetSocketAddress, LyraMessageManager> eldest) {
        if (size() > this.capacity) {
          LyraMessageManager proxy = eldest.getValue();
          RPC.stopProxy(proxy);
          return true;
        }
        return false;
      }
    };
  }
  
  private InetSocketAddress getLyraRecoveryAddress(String peerName, int myIdx, int handlers) {
    String[] peerAddrParts = peerName.split(":");
    if (peerAddrParts.length != 2) {
      throw new ArrayIndexOutOfBoundsException(
          "Peername must consist of exactly ONE \":\"! Given peername was: "
              + peerName);
    }
    int msgPort = Integer.valueOf(peerAddrParts[1]) 
        + Constants.DEFAULT_PEER_MSG_PORT_SPAN + portOffset + (myIdx%handlers) * (conf.getInt(Constants.MAX_TASKS_PER_GROOM, 3));
    return new InetSocketAddress(peerAddrParts[0], msgPort);
  }

  protected final LyraMessageManager getBSPPeerConnection(
      InetSocketAddress addr) throws IOException {
    LyraMessageManager bspPeerConnection;
    if (!peersLRUCache.containsKey(addr)) {
      bspPeerConnection = (LyraMessageManager) RPC.getProxy(
          LyraMessageManager.class, HamaRPCProtocolVersion.versionID, addr,
          this.conf);
      peersLRUCache.put(addr, bspPeerConnection);
    } else {
      bspPeerConnection = peersLRUCache.get(addr);
    }
    return bspPeerConnection;
  }
  
  @Override
  public void setConf(Configuration conf) {
    this.conf = conf;
  }

  @Override
  public Configuration getConf() {
    return conf;
  }
  
  @Override
  public void close() {
    peerRecoverySocketCache.clear();
  }
  
  
  /**
   *  FIXME: not thread safe, only one thread can call this function
   *  there is a buffer in this sender, so the caller can send a null message to clear the buffer.
   *  A better way is that user uses clearTransferBuffer to send the message in buffer
   */
  @SuppressWarnings({ "rawtypes", "unchecked" })
  @Override
  public void transfer(String peerName, LyraRecoverMessage message) throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target lyra msg-server.
    if (peerRecoverySocketCache.containsKey(peerName)) {
      addr = peerRecoverySocketCache.get(peerName);
    } else {
      addr = getLyraRecoveryAddress(peerName, myId, handlers);
      peerRecoverySocketCache.put(peerName, addr);
    }
    
    LyraMessageManager bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      ArrayList<LyraRecoverMessage> buffer = recoverSndMsgs.get(peerName);
      if (buffer == null){
        buffer = new ArrayList<LyraRecoverMessage>();
        recoverSndMsgs.put(peerName, buffer);
      }
      if (message != null){
        buffer.add(message);
      }
      if (buffer.size() > this.threshold || message == null){ //uglyCode, through send a special message (null) to clear the buffer
        if (buffer.size() > 0){
          long startTsf = System.currentTimeMillis();
          bspPeerConnection.doTransfer(new LyraRecoverMessageBundle(buffer), new IntWritable(myId));
          LOG.info("Time to transfer: " + (System.currentTimeMillis()-startTsf));
        }
        buffer.clear();
      }
    }
  }
  
  
  /**
   *  FIXME: not thread safe, only one thread can call this function
   */
  @SuppressWarnings({ "rawtypes", "unchecked" })
  @Override
  public void clearTransferBuffer(String peerName) throws IOException {
    InetSocketAddress addr = null;
    
    // Get socket for target lyra msg-server.
    if (peerRecoverySocketCache.containsKey(peerName)) {
      addr = peerRecoverySocketCache.get(peerName);
    } else {
      addr = getLyraRecoveryAddress(peerName, myId, handlers);
      peerRecoverySocketCache.put(peerName, addr);
    }
    
    LyraMessageManager bspPeerConnection = this.getBSPPeerConnection(addr);
    if (bspPeerConnection == null) {
      throw new IllegalArgumentException("Can not find " + addr.toString()
          + " to transfer messages to!");
    } else {
      ArrayList<LyraRecoverMessage> buffer = recoverSndMsgs.get(peerName);
      if (buffer == null || buffer.size() == 0){
        return;
      }
      if (buffer.size() > 0) {
        long startTsf = System.currentTimeMillis();
        bspPeerConnection.doTransfer(new LyraRecoverMessageBundle(buffer),
            new IntWritable(myId));
        LOG.info("Time to transfer: " + (System.currentTimeMillis() - startTsf));
        buffer.clear();
      }
      return;
    }
  }
  
  
  @Override
  public long getProtocolVersion(String protocol, long clientVersion) {
    return versionID;
  }
  
}
