/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp.sync;

import java.util.List;

import org.apache.hadoop.conf.Configuration;
import org.apache.hama.bsp.BSPJobID;
import org.apache.hama.bsp.TaskAttemptID;

/**
 * PeerSyncClient defines the behavior that a BSPPeer performs to maintain
 * synchronized global state as it progresses. 
 */

public interface PeerSyncClient extends SyncClient{

  /**
   * Init will be called within a spawned task, it should be used to initialize
   * the inner structure and fields, e.G. a zookeeper client or an rpc
   * connection to the real sync daemon.
   * 
   * @throws Exception
   */
  public void init(Configuration conf, BSPJobID jobId, TaskAttemptID taskId)
      throws Exception;


  /**
   * Enters the barrier before the commitState sending in each superstep.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param superstep the superstep of the task
   * @param state (0: normal state, 1: recover state)
   * @return if the element at index 0 is 0, the following elements are the newbies joining; if the element at index 0 is 1, the task are failed, no newbies will come to replace it
   * @throws SyncException
   */
  public List<Integer> lyraBarrier(BSPJobID jobId, TaskAttemptID taskId, long superstep, int state)
      throws SyncException;
  
  /**
   * awake others from barrier in the max superstep
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param superstep the superstep of the task, but actually it is useless
   * @param state (0: normal state, 1: recover state)
   * @return the superstep other sleep in
   * @throws SyncException
   */
  public long lyraAwakeOthersFromBarrier(BSPJobID jobId, TaskAttemptID taskId, long superstep, int state)
      throws SyncException;
  
  
  /**
   * Enters the barrier before the message sending in each superstep.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param superstep the superstep of the task
   * @throws SyncException
   */
  public void enterBarrier(BSPJobID jobId, TaskAttemptID taskId, long superstep)
      throws SyncException;

  /**
   * Leaves the barrier after all communication has been done, this is usually
   * the end of a superstep.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param superstep the superstep of the task
   * @throws SyncException
   */
  public void leaveBarrier(BSPJobID jobId, TaskAttemptID taskId, long superstep)
      throws SyncException;

  /**
   * Registers a specific task with a its host and port to the sync daemon.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param hostAddress the host where the sync server resides
   * @param port the port where the sync server is up
   */
  public void register(BSPJobID jobId, TaskAttemptID taskId,
      String hostAddress, long port);
  
  
  /**
   * Registers a specific task with a its host and port to the sync daemon.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param hostAddress the host where the sync server resides
   * @param port the port where the sync server is up
   */
  public void lyraRegister(BSPJobID jobId, TaskAttemptID taskId,
      String hostAddress, long port);
  
  /**
   * deregisters a specific task with a its host and port to the sync daemon.
   * 
   * @param jobId the jobs ID
   * @param taskId the tasks ID
   * @param hostAddress the host where the sync server resides
   * @param port the port where the sync server is up
   */
  public void lyraDeregister(BSPJobID jobId, TaskAttemptID taskId);
  
  /**
   * this interface will allow the caller store some persistent info on zookeeper
   * @param jobId
   * @param taskId
   * @param info
   * @return
   */
  public void lyraPersistentPut(BSPJobID jobId, int taskId, String info);
  
  
  /**
   * this interface will allow the caller get its persistent info on zookeeper
   * @param jobId
   * @param taskId
   * @return
   */
  public String lyraPersistentGet(BSPJobID jobId, int taskId);
  
  

  /**
   * Returns all registered tasks within the sync daemon. They have to be
   * ordered ascending by their task id.
   * 
   * @param taskId the tasks ID
   * @return an <b>ordered</b> string array of host:port pairs of all tasks
   *         connected to the daemon.
   */
  public String[] lyraGetAllPeerNames(TaskAttemptID taskId);
  
  
  /**
   * Returns all registered tasks within the sync daemon. They have to be
   * ordered ascending by their task id.
   * 
   * @param taskId the tasks ID
   * @return an <b>ordered</b> string array of host:port pairs of all tasks
   *         connected to the daemon.
   */
  public String[] getAllPeerNames(TaskAttemptID taskId);

  /**
   * TODO this has currently no use. Could later be used to deregister tasks
   * from the barrier during runtime if they are finished. Something equal to
   * voteToHalt() in Pregel.
   * 
   * @param jobId
   * @param taskId
   * @param hostAddress
   * @param port
   */
  public void deregisterFromBarrier(BSPJobID jobId, TaskAttemptID taskId,
      String hostAddress, long port);

  /**
   * This stops the sync daemon. Only used in YARN.
   */
  public void stopServer();

}
