/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp.sync;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.ipc.RPC;
import org.apache.hadoop.ipc.RPC.Server;
import org.apache.hama.Constants;
import org.apache.hama.bsp.TaskAttemptID;
import org.apache.hama.ipc.HamaRPCProtocolVersion;

/**
 * This client class abstracts the use of our RPC sync code.
 * 
 */
public class RPCSyncClient<M extends Writable> implements RPCSyncProtocol {
  
  public static final Log LOG = LogFactory
      .getLog(RPCSyncClient.class);
  
  private Configuration conf;
  private InetSocketAddress syncAddress;
  private Server server = null; 
  private boolean isMaster = false;
  private TaskAttemptID taskId;

  private List<RPCSyncProtocol> slaveConnections;

  private InetSocketAddress masterAddress;
  private RPCSyncProtocol masterConnection;
    
  private int barrierSense;
  private long barrierRelease;
  private AtomicInteger barrierCounter;
  private int numPeers;
  
  private final Lock barrierLock = new ReentrantLock();
  private final Condition barrierCond = barrierLock.newCondition();
  
  public void init(Configuration conf, final TaskAttemptID taskId, 
      String[] allPeers) throws Exception {
    this.conf = conf;
    
    this.taskId = taskId;
    this.isMaster = (this.taskId.getTaskID().getId() == 0);

    this.numPeers = allPeers.length;
    
    this.barrierSense = 1;
    this.barrierRelease = -1L;
    this.barrierCounter = new AtomicInteger(0);

    String syncHost = conf.get(Constants.PEER_HOST,
        Constants.DEFAULT_PEER_HOST);
    // NOTE: set PEER_SYNC_PORT in GroomServer.main()
    int syncPort = conf
        .getInt(Constants.PEER_SYNC_PORT, 
            Constants.DEFAULT_PEER_PORT + Constants.DEFAULT_PEER_SYNC_PORT_SPAN);

    syncAddress = new InetSocketAddress(syncHost, syncPort);
    LOG.info("Initialize lyra barrier at " + syncAddress);
    
    startRPCServer(conf, syncAddress);

    if (isMaster) {
      // skip master address (index = 0)
      slaveConnections = new ArrayList<RPCSyncProtocol>(numPeers-1);
      for (int idx = 1; idx < numPeers; idx++) { 
        InetSocketAddress address = getSyncAddress(allPeers[idx]);
        LOG.info("Peer#" + idx + " lyra barrier at " + address);
        slaveConnections.add((RPCSyncProtocol) RPC.getProxy(
            RPCSyncProtocol.class, HamaRPCProtocolVersion.versionID, 
            address, this.conf));
      }
    } else {
      masterAddress = getSyncAddress(allPeers[0]);
      masterConnection = (RPCSyncProtocol) RPC.getProxy(
          RPCSyncProtocol.class, HamaRPCProtocolVersion.versionID, 
          masterAddress, this.conf);
      LOG.info("Master lyra barrier at " + masterAddress);  
    }
  }
  
  private InetSocketAddress getSyncAddress(String peerName) {
    String[] peerAddrParts = peerName.split(":");
    if (peerAddrParts.length != 2) {
      throw new ArrayIndexOutOfBoundsException(
          "Peername must consist of exactly ONE \":\"! Given peername was: "
              + peerName);
    }
    return new InetSocketAddress(peerAddrParts[0], 
        Integer.valueOf(peerAddrParts[1]) + Constants.DEFAULT_PEER_SYNC_PORT_SPAN);
  }
  
  private void startRPCServer(Configuration conf,
      InetSocketAddress syncAddress) {
    try {
      this.server = RPC.getServer(this, syncAddress.getHostName(),
          syncAddress.getPort(), conf);
      server.start();
    } catch (IOException e) {
      LOG.error("Fail to start lyra sync-server!", e);
      throw new RuntimeException("lyra sync-erver could not be launched!");
    }
  }
  
  public String getSyncName() {
    return syncAddress.getHostName() + ":" + syncAddress.getPort();
  }
  
  public void close() throws IOException {
    if (server != null) {
      server.stop();
    }
  }
  
  private boolean isReady() {
    return ((barrierSense == -1 && barrierCounter.get() == 0) 
        || (barrierSense == 1 && barrierCounter.get() == this.numPeers));
  }
  
  private void checkAndSet(long release) {
    if (barrierRelease == -1)
      barrierRelease = release;
    else if (barrierRelease != release)
      LOG.error("wrong release: " + barrierRelease + " != " + release);
  }
  
  public void barrier(long superstep)
      throws SyncException {
    LOG.debug("[" + getSyncName() + "] start barrier: " + superstep);
    
    try {
      barrierLock.lock();
      checkAndSet(superstep);
      if (isMaster) {
        // phase 1: wait sync msgs from slaves      
        barrierCounter.addAndGet(barrierSense);
        if (!isReady())
          barrierCond.await();
        LOG.debug("master: counter = " + barrierCounter.get());
        
        // phase 2: send ack msgs to slaves
        LOG.debug("master: send ack msgs to slaves");
        for (RPCSyncProtocol connection : slaveConnections)
          connection.ack(new LongWritable(superstep));

        barrierSense = -barrierSense;
        barrierRelease = -1; // clear

        LOG.debug("master: barrier done.");
      } else {
        // phase 1: send sync msgs to master
        LOG.debug("slave: send sync msg to master");
        masterConnection.sync(new LongWritable(superstep));
        
        // phase 2: wait for ack msg from master
        barrierCond.await();
        
        barrierRelease = -1; // clear        
        LOG.debug("slave: barrier done.");
      }
    } catch (Exception e) {
      throw new SyncException(e.toString());
    } finally {
      barrierLock.unlock();
    }
  }

  
  @Override
  public void sync(LongWritable value) throws IOException {
    barrierLock.lock();
    checkAndSet(value.get());
    int done = barrierCounter.addAndGet(barrierSense);
    LOG.debug("master: sync " + done + " peers.");
    if (isReady())
      barrierCond.signal();
    barrierLock.unlock();
  }

  @Override
  public void ack(LongWritable value) throws IOException {
    barrierLock.lock();
    checkAndSet(value.get());
    LOG.debug("slave: ack " + barrierRelease);
    barrierCond.signal();
    barrierLock.unlock();
  }

  public long getProtocolVersion(String protocol, 
      long clientVersion) throws IOException {
    return versionID;
  }
}
