/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.bsp.AbstractVertex.VertexType;
import org.apache.hama.graph.RefEdge.RefEdgeType;

/**
 * A message that is either MapWritable (for meta communication purposes) or a
 * real message (vertex ID and value). It can be extended by adding flags, for
 * example for a graph repair call.
 */
public final class GraphJobMessage implements Writable {
  private static final Log LOG = LogFactory.getLog(GraphJobMessage.class);
  
  public static final int MAP_FLAG = 0x01;
  public static final int REPAIR_FLAG = 0x04;
  public static final int PARTITION_FLAG = 0x08;
  public static final int SIZE_FLAG = 0x10;

  public static final int REPLICA_FLAG = 0x20; // (BSP-g) build replica vertex in setup
  public static final int REGISTER_FLAG = 0x40; // (BSP-g) register replicas to master

  public static final int MIGRATE_FLAG = 0x80;
  public static final int MIGRATE_REPLICA_FLAG = 0x100;
  public static final int MIGRATE_MORE_EDGE_FLAG = 0x200;
  public static final int MIGRATE_REGISTER_FLAG = 0x400;
  
  // statically defined because it is process-wide information, therefore in caps
  // considered as a constant
  public static Class<?> VERTEX_CLASS;
  public static Class<? extends Writable> VERTEX_ID_CLASS;
  public static Class<? extends Writable> VERTEX_VALUE_CLASS;
  public static Class<? extends Writable> EDGE_VALUE_CLASS;

  // default flag to -1 "unknown"
  private int flag = -1;
  private MapWritable map;
  private Writable vertexId;
  private Writable vertexValue;
  private Vertex<?, ?, ?> vertex;
  private Map<Writable, Integer> vertexIdMap;
  private List<?> edges;
  private List<?> refEdges;
  private IntWritable numberVertices;
  private IntWritable numberReplicas;
  private IntWritable numberNewReplica;


  private boolean isMirror = false;
  private boolean selectiveActivate = false;
  

  private Writable destinationVertexId;
  private Writable edgeValue;
  
  private IntWritable peerIndex;
  private IntWritable localID;
  private IntWritable masterLID;
  
  
  public GraphJobMessage() {
  }

  public GraphJobMessage(MapWritable map) {
    this.flag = MAP_FLAG;
    this.map = map;
  }

  public GraphJobMessage(Writable vertexId) {
    this.flag = REPAIR_FLAG;
    this.vertexId = vertexId;
  }
  
  public GraphJobMessage(Vertex<?, ?, ?> vertex) {
    this.flag = PARTITION_FLAG;
    this.vertex = vertex;
  }

  public GraphJobMessage(int peerIndex, IntWritable numberVertices, IntWritable numberReplicas, IntWritable numberNewReplica) {
    this.flag = SIZE_FLAG;
    this.peerIndex = new IntWritable(peerIndex);
    this.numberVertices = numberVertices;
    this.numberReplicas = numberReplicas;
    this.numberNewReplica = numberNewReplica;
  }

  /**
   * build in-edges of destination vertex, and create replica if necessary
   * @param vertexId
   * @param vertexValue
   * @param edges
   * @param peerIndex
   */
  public GraphJobMessage(Writable vertexId, Writable vertexValue, List<?> edges,  IntWritable peerIndex) {
    this.flag = REPLICA_FLAG;
    this.vertexId = vertexId;
    this.vertexValue = vertexValue;
    this.edges = edges;
    this.peerIndex = peerIndex;
  }

  public GraphJobMessage(IntWritable peerIndex, 
      Writable vertexId, IntWritable localID, List<?> refEdges) {
    this.flag = REGISTER_FLAG;
    this.peerIndex = peerIndex;
    this.vertexId = vertexId;
    this.localID = localID;
    this.refEdges = refEdges;
  }
  

  public GraphJobMessage(Map<Writable, Integer> vertexIDs, int peerIndex) {
    this.flag = MIGRATE_FLAG;
    this.vertexIdMap = vertexIDs;
    this.peerIndex = new IntWritable(peerIndex);
  }
  
  @SuppressWarnings("rawtypes")
  public GraphJobMessage(Vertex<?, ?, ?> vertex, List<RefEdge> outEdgeList, int sourcePeer, int masterLID, boolean isSelectiveActivate) {
    this.flag = MIGRATE_REPLICA_FLAG;
    this.vertex = vertex;
    this.refEdges = outEdgeList;
    this.peerIndex = new IntWritable(sourcePeer);
    this.masterLID = new IntWritable(masterLID);
    this.selectiveActivate = isSelectiveActivate;
  }
  
  @SuppressWarnings("rawtypes")
  public GraphJobMessage(int RepLID, List<RefEdge> outEdgeList, boolean isSelectiveActivate) {
    this.flag = MIGRATE_MORE_EDGE_FLAG;
    this.localID = new IntWritable(RepLID);
    this.refEdges = outEdgeList;
    this.selectiveActivate = isSelectiveActivate;
  }
  
  public GraphJobMessage(int localID, Writable vertexID, int sourcePeer, int masterLID) {
    this.flag = MIGRATE_REGISTER_FLAG;
    this.localID = new IntWritable(localID);
    this.vertexId = vertexID;
    this.peerIndex = new IntWritable(sourcePeer);
    this.masterLID = new IntWritable(masterLID);
  }

  @Override
  public void write(DataOutput out) throws IOException {
    out.writeInt(this.flag);
    if (isMapMessage()) {
      map.write(out);
    } else if (isPartitioningMessage()) {
      vertex.getVertexID().write(out);
      if (vertex.getNewValue() != null) {
        out.writeBoolean(true);
        vertex.getNewValue().write(out);
      } else {
        out.writeBoolean(false);
      }
      List<?> outEdges = vertex.getEdges();
      out.writeInt(outEdges.size());
      for (Object e : outEdges) {
        Edge<?, ?> edge = (Edge<?, ?>) e;
        edge.getDestinationVertexID().write(out);
        if (edge.getValue() != null) {
          out.writeBoolean(true);
          edge.getValue().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    } else if (isMigrateMessage()) {
      out.writeInt(this.vertexIdMap.size());
      for (Entry<Writable, Integer> entry : this.vertexIdMap.entrySet()) {
        entry.getKey().write(out);
        out.writeInt(entry.getValue().intValue());
      }
      this.peerIndex.write(out);
    } else if (isMigrateReplicaMessage()) {
      out.writeBoolean(this.selectiveActivate);
      vertex.writeCommonStatesForRecover(out);
      if (this.selectiveActivate)
        writeRefEdgeListHelperSelectiveActivateVersion(out, refEdges);
      else
        writeRefEdgeListHalper(out, refEdges);
      this.peerIndex.write(out);
      this.masterLID.write(out);
    } else if (this.isMigrateMoreEdgeMessage()) {
      out.writeBoolean(this.selectiveActivate);
      this.localID.write(out);
      if (this.selectiveActivate)
        writeRefEdgeListHelperSelectiveActivateVersion(out, refEdges);
      else
        writeRefEdgeListHalper(out, refEdges);
    } else if (this.isMigrateRegisterMessage()) {
      this.localID.write(out);
      this.peerIndex.write(out);
      this.vertexId.write(out);
      this.masterLID.write(out);
    } else if (isSizeMessage()) {
      numberVertices.write(out);
      numberReplicas.write(out);
      numberNewReplica.write(out);
      peerIndex.write(out);
    } else if (isRepairMessage()) {
      vertexId.write(out);
    } else if (isReplicaMessage()) {
      vertexId.write(out);
      if (vertexValue == null) {
        vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      }
      vertexValue.write(out);
      peerIndex.write(out);
      
      if (edges == null)
	out.writeBoolean(false);
      else {
	out.writeBoolean(true);
	out.writeInt(edges.size());
	for (Object e : edges) {
	  Edge<?, ?> edge = (Edge<?, ?>) e;
	  edge.getDestinationVertexID().write(out);
	  if (edge.getValue() != null) {
	    out.writeBoolean(true);
	    edge.getValue().write(out);
	  } else {
	    out.writeBoolean(false);
	  }
	}
      }
    } else if (isRegisterMessage()) {
      peerIndex.write(out);
      vertexId.write(out);
      localID.write(out);
      
      writeRefEdgeListHalper(out, refEdges);
    } else {
      LOG.error("Unsupported message type (flag: " + this.flag + "");
      throw new UnsupportedOperationException("Unsupported message type (flag: " + this.flag + "");
    }
  }

  @Override
  public void readFields(DataInput in) throws IOException {
    flag = in.readInt();
    if (isMapMessage()) {
      map = new MapWritable();
      map.readFields(in);
    } else if (isPartitioningMessage()) {
      Vertex<Writable, Writable, Writable> vertex = GraphJobRunner
          .newVertexInstance(VERTEX_CLASS, null);
      Writable vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      vertex.setVertexID(vertexId);
      if (in.readBoolean()) {
        Writable vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS,
            null);
        vertexValue.readFields(in);
        vertex.setNewValue(vertexValue); // updated, then commit to lastValue
      }
      
      int size = in.readInt();
      vertex.setEdges(new ArrayList<Edge<Writable, Writable>>(size));
      for (int i = 0; i < size; i++) {
        Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS,
            null);
        edgeVertexID.readFields(in);
        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
          edgeValue.readFields(in);
        }
        vertex.getEdges().add(
            new Edge<Writable, Writable>(edgeVertexID, edgeValue));
      }
      this.vertex = vertex;
      
    } else if (isMigrateMessage()) {
      int size = in.readInt();
      this.vertexIdMap = new HashMap<Writable, Integer>();
      for (int i = 0; i < size; i++) {
        Writable vid = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
        vid.readFields(in);
        int localID = in.readInt();
        this.vertexIdMap.put(vid, new Integer(localID));
      }
      
      this.peerIndex = new IntWritable();
      this.peerIndex.readFields(in);
    } else if (isMigrateReplicaMessage()) {
      this.selectiveActivate = in.readBoolean();
      Vertex<Writable, Writable, Writable> vertex = GraphJobRunner.newVertexInstance(VERTEX_CLASS, null);
      Writable vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertex.setVertexID(vertexId);
      
      vertex.readCommonStatesForRecover(in);
      
      if (this.selectiveActivate)
        this.refEdges = readRefEdgeListHelperSelectiveActivateVersion(in, RefEdgeType.OUTEDGE);
      else
        this.refEdges = readRefEdgeListHelp(in, RefEdgeType.OUTEDGE);
      
      vertex.setVertexType(VertexType.REPLICA);
      this.vertex = vertex;
      
      this.peerIndex = new IntWritable();
      this.peerIndex.readFields(in);

      this.masterLID = new IntWritable();
      this.masterLID.readFields(in);
    } else if (this.isMigrateMoreEdgeMessage()) {
      this.selectiveActivate = in.readBoolean();
      this.localID = new IntWritable();
      this.localID.readFields(in);
      
      if (this.selectiveActivate)
        this.refEdges = readRefEdgeListHelperSelectiveActivateVersion(in, RefEdgeType.OUTEDGE);
      else
        this.refEdges = readRefEdgeListHelp(in, RefEdgeType.OUTEDGE);
    } else if (this.isMigrateRegisterMessage()) {
      this.localID = new IntWritable();
      this.localID.readFields(in);
      this.peerIndex = new IntWritable();
      this.peerIndex.readFields(in);
      
      this.vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      this.vertexId.readFields(in);

      this.masterLID = new IntWritable();
      this.masterLID.readFields(in);
    } else if (isSizeMessage()) {
      numberVertices = new IntWritable();
      numberVertices.readFields(in);
      numberReplicas = new IntWritable();
      numberReplicas.readFields(in);
      numberNewReplica = new IntWritable();
      numberNewReplica.readFields(in);
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
    } else if (isRepairMessage()) {
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
    } else if (isReplicaMessage()) {
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      vertexValue = ReflectionUtils.newInstance(VERTEX_VALUE_CLASS, null);
      vertexValue.readFields(in);
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
      
      if (in.readBoolean()) {
	int size = in.readInt();
	List<Edge<Writable, Writable>> tmpEdges = new ArrayList<Edge<Writable, Writable>>(size);
	for (int i = 0; i < size; i++) {
	  Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS,
							      null);
	  edgeVertexID.readFields(in);
	  Writable edgeValue = null;
	  if (in.readBoolean()) {
	    edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
	    edgeValue.readFields(in);
	  }
	  tmpEdges.add(new Edge<Writable, Writable>(edgeVertexID, edgeValue));
	}
      
	this.edges = tmpEdges;
      } else {
	this.edges = null;
      }
    } else if (isRegisterMessage()) {
      peerIndex = new IntWritable();
      peerIndex.readFields(in);
      vertexId = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
      vertexId.readFields(in);
      localID = new IntWritable();
      localID.readFields(in);
      
      refEdges = readRefEdgeListHelp(in, RefEdgeType.OUTEDGE);
    } else {
      LOG.error("Unsupported message type (flag: " + this.flag + "");
      throw new UnsupportedOperationException("Unsupported message type (flag: " + this.flag + "");
    }
  }
  
  @SuppressWarnings("rawtypes")
  private List<RefEdge> readRefEdgeListHelp(DataInput in, RefEdge.RefEdgeType expectType) throws IOException{
    if (in.readBoolean()){
      int size = in.readInt();
      List<RefEdge> result = new ArrayList<RefEdge>(size);
      for (int i = 0; i < size; i++){
        Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
        edgeVertexID.readFields(in);
        int lVID = in.readInt();
        int lVEID = in.readInt();

        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
          edgeValue.readFields(in);
        }
        @SuppressWarnings("unchecked")
        RefEdge tmp = new RefEdge(edgeVertexID, lVID, edgeValue, expectType);
        tmp.setvEID(lVEID);
        result.add(tmp);
      }
      return result;
    } else {
      return null;
    }
  }
  
  private void writeRefEdgeListHalper(DataOutput out, List<?> store) throws IOException{
    if (store == null){
      out.writeBoolean(false);
    } else {
      out.writeBoolean(true);
      out.writeInt(store.size());
      for (int i = 0; i < store.size(); i++){
        @SuppressWarnings("rawtypes")
        RefEdge edge = (RefEdge) store.get(i);
        edge.getVertexID().write(out);
        out.writeInt(edge.getlVID());
        out.writeInt(edge.getvEID());
        if (edge.getCost() != null) {
          out.writeBoolean(true);
          edge.getCost().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    }
  }
  
  @SuppressWarnings("rawtypes")
  private List<RefEdge> readRefEdgeListHelperSelectiveActivateVersion(DataInput in, RefEdge.RefEdgeType expectType) throws IOException {
    if (in.readBoolean()){
      int size = in.readInt();
      List<RefEdge> result = new ArrayList<RefEdge>(size);
      for (int i = 0; i < size; i++){
        Writable edgeVertexID = ReflectionUtils.newInstance(VERTEX_ID_CLASS, null);
        edgeVertexID.readFields(in);
        int lVID = in.readInt();
        int lVEID = in.readInt();
        boolean activated = in.readBoolean();

        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(EDGE_VALUE_CLASS, null);
          edgeValue.readFields(in);
        }
        @SuppressWarnings("unchecked")
        RefEdge tmp = new RefEdge(edgeVertexID, lVID, edgeValue, expectType);
        tmp.setvEID(lVEID);
        if (activated)
          tmp.activate();
        result.add(tmp);
      }
      return result;
    } else {
      return null;
    }
  }
  
  private void writeRefEdgeListHelperSelectiveActivateVersion(DataOutput out, List<?> store) throws IOException {
    if (store == null){
      out.writeBoolean(false);
    } else {
      out.writeBoolean(true);
      out.writeInt(store.size());
      for (int i = 0; i < store.size(); i++){
        @SuppressWarnings("rawtypes")
        RefEdge edge = (RefEdge) store.get(i);
        edge.getVertexID().write(out);
        out.writeInt(edge.getlVID());
        out.writeInt(edge.getvEID());
        out.writeBoolean(edge.isAwake());
        if (edge.getCost() != null) {
          out.writeBoolean(true);
          edge.getCost().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    }
  }
  

  

  public MapWritable getMap() {
    return map;
  }

  public Writable getVertexId() {
    return vertexId;
  }

  public Writable getVertexValue() {
    return vertexValue;
  }

  public Writable getEdgeValue() {
    return edgeValue;
  }
  
  public List<?> getEdges() {
    return this.edges;
  }
  
  
  public List<?> getRefEdges() {
    return refEdges;
  }

  public Vertex<?, ?, ?> getVertex() {
    return vertex;
  }
  
  public Map<Writable, Integer> getVertexIdMap() {
    return this.vertexIdMap;
  }

  public IntWritable getVerticesSize() {
    return numberVertices;
  }
  
  public IntWritable getReplicasSize() {
    return numberReplicas;
  }
  
  public IntWritable getNewReplica() {
    return numberNewReplica;
  }
  
  public Writable getDestinationVertexId() { // (BSP-g)
    return destinationVertexId;
  }
  
  public IntWritable getPeerIndex() { // (BSP-g)
    return peerIndex;
  }
  
  public IntWritable getLocalID() {
    return localID;
  }

  public IntWritable getMasterLID() {
    return masterLID;
  }


  public boolean isMirrorRep() {
    return this.isMirror;
  }

  public boolean isMapMessage() {
    return flag == MAP_FLAG;
  }

  public boolean isRepairMessage() {
    return flag == REPAIR_FLAG;
  }

  public boolean isPartitioningMessage() {
    return flag == PARTITION_FLAG;
  }
  
  public boolean isMigrateMessage() {
    return flag == MIGRATE_FLAG;
  }
  
  public boolean isMigrateReplicaMessage() {
    return flag == MIGRATE_REPLICA_FLAG;
  }
  
  public boolean isMigrateMoreEdgeMessage() {
    return flag == MIGRATE_MORE_EDGE_FLAG;
  }
  
  public boolean isMigrateRegisterMessage() {
    return flag == MIGRATE_REGISTER_FLAG;
  }

  public boolean isSizeMessage() {
    return flag == SIZE_FLAG;
  }
  
  public boolean isReplicaMessage() {
    return flag == REPLICA_FLAG;
  }
  
  public boolean isRegisterMessage() {
    return flag == REGISTER_FLAG;
  }
  
  @Override
  public String toString() {
    return "GraphJobMessage [flag=" + flag + ", map=" + map 
        + ", vertexId=" + vertexId + ", vertexValue=" + vertexValue 
        + ", vertex=" + vertex + ", #vertices=" + numberVertices + ", #replicas=" + numberReplicas
        + ", destinationVertexId=" + destinationVertexId
        + ", peerIndex=" + peerIndex
        + "]";
  }

}
