/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.BooleanWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.Constants;
import org.apache.hama.bsp.AbstractGlobalState;
import org.apache.hama.bsp.AbstractRefEdge;
import org.apache.hama.bsp.AbstractVertex.VertexType;
import org.apache.hama.bsp.BSP;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.LyraMessageBundle;
import org.apache.hama.bsp.LyraMirrorMessage;
import org.apache.hama.bsp.LyraRecoverMessage;
import org.apache.hama.bsp.LyraSyncBundle;
import org.apache.hama.bsp.LyraSyncMessage;
import org.apache.hama.bsp.Partitioner;
import org.apache.hama.bsp.TaskStatus;
import org.apache.hama.bsp.sync.SyncException;
import org.apache.hama.graph.RefEdge.RefEdgeType;
import org.apache.hama.util.KeyValuePair;

/**
 * Fully generic graph job runner.
 * 
 * @param <V> the id type of a vertex.
 * @param <E> the value type of an edge.
 * @param <M> the value type of a vertex.
 */
public final class GraphJobRunner<V extends Writable, E extends Writable, M extends Writable>
    extends BSP<Writable, Writable, Writable, Writable, GraphJobMessage> {

  public static enum GraphJobCounter {
    MULTISTEP_PARTITIONING, ITERATIONS, INPUT_VERTICES, AGGREGATE_VERTICES
  }

  private static final Log LOG = LogFactory.getLog(GraphJobRunner.class);

  // make sure that these values don't collide with the vertex names
  public static final String S_FLAG_STAT = "hama.stat";
  public static final String S_FLAG_HALT = "hama.stat.0";
  public static final String S_FLAG_ALIVE_COUNTS = "hama.stat.1";
  public static final String S_FLAG_LOCAL_MESSAGE_COUNTS = "hama.stat.2";
  public static final String S_FLAG_REMOTE_MESSAGE_COUNTS = "hama.stat.3";
  public static final String S_FLAG_LAST_ACTIVE_COUNTS = "hama.stat.4";
  public static final String S_FLAG_SYNC_MESSAGE_COUNTS = "hama.stat.5";

  public static Text FLAG_STAT_HALT;
  public static Text FLAG_STAT_ALIVE_COUNTS;
  public static Text FLAG_STAT_LOCAL_MESSAGE_COUNTS;
  public static Text FLAG_STAT_REMOTE_MESSAGE_COUNTS;
  public static Text FLAG_STAT_LAST_ACTIVE_COUNTS;
  public static Text FLAG_STAT_SYNC_MESSAGE_COUNTS;

  public static final String S_FLAG_AGGREGATOR_VALUE = "hama.1";
  public static final String S_FLAG_AGGREGATOR_INCREMENT = "hama.2";

  public static final String MESSAGE_COMBINER_CLASS = "hama.vertex.message.combiner.class";
  public static final String GRAPH_REPAIR = "hama.graph.repair";
  public static final String GRAPH_SELECTIVE_ACTIVATE = "hama.graph.selective.activate";

  private Configuration conf;
  private Partitioner<V, M> partitioner;
  private boolean isFennel = false;
  private int multiSteps = 1;

  private int localVertexNum = 0;
  
  private Vertex<V, E, M>[] vertices;
  private Map<V, Vertex<V, E, M>> verticesMap = new HashMap<V, Vertex<V, E, M>>();
  private Map<V, Vertex<V, E, M>> replicasMap = new HashMap<V, Vertex<V, E, M>>();

  private LyraSyncBundle[] bundles;
  private int bundleSize = 5000;
  // stat
  private long globalActiveVertexCount = 0;
  private long globalAliveVertexCount = 0;
  private long globalLocalMessageCount = 0;
  private long globalRemoteMessageCount = 0;
  private long globalSyncMessageCount = 0;
  private int globalReplicaAdd = 0;
  private AbstractGlobalState globalStat = new AbstractGlobalState();

  private long localActiveVertexCount = 0;
  private long localAliveVertexCount = 0;
  private long localLocalMessageCount = 0;
  private long localRemoteMessageCount = 0;
  private long localSyncMessageCount = 0;
  
  private int localMasterNum = 0;
  private int localReplicaNum = 0;
  private int localReplicaAdd = 0;
  private int localMirror = 0;
  private MapWritable localStat = new MapWritable();

  private long numberVertices = 0;
  private long numberReplicas = 0;
  //private long numberMirror = 0;
  // -1 is deactivated
  private int maxIteration = -1;
  private long iteration;
  private float minRate = -1f;
  private float rate;
  private volatile boolean converged = true;

  private Class<V> vertexIdClass;
  private Class<M> vertexValueClass;
  private Class<E> edgeValueClass;
  private Class<Vertex<V, E, M>> vertexClass;

  private AggregationRunner<V, E, M> aggregationRunner;

  private BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer;

  private boolean selectiveActivate = false;
  private boolean runtimePartitioning = false;
  
  private boolean danglingOPT = false;


  // execution time breakdown
  private long bspTOTACT = 0;
  private long setupTOT = 0, bspTOT = 0, onceTOT = 0;
  private long setupLD = 0, setupREP = 0, setupBUILD = 0, setupPICK = 0, setupREV = 0, setupINIT = 0, setupMIRROR = 0;
  private long bspSYN = 0, bspSND = 0, bspPRS = 0, bspMST = 0, bspSTP = 0,
      bspCommit = 0;
  private long onceSYN = 0, onceSYN2 = 0, onceSND = 0, onceMST = 0,
      onceSTP = 0, onceCommit = 0;
  private long avgSetupTOT = 0, avgSetupLD = 0, avgSetupREP = 0,
      avgSetupREV = 0, avgSetupINIT = 0, avgSetupPICK = 0, avgSetupMIRROR = 0, avgSetupBUILD = 0;
  private long avgBspTOT = 0, avgBspSYN = 0, avgBspSND = 0, avgBspPRS = 0,
      avgBspMST = 0, avgBspSTP = 0, avgBspCMT = 0, avgBspRepCMT = 0;
  private long sumMsgTOT = 0, msgRCV = 0;

  List<Integer> migrationTask = null; // use to skip send message to this machine 


  @SuppressWarnings("unchecked")
  @Override
  public final void setup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Setup graph!");

    setupTOT = System.currentTimeMillis();

    setupFields(peer);

    initLyraMessageService(peer);

    if (peer.getState() == TaskStatus.State.RUNNING) {
      // if normal start, build the graph topology and init some state

      // [SETUP-LD]: load hama graph
      if (!isFennel){
        loadVertices(peer);
        // [SETUP-REP]: build replicated graph
        setupREP = System.currentTimeMillis();
        loadReplicas(peer);
        setupREP = System.currentTimeMillis() - setupREP;
      } else {
        fennelLoadVertices(peer);
        // [SETUP-REP]: build replicated graph
        setupREP = System.currentTimeMillis();
        fennelLoadReplicas(peer);
        setupREP = System.currentTimeMillis() - setupREP;
      }
      setupPICK = System.currentTimeMillis();
      addAndPickMirrors(peer);
      this.localMasterNum = this.verticesMap.size();
      this.localReplicaNum = this.replicasMap.size();
      setupPICK = System.currentTimeMillis() - setupPICK;

      setupBUILD = System.currentTimeMillis();
      assignLocalID();
      buildTopology();
      peer.lyraRegisterInfo(""+vertices.length);
      
      initRunner(peer);

      setupBUILD = System.currentTimeMillis()-setupBUILD;
      
      setupREV = System.currentTimeMillis();
      registerReplicas(peer);
      setupREV = System.currentTimeMillis() - setupREV;
      
      ownerMap.clear();
      ownerMap = null;
      verticesMap.clear();
      verticesMap = null;
      replicasMap.clear();
      replicasMap = null;
      
      setupMIRROR = System.currentTimeMillis();
      replicateMirrorInfo(peer);
      setupMIRROR = System.currentTimeMillis() - setupMIRROR;

      
      countGlobal(peer);

      // [SETUP-INIT]: initialize all vertices
      setupINIT = System.currentTimeMillis();

      commit(peer);
      globalStat.commitState();// not need actually

      init(peer);

      peer.prepare();

      send(peer);



      peer.lyraSync(Constants.BARRIER_NORMAL); // 1st sync

      commit(peer);

      globalStat.commitState();

      System.gc();
      peer.lyraSync(Constants.BARRIER_NORMAL); // 2nd sync
      setupINIT = System.currentTimeMillis() - setupINIT;

      iteration++;
      setupTOT = System.currentTimeMillis() - setupTOT;
      setupLD = setupTOT - setupREP - setupREV - setupINIT - setupPICK - setupMIRROR - setupBUILD;
    } else {
      String verticesSize = peer.lyraGetInfo();
      LOG.info("The vertices number stored on zk:" + verticesSize);
      vertices = new Vertex[Integer.parseInt(verticesSize)];
    }

  }

  @Override
  public final void bsp(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    List<Integer> barrierStatus = null;
    if (peer.getState() == TaskStatus.State.RUNNING) {
      LOG.info("Starting graph processing!");
    } else {
      LOG.info("Starting recover graph topology and runtime state");
      LOG.info("awake others from barriers");
      peer.lyraSAwakeOtherFromSync(Constants.BARRIER_AWAKE);
      LOG.info("start to recieve message and recover");
      recoverFromPeer(peer);
      LOG.info("Recover finished. Start the computation from iteration "
          + this.iteration);
    }

    // until vertices convergence or max iteration
    bspTOTACT = System.currentTimeMillis();
    while (!isDone()) {
      reset(); // reset global stats
      // [BSP-ABT]: collect global stats and arbitrate next step
      onceMST = System.currentTimeMillis();
      arbitrate(peer);
      if (isConverged()) // if converged
        break;
      onceMST = System.currentTimeMillis() - onceMST;
      bspMST += onceMST;

      // [BSP-CMP]: loop over vertices and do commit and computation
      onceSTP = System.currentTimeMillis();
      compute(peer);
      peer.prepare();
      // [BSP-SND]: send sync and stat msgs
      onceSND = System.currentTimeMillis();
      send(peer);
      onceSND = System.currentTimeMillis() - onceSND;
      bspSND += onceSND;

      onceSTP = System.currentTimeMillis() - onceSTP - onceSND;
      bspSTP += onceSTP;

      
      // [BSP-SYN1]: sync with all peers
      onceSYN = System.currentTimeMillis();
      barrierStatus = peer.lyraSync(Constants.BARRIER_NORMAL); // 1st sync
      onceSYN = System.currentTimeMillis() - onceSYN;
      bspSYN += onceSYN;
      
      if (barrierStatus.size() != 0) {
        LOG.info("failure happened before the first barrier");
        int mode = barrierStatus.remove(0);
        if (mode == Constants.NEWBIE_HANDLER){
          recoverPeer(peer, barrierStatus);
        } else {
          migratePeer(peer, barrierStatus);
        }
        LOG.info("Recover over");
        clearDummyState();
        LOG.info("Clear the partial state");

        barrierStatus = null;
        continue; // mainly skip the next barrier
      } else {
        onceCommit = System.currentTimeMillis();
        commit(peer);
        globalStat.commitState();

        onceCommit = System.currentTimeMillis() - onceCommit;
        bspCommit += onceCommit;
      }

      // [BSP-SYN2]: sync with all workers
      onceSYN2 = System.currentTimeMillis();
      barrierStatus = peer.lyraSync(Constants.BARRIER_NORMAL); // 2nd sync
      onceSYN2 = System.currentTimeMillis() - onceSYN2;
      bspSYN += onceSYN2;

      this.iteration++;
      if (isMasterTask(peer)) {
        peer.getCounter(GraphJobCounter.ITERATIONS).increment(1);
      }
      if (barrierStatus.size() != 0) {
        LOG.info("failure happened after the first barrier");
        int mode = barrierStatus.remove(0);
        if (mode == Constants.NEWBIE_HANDLER){
          recoverPeer(peer, barrierStatus);
        } else {
          migratePeer(peer, barrierStatus);
        }
        LOG.debug("Recover over");
        barrierStatus = null;
      }
      onceTOT = onceMST + onceSTP + +onceSND + onceSYN + onceSYN2 + onceCommit;

      // [BSP-DMP]: dump superstep info
      if (isMasterTask(peer)) {
        LOG.info("==== iter#" + iteration + ":" + " [msg] #rcv-sync: "
            + globalSyncMessageCount + " #snd-local: "
            + globalLocalMessageCount + " #snd-remote: "
            + globalRemoteMessageCount + " [vtx] #active: "
            + globalActiveVertexCount + " #alive: " + globalAliveVertexCount
            + " [cvg] rate(%): " + (this.rate * 100f) + " ====");
      }

      LOG.info("==== iter#" + iteration + ":" + " [tot] " + onceTOT + " [mst] "
          + onceMST + " [cmp] " + onceSTP + " [snd] " + onceSND + " [syn] "
          + (onceSYN + onceSYN2) + " (" + onceSYN + " + " + onceSYN2 + ")"
          + " [cmt] " + onceCommit
          + " ==== \n");

    }
    bspTOTACT = System.currentTimeMillis() - bspTOTACT;
    bspTOT = bspSYN + bspPRS + bspMST + bspSTP + bspCommit + bspSND;


    if (isDone())
      LOG.info("BSP Job Done: max-iter (" + iteration + ") rate(%): "
          + (this.rate * 100f));
    else
      LOG.info("BSP Job Done: converged (" + iteration + ") rate(%): "
          + (this.rate * 100f));

    average(peer);
    breakdown();
    
  }

  /**
   * Just write <ID as Writable, Value as Writable> pair as a result. Note that
   * this will also be executed when failure happened.
   */
  @Override
  public final void cleanup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    for (Vertex<V, E, M> e : vertices) {
      if (e != null && e.isMaster()){
        peer.write(e.getVertexID(), e.getNewValue()); // (BSP-g)
      }
    }
  }
  


  private void initRunner(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    peer.registerMessaging(vertices, globalStat);
    bundles = new LyraSyncBundle[peer.getNumPeers()];
    for (int i = 0; i < bundles.length; i++)
      bundles[i] = new LyraSyncBundle();
    LOG.info("Init runner done.");
  }

  private void initLyraMessageService(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    try {
      if (peer.getState() == TaskStatus.State.RUNNING) {
        peer.initilizeLyraMessenger();
      } else {
        int handlers = this.conf.getInt(Constants.RECOVER_RPC_NUM, 4);
        peer.initilizeLyraRecoverMessenger(handlers);
      }
    } catch (ClassNotFoundException e) {
      Assert.fail("ClassNotFoundException during register replicase!");
    }
  }
  

  private final void breakdown() {
    LOG.info("Peer#" + peer.getPeerIndex());
    LOG.info("==== setup tot: " + setupTOT + " load: " + setupLD + " rep: "
        + setupREP + " rev: " + setupREV + " init: " + setupINIT + " pick: " + setupPICK + " mirror: " + setupMIRROR + " build: " + setupBUILD + " ====");
    LOG.info("====   bsp tot: " + bspTOT + " syn: " + bspSYN + " prs: "
        + bspPRS + " mst: " + bspMST + " stp: " + bspSTP + " snd: " + bspSND
        + " cmt: " + bspCommit + " ====");
    LOG.info("====   bsp tot: " + bspTOTACT + " actually ====");
    LOG.info("====   msg tot: " + msgRCV + " ====");

    if (isMasterTask(peer)) {
      LOG.info("Average");
      LOG.info("==== setup tot: " + avgSetupTOT + " load: " + avgSetupLD
          + " rep: " + avgSetupREP + " rev: " + avgSetupREV + " init: "
          + avgSetupINIT + " pick: " + avgSetupPICK + " mirror: " +avgSetupMIRROR + " build: " +avgSetupBUILD + " ====");
      LOG.info("====   bsp tot: " + avgBspTOT + " syn: " + avgBspSYN + " prs: "
          + avgBspPRS + " mst: " + avgBspMST + " stp: " + avgBspSTP + " snd: "
          + avgBspSND + " cmt: " + avgBspCMT + "(repComit:" + avgBspRepCMT
          + ")" + " ====");

      LOG.info("Sum");
      LOG.info("====   msg tot: " + sumMsgTOT + " ====");
    }
  }

  private final void average(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    int peerIndex = peer.getPeerIndex();

    peer.prepare();

    globalStat.commitState();

    peer.lyraSync(Constants.BARRIER_NORMAL);

    localStat.put(new Text("bsp-tot" + peerIndex), new LongWritable(bspTOT));
    localStat.put(new Text("bsp-syn" + peerIndex), new LongWritable(bspSYN));
    localStat.put(new Text("bsp-prs" + peerIndex), new LongWritable(bspPRS));
    localStat.put(new Text("bsp-mst" + peerIndex), new LongWritable(bspMST));
    localStat.put(new Text("bsp-stp" + peerIndex), new LongWritable(bspSTP));
    localStat.put(new Text("bsp-snd" + peerIndex), new LongWritable(bspSND));
    localStat.put(new Text("bsp-cmt" + peerIndex), new LongWritable(bspCommit));

    localStat.put(new Text("setup-tot" + peerIndex), new LongWritable(setupTOT));
    localStat.put(new Text("setup-ld" + peerIndex), new LongWritable(setupLD));
    localStat.put(new Text("setup-rep" + peerIndex), new LongWritable(setupREP));
    localStat.put(new Text("setup-rev" + peerIndex), new LongWritable(setupREV));
    localStat.put(new Text("setup-init" + peerIndex), new LongWritable(setupINIT));
    localStat.put(new Text("setup-pick" + peerIndex), new LongWritable(setupPICK));
    localStat.put(new Text("setup-mirror" + peerIndex), new LongWritable(setupMIRROR));
    localStat.put(new Text("setup-build" + peerIndex), new LongWritable(setupBUILD));

    peer.syncStat(getMasterTask(peer), localStat);
    localStat.clear();

    peer.lyraSync(Constants.BARRIER_NORMAL);

    globalStat.commitState();

    if (isMasterTask(peer)) {
      long btot = 0, bsyn = 0, bprs = 0, bmst = 0, bstp = 0, bsnd = 0, bcmt = 0, bRepCmt = 0;
      long stot = 0, sld = 0, srep = 0, srev = 0, sinit = 0, spick = 0, smirror = 0, sbuild = 0;
      int numberPeers = peer.getNumPeers();

      for (Entry<Writable, Writable> e : globalStat.getState().entrySet()) {
        String key = e.getKey().toString();

        if (key.startsWith("bsp-tot")) {
          btot += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-syn")) {
          bsyn += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-prs")) {
          bprs += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-mst")) {
          bmst += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-stp")) {
          bstp += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-snd")) {
          bsnd += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-cmt")) {
          bcmt += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-tot")) {
          stot += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-ld")) {
          sld += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-rep")) {
          srep += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-rev")) {
          srev += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-init")) {
          sinit += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-pick")) {
          spick += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-mirror")) {
          smirror += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-build")) {
          sbuild += ((LongWritable) e.getValue()).get();
        } else {
          throw new UnsupportedOperationException("unknown time: " + key);
        }
      }

      avgBspTOT = btot / numberPeers;
      avgBspSYN = bsyn / numberPeers;
      avgBspPRS = bprs / numberPeers;
      avgBspMST = bmst / numberPeers;
      avgBspSTP = bstp / numberPeers;
      avgBspSND = bsnd / numberPeers;
      avgBspCMT = bcmt / numberPeers;
      avgBspRepCMT = bRepCmt / numberPeers;

      avgSetupTOT = stot / numberPeers;
      avgSetupLD = sld / numberPeers;
      avgSetupREP = srep / numberPeers;
      avgSetupREV = srev / numberPeers;
      avgSetupINIT = sinit / numberPeers;
      avgSetupPICK = spick / numberPeers;
      avgSetupMIRROR = smirror / numberPeers;
      avgSetupBUILD = sbuild / numberPeers;
    }
  }


  private final boolean isDone() {
    return (this.maxIteration > 0) && (this.iteration > this.maxIteration);
  }

  private final boolean isConverged() {
    return (this.converged || (this.minRate > 0.0f)
        && (this.rate < this.minRate));
  }

  private final void reset() {
    // recieve
    globalActiveVertexCount = 0;
    globalAliveVertexCount = 0;
    globalLocalMessageCount = 0;
    globalRemoteMessageCount = 0;
    globalSyncMessageCount = 0;
    this.converged = true;

    // send
    localActiveVertexCount = 0;
    localAliveVertexCount = 0;
    localLocalMessageCount = 0;
    localRemoteMessageCount = 0;
    localSyncMessageCount = 0;
  }

  private void arbitrate(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    
    localSyncMessageCount = peer.getNumCurrentLyraMessages();
    localSyncMessageCount -= peer.getNumPeers(); // minus the global-stat msgs
                                                 // from each peer

    for (Entry<Writable, Writable> e : globalStat.getState().entrySet()) {
      String key = e.getKey().toString();
      if (key.startsWith(S_FLAG_HALT)) {
        if (((BooleanWritable) e.getValue()).get() == false)
          this.converged = false;
      } else if (key.startsWith(S_FLAG_ALIVE_COUNTS)) {
        globalAliveVertexCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_LOCAL_MESSAGE_COUNTS)) {
        globalLocalMessageCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_REMOTE_MESSAGE_COUNTS)) {
        globalRemoteMessageCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_LAST_ACTIVE_COUNTS)) {
        globalActiveVertexCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_SYNC_MESSAGE_COUNTS)) {
        globalSyncMessageCount += ((LongWritable) e.getValue()).get();
      } else {
        throw new UnsupportedOperationException("unknown stat: " + key);
      }
    }

    if (aggregationRunner.isEnabled())
      // update aggregation value if necessary (additional sync)
      throw new UnsupportedOperationException("aggregation unsupported now!");

    this.rate = ((float) this.globalAliveVertexCount / (float) this.numberVertices);
    // this.rate = ((float)this.globalActiveVertexCount /
    // (float)this.numberVertices);

    sumMsgTOT += globalSyncMessageCount;
    msgRCV += localSyncMessageCount;

    LOG.debug("arbitrate #sync-msgs: " + localSyncMessageCount + " rate(%): "
        + (this.rate * 100f));
  }

  private void commit(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    for (int i = 0; i < this.localVertexNum; i++){
      Vertex<V, E, M> vertex = vertices[i];
      vertex.commitState();
    }
  }
  /**
   * Do the main logic of a superstep, namely checking if vertices are active,
   * feeding compute with messages and controlling combiners/aggregators.
   */
  private void compute(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    long activeVertexCount = 0;
    long aliveVertexCount = 0;
    long localMessageCount = 0;
    long remoteMessageCount = 0;

    if (this.migrationTask == null){ // no migration, masters are also at the head part of the vertices, we can only traverse masters
      for (int i = 0; i < this.localMasterNum; i++) {
        Vertex<V, E, M> vertex = vertices[i];
        if (vertex.isActive()) {
          activeVertexCount++;
          // TODO: modify compute function interface?
          vertex.compute(null);
          localMessageCount += vertex.localActivateVertices();
          long count = vertex.remoteSync(bundles);
          remoteMessageCount += count;
          if (!vertex.isConverged())
            aliveVertexCount++;
        }
      }
    } else { // after migration happens, the new master are distributed and not together at the head of vertices array, we need traverse all the vertex
      for (int i = 0; i < this.localVertexNum; i++) {
        Vertex<V, E, M> vertex = vertices[i];
        if (vertex != null && vertex.getVertexType() == VertexType.MASTER && vertex.isActive()) {
          activeVertexCount++;
          // TODO: modify compute function interface?
          vertex.compute(null);
          localMessageCount += vertex.localActivateVertices();
          long count = vertex.remoteSync(bundles);
          remoteMessageCount += count;
          if (!vertex.isConverged())
            aliveVertexCount++;
        }
      }
    }
    LOG.debug("ALIVE NUM: "+activeVertexCount);

    // [unused] send aggregation value if necessary (need additional sync)
    // aggregationRunner.sendAggregatorValues(peer);

    localActiveVertexCount = activeVertexCount;
    localAliveVertexCount = aliveVertexCount;
    localLocalMessageCount = localMessageCount;
    localRemoteMessageCount = remoteMessageCount;
  }

  /**
   * Seed the vertices first with their own values in compute. This is the first
   * superstep after the vertices have been loaded.
   */
  private void init(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException {
    long activeVertexCount = 0;
    long aliveVertexCount = 0;
    long localMessageCount = 0;
    long remoteMessageCount = 0;

    for (Vertex<V, E, M> vertex : vertices) {
      if (vertex != null && vertex.getVertexType() == VertexType.MASTER){
        activeVertexCount++;

        // M lastValue = vertex.getLastValue(); // (BSP-g)
        // TODO: modify compute function interface?
        vertex.compute(null);
        localMessageCount += vertex.localActivateVertices();
        remoteMessageCount += vertex.remoteSync(bundles);

        // aggregationRunner.aggregateVertex(lastValue, vertex);

        if (!vertex.isConverged())
          aliveVertexCount++;
      }
    }

    // [unused] send aggregation value if necessary (need additional sync)
    // aggregationRunner.sendAggregatorValues(peer);

    localActiveVertexCount = activeVertexCount;
    localAliveVertexCount = aliveVertexCount;
    localLocalMessageCount = localMessageCount;
    localRemoteMessageCount = remoteMessageCount;
  }

  private void send(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    boolean noNodeFailure = true;
    int i = peer.getPeerIndex();
    int j = 0;
    for (j = 0; j < bundles.length; j++) {
      if (!bundles[i].isEmpty()) {
        int n = bundles[i].size() / bundleSize;
        if (n > 1) {
          List<LyraSyncBundle> list = bundles[i].split(n);
          for (LyraSyncBundle bundle : list){
            noNodeFailure = peer.syncVertex(peer.getPeerName(i), bundle);
            if (!noNodeFailure)
              break;
          }
        } else {
          noNodeFailure = peer.syncVertex(peer.getPeerName(i), bundles[i]);
        }

        bundles[i] = new LyraSyncBundle();
        if (!noNodeFailure)
          break;
      }
      i = (i+1)%peer.getNumPeers();
    }
    if (!noNodeFailure){
      // clear state
      for (i = 0; i < bundles.length; i++)
        bundles[i] = new LyraSyncBundle();
      localStat.clear();
      return;
    }
      

    // send local-stat to all peers
    if ((localAliveVertexCount > 0) || (localLocalMessageCount > 0)
        || (localRemoteMessageCount > 0))
      localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(false));
    else
      localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(true));
    localStat.put(GraphJobRunner.FLAG_STAT_LAST_ACTIVE_COUNTS,
        new LongWritable(localActiveVertexCount));
    localStat.put(GraphJobRunner.FLAG_STAT_ALIVE_COUNTS, new LongWritable(
        localAliveVertexCount));
    localStat.put(GraphJobRunner.FLAG_STAT_LOCAL_MESSAGE_COUNTS,
        new LongWritable(localLocalMessageCount));
    localStat.put(GraphJobRunner.FLAG_STAT_REMOTE_MESSAGE_COUNTS,
        new LongWritable(localRemoteMessageCount));
    localStat.put(GraphJobRunner.FLAG_STAT_SYNC_MESSAGE_COUNTS,
        new LongWritable(localSyncMessageCount));

    i = peer.getPeerIndex();
    j = 0;
    for (j = 0; j < bundles.length; j++) {
      if (!(this.migrationTask != null && migrationTask.contains(i)))
        peer.syncStat(peer.getPeerName(i),localStat);
      i = (i+1)%peer.getNumPeers();
    }
    localStat.clear();

    LOG.debug("worker:" + " [msg] #local = " + localLocalMessageCount
        + " #remote = " + localRemoteMessageCount + " [vtx] #active: "
        + localActiveVertexCount + " #alive: " + localAliveVertexCount
        + " [rep] #synced: " + localSyncMessageCount);
  }

  @SuppressWarnings("unchecked")
  private void setupFields(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    this.peer = peer;
    this.conf = peer.getConfiguration();

    maxIteration = peer.getConfiguration().getInt("hama.graph.max.iteration",
        -1);
    minRate = peer.getConfiguration().getFloat("hama.graph.min.rate", -1f);
   
    vertexIdClass = (Class<V>) conf.getClass(GraphJob.VERTEX_ID_CLASS_ATTR,
        Text.class, Writable.class);
    vertexValueClass = (Class<M>) conf.getClass(
        GraphJob.VERTEX_VALUE_CLASS_ATTR, IntWritable.class, Writable.class);
    edgeValueClass = (Class<E>) conf.getClass(
        GraphJob.VERTEX_EDGE_VALUE_CLASS_ATTR, IntWritable.class,
        Writable.class);
    vertexClass = (Class<Vertex<V, E, M>>) conf.getClass(
        "hama.graph.vertex.class", Vertex.class);

    // set the classes statically, so we can save memory per message
    GraphJobMessage.VERTEX_ID_CLASS = vertexIdClass;
    GraphJobMessage.VERTEX_VALUE_CLASS = vertexValueClass;
    GraphJobMessage.VERTEX_CLASS = vertexClass;
    GraphJobMessage.EDGE_VALUE_CLASS = edgeValueClass;

    LyraSyncMessage.VERTEX_ID_CLASS = vertexIdClass;
    LyraSyncMessage.VERTEX_VALUE_CLASS = vertexValueClass;
    
    LyraRecoverMessage.VERTEX_CLASS = vertexClass;
    // ugly
    Vertex.V_CLASS = vertexIdClass;
    Vertex.E_CLASS = edgeValueClass;
    Vertex.M_CLASS = vertexValueClass;

    partitioner = (Partitioner<V, M>) ReflectionUtils.newInstance(
        conf.getClass("bsp.input.partitioner.class", HashPartitioner.class),
        conf);
    partitioner.setConf(conf);
    isFennel = conf.getBoolean(Constants.GRAPH_FENNEL, false);
    if (isFennel){
      Assert.assertEquals(true, partitioner.getClass().equals(HashPartitioner.class));
      if (conf.get("hama.graph.multi.step.partitioning.interval") == null )
        conf.set("hama.graph.multi.step.partitioning.interval", "50000"); // for fennel, we need a small interval
    }

    aggregationRunner = new AggregationRunner<V, E, M>();
    aggregationRunner.setupAggregators(peer);

    // unify stat name
    FLAG_STAT_HALT = new Text(S_FLAG_HALT + "-"
        + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_ALIVE_COUNTS = new Text(S_FLAG_ALIVE_COUNTS + "-"
        + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_LOCAL_MESSAGE_COUNTS = new Text(S_FLAG_LOCAL_MESSAGE_COUNTS + "-"
        + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_REMOTE_MESSAGE_COUNTS = new Text(S_FLAG_REMOTE_MESSAGE_COUNTS
        + "-" + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_LAST_ACTIVE_COUNTS = new Text(S_FLAG_LAST_ACTIVE_COUNTS + "-"
        + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_SYNC_MESSAGE_COUNTS = new Text(S_FLAG_SYNC_MESSAGE_COUNTS + "-"
        + Integer.toString(peer.getPeerIndex()));

    // TODO: Now, we only provide partitioning code
    runtimePartitioning = conf.getBoolean(
        GraphJob.VERTEX_GRAPH_RUNTIME_PARTIONING, true);
    Assert.assertTrue(runtimePartitioning);

    selectiveActivate = conf.getBoolean(GRAPH_SELECTIVE_ACTIVATE, false);
    if (selectiveActivate)
      LOG.info("Selective activation is enabled");
    danglingOPT = conf.getBoolean(Constants.GRAPH_DANGLING_OPT, false);
    if (danglingOPT)
      LOG.info("dangling_opt is enabled");
  }

  /**
   * Partitions our vertices through multiple supersteps to save memory.
   */
  private int partitionMultiSteps(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      long splitSize) throws IOException, SyncException, InterruptedException {
    int multiSteps = 1;

    MapWritable ssize = new MapWritable();
    ssize
        .put(new IntWritable(peer.getPeerIndex()), new LongWritable(splitSize));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      long maxSplitSize = 0L;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          long curr = ((LongWritable) e.getValue()).get();
          if (maxSplitSize < curr) {
            maxSplitSize = curr;
          }
        }
      }

      int steps = (int) (maxSplitSize / conf.getLong( // 20 mb
          "hama.graph.multi.step.partitioning.interval", 20000000)) + 1;

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("max"), new IntWritable(steps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    if (isMasterTask(peer)) {
      peer.getCounter(GraphJobCounter.MULTISTEP_PARTITIONING).increment(
          multiSteps);
    }

    return multiSteps;
  }

  @SuppressWarnings("unchecked")
  private void createVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    GraphJobMessage msg = null;
    int peerIdx = peer.getPeerIndex();
    while ((msg = peer.getCurrentMessage()) != null) {
      Vertex<V, E, M> vertex = (Vertex<V, E, M>) msg.getVertex();
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);
      vertex.setMasterPeerIndex(peerIdx);

      Vertex<V, E, M> old = verticesMap.put(vertex.getVertexID(), vertex);
      if (old != null) // vertex with partitioned info
        vertex.merge(old);

      vertex.setup(conf); // call setup after merge

      if (selectiveActivate)
        vertex.enableSelectiveActivate();
      if (danglingOPT)
        vertex.enableDanglingOpt();
    }
  }
  
  
  @SuppressWarnings("unchecked")
  private void createVerticesAndAddOwnerMap(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer, MapWritable owerMap, int[] verticesOnNode)
      throws IOException {
    GraphJobMessage msg = null;
    int peerIdx = peer.getPeerIndex();
    while ((msg = peer.getCurrentMessage()) != null) {
      if (msg.isMapMessage()){
        for (Entry<Writable, Writable> entry : msg.getMap().entrySet()){
          int owner = ((IntWritable)(entry.getValue())).get();
          verticesOnNode[owner]++;
        }
        owerMap.putAll(msg.getMap());
        continue;
      }
      Vertex<V, E, M> vertex = (Vertex<V, E, M>) msg.getVertex();
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);
      vertex.setMasterPeerIndex(peerIdx);

      Vertex<V, E, M> old = verticesMap.put(vertex.getVertexID(), vertex);
      if (old != null) // vertex with partitioned info
        vertex.merge(old);

      vertex.setup(conf); // call setup after merge

      if (selectiveActivate)
        vertex.enableSelectiveActivate();
      if (danglingOPT)
        vertex.enableDanglingOpt();
    }
  }
  
  
  
  static double alpha = 0;
  static double gamma = 1.5;
  static double MAXCAP = 0;
  MapWritable ownerMap = new MapWritable();
  
  private void assignAlpha(int nodeNum){
    int VERTICESNUM = conf.getInt(Constants.GRAPH_VERTEX_NUM, 1000);
    int EDGENUM = conf.getInt(Constants.GRAPH_EDGE_NUM, 5000);
    alpha = (Math.pow(nodeNum, 0.5) * EDGENUM)/(Math.pow(VERTICESNUM, 1.5));
    MAXCAP = 1.1 * (VERTICESNUM  / nodeNum);
    LOG.debug("vertexNum:" + VERTICESNUM + " EDGE:" + EDGENUM + " alpha:" + alpha + " maxCap:" + MAXCAP);
  }
  
  private static double computeRank(int neighbor, int vOnNode){
    if (vOnNode < MAXCAP){
      return neighbor - alpha*gamma*Math.pow(vOnNode, gamma-1);
    } else {
      return -1000000000000000000D;
    }
  }
  
  // SideEffect: verticesOnNode[owner]++, and put pair(vid -> PositionRecord ) into newRecord
  private int getOwner(Vertex<V, E, M> me,  int[] verticesOnNode, MapWritable ownerMap, MapWritable newRecord){
    int owner = 0;
    double top = -1000000000000000000D;
    for (int i = 0; i < verticesOnNode.length; i++){
      int neighbor = 0;
      for (Edge<V, E> out : me.getEdges()){
        IntWritable pr = (IntWritable) ownerMap.get(out.getDestinationVertexID());
        if (pr == null){
          pr = (IntWritable) newRecord.get(out.getDestinationVertexID());
        }
        if (pr!= null && pr.get()== i){
          neighbor++;
        }
      }
      double rank = computeRank(neighbor,verticesOnNode[i]);
      if (rank > top){
        top = rank;
        owner = i;
      }
    }
    verticesOnNode[owner]++;
    newRecord.put(me.getVertexID(), new IntWritable(owner));
    return owner;
  }
  
  

  /**
   * Loads vertices into memory of each peer.
   * TODO this needs to be simplified.
   */
  @SuppressWarnings("unchecked")
  private void loadVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start loading vertices");

    final VertexInputReader<Writable, Writable, V, E, M> reader = (VertexInputReader<Writable, Writable, V, E, M>) ReflectionUtils
        .newInstance(conf.getClass(GraphJob.VERTEX_GRAPH_INPUT_READER,
            VertexInputReader.class), conf);

    final boolean repairNeeded = conf.getBoolean(GRAPH_REPAIR, false);
    final long splitSize = peer.getSplitSize();
    final int partitioningSteps = partitionMultiSteps(peer, splitSize);
    multiSteps = partitioningSteps; // used by load replicas
    final long interval = splitSize / partitioningSteps;
    final boolean selfReference = conf.getBoolean("hama.graph.self.ref", false);

    // new vertex container
    Vertex<V, E, M> vertex = newVertexInstance(vertexClass, conf);
    vertex.runner = this;
    vertex.setVertexType(VertexType.MASTER);

    long startPos = peer.getPos();
    if (startPos == 0)
      startPos = 1L;

    KeyValuePair<Writable, Writable> next = null;
    int numberPeers = peer.getNumPeers();
    boolean pending = false; // fix the tail-vertex issue
    int steps = 1;

    while ((next = peer.readNext()) != null) {
      boolean finished = false;
      try {
        finished = reader.parseVertex(next.getKey(), next.getValue(), vertex);
        pending = true;
      } catch (Exception e) {
        e.printStackTrace();
        throw new IOException("exception occured during parsing vertex!"
            + e.toString());
      }

      if (!finished)
        continue;

      if (vertex.getEdges() == null) {
        if (selfReference) {
          vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
              .getVertexID(), null)));
        } else {
          vertex.setEdges(Collections.EMPTY_LIST);
        }
      }

      if (selfReference)
        vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));

      if (runtimePartitioning) {
        int partition = partitioner.getPartition(vertex.getVertexID(),
            vertex.getNewValue(), numberPeers);
        peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
      } else {
        vertex.setup(conf);
        if (selectiveActivate)
          vertex.enableSelectiveActivate();
        if (danglingOPT)
          vertex.enableDanglingOpt();
        vertex.setMasterPeerIndex(peer.getPeerIndex());
        verticesMap.put(vertex.getVertexID(), vertex);
      }

      // new vertex container
      vertex = newVertexInstance(vertexClass, conf);
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);

      if (runtimePartitioning) {
        if (steps < partitioningSteps && (peer.getPos() - startPos) >= interval) {
          peer.sync();
          steps++;
          createVertices(peer);
          startPos = peer.getPos();
        }
      }
    }

    // flush the tail-vertex
    if (pending) {
      boolean tail = false;
      try {
        tail = reader.parseVertex(null, null, vertex);
      } catch (Exception e) {
        throw new IOException("exception occured during parsing tail vertex!"
            + e.toString());
      }

      if (tail) {
        if (vertex.getEdges() == null) {
          if (selfReference) {
            vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
                .getVertexID(), null)));
          } else {
            vertex.setEdges(Collections.EMPTY_LIST);
          }
        }

        if (selfReference)
          vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));

        if (runtimePartitioning) {
          int partition = partitioner.getPartition(vertex.getVertexID(),
              vertex.getNewValue(), peer.getNumPeers());
          peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
        } else {
          vertex.setup(conf);
          vertex.setMasterPeerIndex(peer.getPeerIndex());
          if (selectiveActivate)
            vertex.enableSelectiveActivate();
          if (danglingOPT)
            vertex.enableDanglingOpt();
          verticesMap.put(vertex.getVertexID(), vertex);
        }

        LOG.debug("Oops! Fix the tail-vertex(" + vertex.getVertexID()
            + ") issue.");
      }
    }

    // last step
    if (runtimePartitioning) {
      while(steps <= partitioningSteps){
        peer.sync();
        createVertices(peer);
        steps++;
      }
    }

    LOG.info("Load vertices finished at " + peer.getSuperstepCount()
        + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());

    
    /*
     * If the user want to repair the graph, it should traverse through that
     * local chunk of adjancency list and message the corresponding peer to
     * check whether that vertex exists. In real-life this may be dead-ending
     * vertices, since we have no information about outgoing edges. Mainly this
     * procedure is to prevent NullPointerExceptions from happening.
     */
    if (repairNeeded)
      repair(peer, partitioningSteps, selfReference);
  }
  
  
  /**
   * Loads vertices into memory of each peer. TODO this needs to be simplified.
   * @param buffer 
   */
  @SuppressWarnings("unchecked")
  private void fennelLoadVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    // Initialise fennel partition data structure
    int numPeer = peer.getNumPeers();
    int[] verticesOnNode = new int[numPeer];
    for (int i = 0; i < numPeer; i++)
      verticesOnNode[i] = 0;    
    assignAlpha(numPeer);
    MapWritable newRecord = new MapWritable();
    
    
    LOG.info("Start fennel loading vertices");

    final VertexInputReader<Writable, Writable, V, E, M> reader = (VertexInputReader<Writable, Writable, V, E, M>) ReflectionUtils
        .newInstance(conf.getClass(GraphJob.VERTEX_GRAPH_INPUT_READER,
            VertexInputReader.class), conf);

    final boolean repairNeeded = conf.getBoolean(GRAPH_REPAIR, false);
    final long splitSize = peer.getSplitSize();
    final int partitioningSteps = partitionMultiSteps(peer, splitSize);
    multiSteps = (partitioningSteps/20)+1; // used by load replicas
    final long interval = splitSize / partitioningSteps;
    final boolean selfReference = conf.getBoolean("hama.graph.self.ref", false);

    // new vertex container
    Vertex<V, E, M> vertex = newVertexInstance(vertexClass, conf);
    vertex.runner = this;
    vertex.setVertexType(VertexType.MASTER);

    long startPos = peer.getPos();
    if (startPos == 0)
      startPos = 1L;

    KeyValuePair<Writable, Writable> next = null;
    int steps = 1;

    while ((next = peer.readNext()) != null) {
      boolean finished = false;
      try {
        finished = reader.parseVertex(next.getKey(), next.getValue(), vertex);
      } catch (Exception e) {
        throw new IOException("exception occured during parsing vertex!"
            + e.toString());
      }

      if (!finished){
         continue;
      }

      if (vertex.getEdges() == null) {
        if (selfReference) {
          vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
              .getVertexID(), null)));
        } else {
          vertex.setEdges(Collections.EMPTY_LIST);
        }
      }

      if (selfReference)
        vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));

      if (runtimePartitioning) {
        // int partition = partitioner.getPartition(vertex.getVertexID(), vertex.getNewValue(), numberPeers);
        int partition = getOwner(vertex, verticesOnNode, ownerMap, newRecord);
        peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
      } else {
        vertex.setup(conf);
        vertex.setMasterPeerIndex(peer.getPeerIndex());
        if (selectiveActivate)
          vertex.enableSelectiveActivate();
        if (danglingOPT)
          vertex.enableDanglingOpt();
        verticesMap.put(vertex.getVertexID(), vertex);
      }

      // new vertex container
      vertex = newVertexInstance(vertexClass, conf);
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);

      if (runtimePartitioning) {
        if (steps < partitioningSteps && (peer.getPos() - startPos) >= interval) {
          for (int i = 0; i < numPeer; i++){
            if (i != peer.getPeerIndex()){
              peer.send(peer.getPeerName(i), new GraphJobMessage(newRecord));
            }
          }
                    
          peer.sync();
          ownerMap.putAll(newRecord);
          newRecord.clear();

          steps++;
          createVerticesAndAddOwnerMap(peer, ownerMap, verticesOnNode);
          startPos = peer.getPos();
        }
      }
    }


    // last step
    if (runtimePartitioning) {
      while(steps <= partitioningSteps){
        for (int i = 0; i < numPeer; i++){
          if (i != peer.getPeerIndex()){
            peer.send(peer.getPeerName(i), new GraphJobMessage(newRecord));
          }
        }
        peer.sync();
        ownerMap.putAll(newRecord);
        newRecord.clear();
        
        createVerticesAndAddOwnerMap(peer, ownerMap, verticesOnNode);
        steps++;
      }
    }

    LOG.info("Load vertices finished at " + peer.getSuperstepCount()
        + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());
    LOG.info("OwnerMap size: " + ownerMap.size());
    
    /*
     * If the user want to repair the graph, it should traverse through that
     * local chunk of adjancency list and message the corresponding peer to
     * check whether that vertex exists. In real-life this may be dead-ending
     * vertices, since we have no information about outgoing edges. Mainly this
     * procedure is to prevent NullPointerExceptions from happening.
     */
    if (repairNeeded)
      fennelRepair(peer, (partitioningSteps/20)+1, selfReference, ownerMap);
  }
  

  @SuppressWarnings("unchecked")
  private void fennelRepair(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      int partitioningSteps, boolean selfReference, MapWritable ownerMap) throws IOException,
      SyncException, InterruptedException {
    LOG.info("Start repairing the graph!");
    
    MapWritable newRecord = new MapWritable();
    int numberPeers = peer.getNumPeers();

    int multiSteps = 0;
    MapWritable ssize = new MapWritable();
    ssize.put(new IntWritable(peer.getPeerIndex()),
        new IntWritable(verticesMap.size()));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      int minVerticesSize = Integer.MAX_VALUE;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          int curr = ((IntWritable) e.getValue()).get();
          if (minVerticesSize > curr) {
            minVerticesSize = curr;
          }
        }
      }

      if (minVerticesSize < (partitioningSteps * 2)) {
        multiSteps = minVerticesSize;
      } else {
        multiSteps = partitioningSteps*2;
      }

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("steps"), new IntWritable(multiSteps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    Map<V, Vertex<V, E, M>> tmp = new HashMap<V, Vertex<V, E, M>>();

    int i = 0;
    int syncs = 0;

    for (Vertex<V, E, M> v : verticesMap.values()) {
      for (Edge<V, E> e : v.getEdges()) {
        if (ownerMap.get(e.getDestinationVertexID()) == null){
          int pos = partitioner.getPartition(e.getDestinationVertexID(), null, numberPeers);
          peer.send(peer.getPeerName(pos), new GraphJobMessage(e.getDestinationVertexID()));
          IntWritable pr = new IntWritable(pos);
          newRecord.put(e.getDestinationVertexID(), pr);          
        }
      }

      if (syncs < multiSteps && (i % (verticesMap.size() / multiSteps)) == 0) {
        for (int idx = 0; idx < numberPeers; idx++){
          if (idx != peer.getPeerIndex())
            peer.send(peer.getPeerName(idx), new GraphJobMessage(newRecord));
        }
        peer.sync();
        ownerMap.putAll(newRecord);
        newRecord.clear();
        syncs++;
        GraphJobMessage msg = null;
        while ((msg = peer.getCurrentMessage()) != null) {
          if (msg.isMapMessage()){
            ownerMap.putAll(msg.getMap());
            continue;
          }
          V vertexName = (V) msg.getVertexId();

          Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
          newVertex.setVertexID(vertexName);
          newVertex.runner = this;
          newVertex.setVertexType(VertexType.MASTER);
          newVertex.setMasterPeerIndex(peer.getPeerIndex());
          if (selectiveActivate)
            newVertex.enableSelectiveActivate();
          if (danglingOPT)
            newVertex.enableDanglingOpt();
          if (selfReference) {
            newVertex.setEdges(Collections.singletonList(new Edge<V, E>(
                newVertex.getVertexID(), null)));
          } else {
            newVertex.setEdges(Collections.EMPTY_LIST);
          }
          newVertex.setup(conf);
          tmp.put(vertexName, newVertex);

        }
      }
      i++;
    }

    while (syncs <= multiSteps){
      for (int idx = 0; idx < numberPeers; idx++) {
        if (idx != peer.getPeerIndex())
          peer.send(peer.getPeerName(idx), new GraphJobMessage(newRecord));
      }
      peer.sync();
      ownerMap.putAll(newRecord);
      newRecord.clear();

      GraphJobMessage msg = null;
      while ((msg = peer.getCurrentMessage()) != null) {
        if (msg.isMapMessage()) {
          ownerMap.putAll(msg.getMap());
          continue;
        }

        V vertexName = (V) msg.getVertexId();

        Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
        newVertex.setVertexID(vertexName);
        newVertex.runner = this;
        newVertex.setVertexType(VertexType.MASTER);
        newVertex.setMasterPeerIndex(peer.getPeerIndex());
        if (selectiveActivate)
          newVertex.enableSelectiveActivate();
        if (danglingOPT)
          newVertex.enableDanglingOpt();
        if (selfReference) {
          newVertex.setEdges(Collections.singletonList(new Edge<V, E>(newVertex
              .getVertexID(), null)));
        } else {
          newVertex.setEdges(Collections.EMPTY_LIST);
        }
        newVertex.setup(conf);
        tmp.put(vertexName, newVertex);
        newVertex = null;
      }
      syncs++;
    }

    for (Vertex<V, E, M> e : verticesMap.values()) {
      if (tmp.containsKey((e.getVertexID()))) {
        tmp.remove(e.getVertexID());
        LOG.error("vertex " + e + "shoudl not in the new dangling vertices map");
        System.exit(-1);
      }
    }

    verticesMap.putAll(tmp);
    tmp.clear();

    LOG.info("Repair finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());
    LOG.info("Total # of vertices: " + ownerMap.size());
  }
    

  @SuppressWarnings("unchecked")
  private void repair(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      int partitioningSteps, boolean selfReference) throws IOException,
      SyncException, InterruptedException {
    LOG.info("Start repairing the graph!");

    int multiSteps = 0;
    MapWritable ssize = new MapWritable();
    ssize.put(new IntWritable(peer.getPeerIndex()),
        new IntWritable(verticesMap.size()));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      int minVerticesSize = Integer.MAX_VALUE;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          int curr = ((IntWritable) e.getValue()).get();
          if (minVerticesSize > curr) {
            minVerticesSize = curr;
          }
        }
      }

      if (minVerticesSize < (partitioningSteps * 2)) {
        multiSteps = minVerticesSize;
      } else {
        multiSteps = (partitioningSteps * 2);
      }

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("steps"), new IntWritable(multiSteps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    Map<V, Vertex<V, E, M>> tmp = new HashMap<V, Vertex<V, E, M>>();

    int i = 0;
    int syncs = 0;

    for (Vertex<V, E, M> v : verticesMap.values()) {
      for (Edge<V, E> e : v.getEdges()) {
        peer.send(v.getDestinationPeerName(e),
            new GraphJobMessage(e.getDestinationVertexID()));
      }

      if (syncs < multiSteps && (i % (verticesMap.size() / multiSteps)) == 0) {
        peer.sync();
        syncs++;
        GraphJobMessage msg = null;
        while ((msg = peer.getCurrentMessage()) != null) {
          V vertexName = (V) msg.getVertexId();

          Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
          newVertex.setVertexID(vertexName);
          newVertex.runner = this;
          newVertex.setVertexType(VertexType.MASTER);
          newVertex.setMasterPeerIndex(peer.getPeerIndex());
          if (selfReference) {
            newVertex.setEdges(Collections.singletonList(new Edge<V, E>(
                newVertex.getVertexID(), null)));
          } else {
            newVertex.setEdges(Collections.EMPTY_LIST);
          }
          newVertex.setup(conf);
          if (selectiveActivate)
            newVertex.enableSelectiveActivate();
          if (danglingOPT)
            newVertex.enableDanglingOpt();
          tmp.put(vertexName, newVertex);

        }
      }
      i++;
    }

    while(syncs <= multiSteps){
      peer.sync();
      GraphJobMessage msg = null;
      while ((msg = peer.getCurrentMessage()) != null) {
        V vertexName = (V) msg.getVertexId();

        Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
        newVertex.setVertexID(vertexName);
        newVertex.runner = this;
        newVertex.setVertexType(VertexType.MASTER);
        newVertex.setMasterPeerIndex(peer.getPeerIndex());
        if (selfReference) {
          newVertex.setEdges(Collections.singletonList(new Edge<V, E>(newVertex
              .getVertexID(), null)));
        } else {
          newVertex.setEdges(Collections.EMPTY_LIST);
        }
        newVertex.setup(conf);
        if (selectiveActivate)
          newVertex.enableSelectiveActivate();
        if (danglingOPT)
          newVertex.enableDanglingOpt();
        tmp.put(vertexName, newVertex);
        newVertex = null;
      }
      syncs++;
    }

    for (Vertex<V, E, M> e : verticesMap.values()) {
      if (tmp.containsKey((e.getVertexID()))) {
        tmp.remove(e.getVertexID());
      }
    }

    verticesMap.putAll(tmp);
    tmp.clear();

    LOG.info("Repair finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());
  }

  @SuppressWarnings("unchecked")
  private void createReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isReplicaMessage(), true);

      V sourceVertexId = (V) msg.getVertexId();
      M vertexValue = (M) msg.getVertexValue();
      List<Edge<V, E>> edges = (List<Edge<V, E>>) msg.getEdges();
      int peerIndex = msg.getPeerIndex().get();

      Vertex<V, E, M> replica;
      if (!replicasMap.containsKey(sourceVertexId)) {
        // create replica
        replica = newVertexInstance(vertexClass, conf);
        replica.runner = this;
        if (selectiveActivate)
          replica.enableSelectiveActivate();
        if (danglingOPT)
          replica.enableDanglingOpt();
        replica.setVertexType(VertexType.REPLICA);
        replica.setVertexID(sourceVertexId);
        replica.setNewValue(vertexValue);
        replica.setMasterPeerIndex(peerIndex);
        replica.setEdges(edges);

        replicasMap.put(sourceVertexId, replica);
      } else {
        Assert.assertTrue(false);
      }
    }
  }

  // load replicas
  // TODO: selfReference issue
  @SuppressWarnings("unchecked")
  private void loadReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start loading replicas");

    int sourcePeerIndex = peer.getPeerIndex();
    int step = 1, replicaSteps = multiSteps;
    int count = 0, interval = verticesMap.size() / replicaSteps;

    // send msgs to destination vertices
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      Map<Integer, List<Edge<V, E>>> tmp = new HashMap<Integer, List<Edge<V, E>>>();
      for (Edge<V, E> edge : vertex.getEdges()) {
        int destinationPeerIndex = partitioner.getPartition(
            edge.getDestinationVertexID(), (M) edge.getValue(),
            peer.getNumPeers());
        if (selectiveActivate)
          vertex.addNeighborMap(edge.getDestinationVertexID(),
              destinationPeerIndex);
        
        List<Edge<V, E>> edges = tmp.get(destinationPeerIndex);
        if (edges == null){
          edges = new ArrayList<Edge<V, E>>();
          tmp.put(destinationPeerIndex, edges);
        }
        edges.add(edge);
      }
      vertex.setEdges(null); // this is necessary, because the next setEdge may be not reachable
      for (Entry<Integer, List<Edge<V, E>>> entry : tmp.entrySet()){
        if (entry.getKey() != sourcePeerIndex){
          String peerName = peer.getPeerName(entry.getKey());          
          peer.send(peerName,
              new GraphJobMessage(vertex.getVertexID(), vertex.getNewValue(), entry.getValue(), new IntWritable(peer.getPeerIndex())));
          vertex.addReplicaLocalID(entry.getKey(), -2);
        } else {
          vertex.setEdges(entry.getValue());
        }
      }
      count++;
      
      if (step < replicaSteps && count >= interval) {
        peer.sync();
        createReplicas(peer);
        step++;
        count = 0;
      }
    }
    Assert.assertEquals(step, replicaSteps);

    // last step
    peer.sync();
    createReplicas(peer);

    LOG.info("Load replicas finished at " + peer.getSuperstepCount()
        + " steps.");
    LOG.info("Total # of local replicas: " + replicasMap.size());
    LOG.info("Total # of new replicas: " + localReplicaAdd);
  }
  
  /**
   * pick mirrors from the replicas
   * @param peer
   * @throws IOException
   * @throws SyncException
   * @throws InterruptedException
   */
  private void addAndPickMirrors(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
          throws IOException, SyncException, InterruptedException {
    LOG.info("starting picking mirrors");
    int peerNum = peer.getNumPeers();
    
    for (String peerName : peer.getAllPeerNames())
      peer.send(peerName, new GraphJobMessage(peer.getPeerIndex(), new IntWritable(verticesMap.size()),
          new IntWritable(replicasMap.size()), new IntWritable(this.localReplicaAdd)));
    
    peer.sync();
    LOG.debug("counting over");
    int[] numOfVertices = new int[peer.getNumPeers()];
    int[] numOfReplicas = new int[peer.getNumPeers()];
    int[] indexForReplica = new int[peer.getNumPeers()];
    
    // init mirror state
    int[] numOfMirror = new int[peer.getNumPeers()];
    int[] indexForMirror = new int[peer.getNumPeers()];
    for (int i = 0; i < peer.getNumPeers(); i++){
      indexForMirror[i] = i;
      numOfMirror[i] = 0;
    }
    
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      if (msg.isSizeMessage()) {
        int peerIndex = msg.getPeerIndex().get();
        numOfVertices[peerIndex] = msg.getVerticesSize().get();
        numOfReplicas[peerIndex] = msg.getReplicasSize().get();
        indexForReplica[peerIndex] = peerIndex;
        if (peerIndex == peer.getPeerIndex())
          numOfReplicas[peerIndex] = Integer.MAX_VALUE;
        //this.globalReplicaAdd += msg.getNewReplica().get();
        if (LOG.isDebugEnabled())
          LOG.debug("peer "+peerIndex+" has "+numOfVertices[peerIndex]+" vertices and "+numOfReplicas[peerIndex]+" replicas");
      }
    }
    
    //sort the number of replicas array
    for (int i = 0; i < peer.getNumPeers()-1; i++) {
      for (int j = i; j < peer.getNumPeers(); j++) {
        if (numOfReplicas[i] > numOfReplicas[j]) {
          int t = numOfReplicas[i];
          numOfReplicas[i] = numOfReplicas[j];
          numOfReplicas[j] = t;
          
          t = indexForReplica[i];
          indexForReplica[i] = indexForReplica[j];
          indexForReplica[j] = t;
        }
      }
    }
    
    
    int numOfMirrorRequired = conf.getInt(Constants.MACHINE_FAILURE, Constants.DEFAULT_MACHINE_FAILURE);
    int i = 0;
    int stage = 1;
    
    // heuristic balance algorithm
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      int j = i;
      int numReplica = vertex.getNumOfLocalIDs();
      while (numReplica < numOfMirrorRequired){
        if (indexForReplica[j] != peer.getPeerIndex()
            && vertex.getReplicaLocalIDAtIndex(indexForReplica[j]) == -1L) {
          peer.send(peer.getPeerName(indexForReplica[j]), new GraphJobMessage(vertex.getVertexID(), vertex.getNewValue(), null, new IntWritable(peer.getPeerIndex())));
          vertex.addMirrorIndex(indexForReplica[j]);
          vertex.addReplicaLocalID(indexForReplica[j], -3); // different from normal replica
          numOfReplicas[j] += peerNum;
          numReplica++;
          this.localReplicaAdd++;
          
          numOfMirror[indexForReplica[j]]++;
        }
        j = (j+1) % (peerNum-1);
      }
      if (numOfReplicas[i] > numOfReplicas[stage]){
        stage++;
      }
      i = (i+1)%stage;
    }
    
    
    
    peer.sync();
    createReplicas(peer);
    
    
    i = 0;
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      while (vertex.getNumOfMirror() < numOfMirrorRequired) {
        if (vertex.getReplicaLocalIDAtIndex(i) != -1L && vertex.getReplicaLocalIDAtIndex(i) != -3L) {
          vertex.addMirrorIndex(i);
          vertex.addReplicaLocalID(i, -3);
          numOfMirror[i]++;
        }
        i = (i+1) % peerNum;
      }
    }
    
    LOG.debug("picked mirror from existing replicas");    
   }

  
  // load replicas (fill in-edges and local out-edges)
  // TODO: selfReference issue
  private void fennelLoadReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start fennel loading replicas");

    int sourcePeerIndex = peer.getPeerIndex();
    int step = 1, replicaSteps = multiSteps;
    int count = 0, interval = verticesMap.size() / replicaSteps;

    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      Map<Integer, List<Edge<V, E>>> neighborOwner = new HashMap<Integer, List<Edge<V, E>>>();
      for (Edge<V, E> edge : vertex.getEdges()) {
        int destinationPeerIndex = ((IntWritable) ownerMap.get(edge.getDestinationVertexID())).get();
        if (selectiveActivate)
          vertex.addNeighborMap(edge.getDestinationVertexID(),
              destinationPeerIndex);

        List<Edge<V, E>> edges = neighborOwner.get(destinationPeerIndex);
        if (edges == null){
          edges = new ArrayList<Edge<V, E>>();
          neighborOwner.put(destinationPeerIndex, edges);
        }
        edges.add(edge);
      }
      vertex.setEdges(null); // this is necessary, because the next setEdge may be not reachable
      for (Entry<Integer, List<Edge<V, E>>> entry : neighborOwner.entrySet()){
        if (entry.getKey() != sourcePeerIndex){
        String peerName = peer.getPeerName(entry.getKey());     
          peer.send(peerName,
              new GraphJobMessage(vertex.getVertexID(), vertex.getNewValue(), entry.getValue(), new IntWritable(peer.getPeerIndex())));
          vertex.addReplicaLocalID(entry.getKey(), -2);
        } else {
          vertex.setEdges(entry.getValue());
        }
      }
      count++;
      if (step < replicaSteps && count >= interval) {
        peer.sync();
        createReplicas(peer);
        step++;
        count = 0;
      }
    }

    // last step
    while(step <= replicaSteps){
      peer.sync();
      createReplicas(peer);
      step++;
    }

    LOG.info("Load replicas finished at " + peer.getSuperstepCount()
        + " steps.");
    LOG.info("Total # of local replicas: " + replicasMap.size());
    LOG.info("Total # of new replicas: " + localReplicaAdd);
  }
  
  
  private void zonePartition(Vertex<V, E, M>[] vertices, int start){
    Map<Integer, List<Vertex<V, E, M>>> masterToReplica = new HashMap<Integer, List<Vertex<V, E, M>>>();
    for (int i = start; i < vertices.length; i++)
    {
      List<Vertex<V, E, M>> vl = masterToReplica.get(vertices[i].getMasterPeerIndex());
      if (vl == null){
        vl = new LinkedList<Vertex<V, E, M>>();
        masterToReplica.put(vertices[i].getMasterPeerIndex(), vl);
      }
      vl.add(vertices[i]);
    }
    int i = start;
    for (Entry<Integer, List<Vertex<V, E, M>>> entry : masterToReplica.entrySet()){
      for (Vertex<V, E, M> v: entry.getValue()){
        v.setLocalID(i);
        vertices[i] = v;
        i++;
      }
    }
    masterToReplica = null;
  }
  
  @SuppressWarnings("unchecked")
  private void assignLocalID(){
    float reserveRatio = (1f/this.getPeer().getNumPeers()) * 5 + 1;
    if (reserveRatio > 2)
      reserveRatio = 2;
    int reserveSize = (int) (this.replicasMap.size() * reserveRatio);
    this.vertices = new Vertex[this.verticesMap.size()+reserveSize]; // preserve some space for migration approach 
    LOG.info("vertices length: " + vertices.length);
    this.localVertexNum = this.verticesMap.size()+this.replicasMap.size();
    int j = 0;
    for(Vertex<V, E, M> v : verticesMap.values()){
      vertices[j] = v;
      v.setLocalID(j);
      j++;
    }
    
    for(Vertex<V, E, M> v : replicasMap.values()){
      vertices[j] = v;
      v.setLocalID(j);
      j++;
    }
    if (conf.getBoolean(Constants.SORT, false)){
      LOG.info("vertices size:" + vertices.length + " master size:" + verticesMap.size());
      long sortTime = System.currentTimeMillis();
      zonePartition(vertices, verticesMap.size());
      LOG.info("Time to sort the replicas: " + (System.currentTimeMillis()-sortTime));
    }
  }
  
  @SuppressWarnings({ "rawtypes", "unchecked" })
  private void buildTopology() throws IOException{
    for (Vertex replica : replicasMap.values()) {
      List<Edge<V, E>> edges = replica.getEdges();      
      if (edges != null){
        for (Edge<V, E> edge : edges) {
          Vertex<V, E, M> destinationVertex = verticesMap.get(edge.getDestinationVertexID());
          if (destinationVertex == null)
          {
            LOG.error(replica + " points to a null vertices" + edge);
            throw new IOException(replica + " points to a null vertices" + edge);
          }
          E valueOnEdge = edge.getValue();
        
          RefEdge<V, E, M> in = new RefEdge<V, E, M>((V)replica.getVertexID(), replica.getLocalID(), valueOnEdge, RefEdgeType.INEDGE);
          RefEdge<V, E, M> out = new RefEdge<V, E, M>(destinationVertex.getVertexID(), destinationVertex.getLocalID(), valueOnEdge, RefEdgeType.OUTEDGE);
          destinationVertex.addInEdge(in);
          replica.addOutEdge(out);
          in.setvEID(replica.getOutEdges().size()-1);
          out.setvEID(destinationVertex.getInEdges().size()-1);
        }
        Assert.assertEquals(true, replica.getEdges().size() == replica.getOutEdges().size());
        replica.setEdges(null);
      }
    }

    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      List<Edge<V, E>> edges = vertex.getEdges();
      if (edges != null) {
        for (Edge<V, E> edge : edges) {
          Vertex<V, E, M> destinationVertex = verticesMap.get(edge.getDestinationVertexID());
          if (destinationVertex == null) {
            LOG.error(vertex + " points to a null vertices" + edge);
            throw new IOException(vertex + " points to a null vertices" + edge);
          }

          RefEdge<V, E, M> in = new RefEdge<V, E, M>(vertex.getVertexID(), vertex.getLocalID(), edge.getValue(),RefEdgeType.INEDGE);
          RefEdge<V, E, M> out = new RefEdge<V, E, M>(destinationVertex.getVertexID(), destinationVertex.getLocalID(), edge.getValue(),RefEdgeType.OUTEDGE);
          destinationVertex.addInEdge(in);
          vertex.addOutEdge(out);
          in.setvEID(vertex.getOutEdges().size()-1);
          out.setvEID(destinationVertex.getInEdges().size()-1);
        }
      }
      vertex.setEdges(null);
    }

    RefEdge.vertexSet = vertices; // this is must, because the index topology
    LyraMirrorMessage.vertexSet = vertices; // this is used to replicate mirror information.
  }
  
  
  /**
   * register replicas and store fault tolerance information on master vertices
   * @param peer
   * @throws IOException
   * @throws SyncException
   * @throws InterruptedException
   */
  @SuppressWarnings({ "unchecked", "rawtypes" })
  private void registerReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start registering replicas");
    int registerCount = 0;
    int step = 1, registerSteps = (multiSteps - 1) / 2 + 1;
    int count = 0, interval = replicasMap.size() / registerSteps;

    for (Vertex<V, E, M> vertex : replicasMap.values()) {
      peer.send(
          peer.getPeerName(vertex.getMasterPeerIndex()),
          new GraphJobMessage(new IntWritable(peer.getPeerIndex()), vertex.getVertexID(), new IntWritable(vertex.getLocalID()), vertex.getOutEdges()));

      count++;
      if (step < registerSteps && count >= interval) {
        peer.sync();
        GraphJobMessage msg = null;
        int remoteID = 0;
        registerCount += peer.getNumCurrentMessages();
        while ((msg = peer.getCurrentMessage()) != null) {
          Assert.assertEquals(msg.isRegisterMessage(), true);

          int peerIndex = msg.getPeerIndex().get();
          V vertexID = (V) msg.getVertexId();
          remoteID = msg.getLocalID().get();

          Vertex<V, E, M> master = verticesMap.get(vertexID);
          master.addReplicaLocalID(peerIndex, remoteID);
          master.addReplicaFTInfo(peerIndex, (List<RefEdge>) (msg.getRefEdges()));
        }
        step++;
        count = 0;
      }
    }
    Assert.assertEquals(step, registerSteps);

    // last step
    peer.sync();
    GraphJobMessage msg = null;
    int remoteID = 0;
    registerCount += peer.getNumCurrentMessages();
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isRegisterMessage(), true);

      int peerIndex = msg.getPeerIndex().get();
      V vertexID = (V) msg.getVertexId();
      remoteID = msg.getLocalID().get();

      Vertex<V, E, M> master = verticesMap.get(vertexID);
      master.addReplicaLocalID(peerIndex, remoteID);
      master.addReplicaFTInfo(peerIndex, (List<RefEdge>) (msg.getRefEdges()));
    }
    
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      vertex.addSelfFTInfo();
    }

    LOG.info("Register " + registerCount + " replicas finished at " + peer.getSuperstepCount() + " steps.");
  }
  
  private void replicateMirrorInfo(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    int numOfMirrorRequired = conf.getInt(Constants.MACHINE_FAILURE, Constants.DEFAULT_MACHINE_FAILURE);
    if (numOfMirrorRequired == 0)
      return;
    
    LOG.info("start to replicate the mirror information for fault tolerance");
    
    @SuppressWarnings("unchecked")
    LyraMessageBundle<LyraMirrorMessage>[] bundles = new LyraMessageBundle[this.getPeer().getNumPeers()];
    for (int i = 0; i < bundles.length; i++)
      bundles[i] = new LyraMessageBundle<LyraMirrorMessage>();
    
    if (this.migrationTask == null){
      for (int j = 0; j < this.localMasterNum; j++) {
        Vertex<V, E, M> vertex = vertices[j];
        for (int i = 0; i < vertex.getNumOfMirror(); i++) {
          int host = vertex.getMirrorIndex(i);
          int localID = vertex.getLocalIDList()[host];
          bundles[host].addMessage(new LyraMirrorMessage(vertex, localID));
        }
      }
    } else {
      for (int j = 0; j < this.localVertexNum; j++) {
        Vertex<V, E, M> vertex = vertices[j];
        if (vertex != null && vertex.getVertexType() == VertexType.MASTER) {
          for (int i = 0; i < vertex.getNumOfMirror(); i++) {
            int host = vertex.getMirrorIndex(i);
            int localID = vertex.getLocalIDList()[host];
            bundles[host].addMessage(new LyraMirrorMessage(vertex, localID));
          }
        }
      }
    }
    
    int i = peer.getPeerIndex();
    int j = 0;
    for (j = 0; j < bundles.length; j++) {
      if (!bundles[i].isEmpty()) {
        int n = bundles[i].size() / bundleSize;
        if (n > 1) {
          List<LyraMessageBundle<LyraMirrorMessage>> list = bundles[i].split(n);

          for (LyraMessageBundle<LyraMirrorMessage> bundle : list){
            peer.completeMirror(peer.getPeerName(i), bundle);
          }
        } else {
          peer.completeMirror(peer.getPeerName(i), bundles[i]);
        }
      }
      i = (i+1)%peer.getNumPeers();
    }

  }
  
  private void replicateMirrorInfo(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer, Collection<Vertex<V, E, M>> changedMasters)
      throws IOException, SyncException, InterruptedException {
    int numOfMirrorRequired = conf.getInt(Constants.MACHINE_FAILURE, Constants.DEFAULT_MACHINE_FAILURE);
    if (numOfMirrorRequired == 0)
      return;
    
    LOG.info("start to replicate the mirror information for changed masters");
    
    @SuppressWarnings("unchecked")
    LyraMessageBundle<LyraMirrorMessage>[] bundles = new LyraMessageBundle[this.getPeer().getNumPeers()];
    for (int i = 0; i < bundles.length; i++)
      bundles[i] = new LyraMessageBundle<LyraMirrorMessage>();
    
    for (Vertex<V, E, M> vertex : changedMasters) {
      for (int i = 0; i < vertex.getNumOfMirror(); i++) {
        int host = vertex.getMirrorIndex(i);
        int localID = vertex.getLocalIDList()[host];
        bundles[host].addMessage(new LyraMirrorMessage(vertex, localID));
      }
    }
    
    int i = peer.getPeerIndex();
    int j = 0;
    for (j = 0; j < bundles.length; j++) {
      if (!bundles[i].isEmpty()) {
        int n = bundles[i].size() / bundleSize;
        if (n > 1) {
          List<LyraMessageBundle<LyraMirrorMessage>> list = bundles[i].split(n);

          for (LyraMessageBundle<LyraMirrorMessage> bundle : list){
            peer.completeMirror(peer.getPeerName(i), bundle);
          }
        } else {
          peer.completeMirror(peer.getPeerName(i), bundles[i]);
        }
      }
      i = (i+1)%peer.getNumPeers();
    }

  }

  /**
   * 
   * @param replica vertex's type can be mirror, replica, but it should not be master
   * @param targetPeerIndices
   * @return whether this vertex will not participate in recover or not
   */
  private boolean isRecoverMirror(Vertex<V, E, M> replica, List<Integer> targetPeerIndices){
    if (replica.isMirrorReplica == false)
      return false;
    if (!targetPeerIndices.contains(new Integer(replica.getMasterPeerIndex())))
      return false;
    int mirrorIdx = -1;
    for (int i = 0; i < replica.getNumOfMirror(); i++){
      mirrorIdx = replica.getMirrorIndex(i);
      if (!targetPeerIndices.contains(new Integer(mirrorIdx))){
        break;
      }
    }
    Assert.assertEquals(true, (mirrorIdx != -1));
    if (mirrorIdx == this.getPeer().getPeerIndex())
      return true;
    else
      return false;
  }
  
  private static int rrPickIdx = 0;
  
  /**
   * use the correct location information to create new mirror on the surviving machine
   * @param peer
   * @param crashedPeerIndices both new crashed and already crashed peers
   * @param mirror
   * @throws IOException
   * @throws SyncException
   * @throws InterruptedException
   */
  private void replenishMirror(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      List<Integer> crashedPeerIndices, Vertex<V, E, M> mirror) throws IOException, SyncException,
      InterruptedException {
    int numOfMirrorRequired = conf.getInt(Constants.MACHINE_FAILURE, Constants.DEFAULT_MACHINE_FAILURE);
    int numOfReplica = mirror.getNumOfLocalIDs();
    int numofPeer = peer.getNumPeers();
    int myIdx = peer.getPeerIndex();

    do {
      rrPickIdx = (rrPickIdx+1) %  numofPeer;
      
      if (numOfReplica < numOfMirrorRequired){
        if (rrPickIdx != peer.getPeerIndex() && !crashedPeerIndices.contains(rrPickIdx)
            && mirror.getReplicaLocalIDAtIndex(rrPickIdx) == -1) {
          peer.send(peer.getPeerName(rrPickIdx), new GraphJobMessage(mirror, null, mirror.getMasterPeerIndex(), mirror.getLocalID(), selectiveActivate));
          mirror.addReplicaLocalID(rrPickIdx, -3); // different from normal replica
          numOfReplica++;
        }
      }
      
      if (mirror.getNumOfMirror() < numOfMirrorRequired){
        if (mirror.getReplicaLocalIDAtIndex(rrPickIdx) != -1 && !mirror.judgeMirrorHost(rrPickIdx) && rrPickIdx != myIdx) { // pick a mirror from existed replica
          mirror.addMirrorIndex(rrPickIdx);
        }
      } else {
        break;
      }
    } while (true);
  }
  
 
  /**
   * In this implementation, we does not allow the hybrid recovery approach, which means either both approaches during once recovery or one recovery approach after the other recovery approach  
   * @param peer
   * @param crashedPeerIndices
   * @throws IOException
   * @throws SyncException
   * @throws InterruptedException
   */
  @SuppressWarnings({ "rawtypes", "unchecked" })
  private void migratePeer(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      List<Integer> crashedPeerIndices) throws IOException, SyncException,
      InterruptedException {
    LOG.info("start migrating");
    if (migrationTask == null)
      migrationTask = crashedPeerIndices;
    else 
      migrationTask.addAll(crashedPeerIndices);
    
    long migrateTot = System.currentTimeMillis();
    
    boolean fullRecovery = false;
    
    int myIdx = peer.getPeerIndex();
    
    Map<Writable, Integer> vIdToLId = new HashMap<Writable, Integer>();
    
    List<Vertex<V, E, M>> newMasters = new ArrayList<Vertex<V, E, M>>();
    Map<V, Vertex<V, E, M>> changedMasters = new HashMap<V, Vertex<V, E, M>>();

    long migrateP1 = System.currentTimeMillis();
    
    for (int i = 0; i < this.localVertexNum; i++) {
      Vertex<V, E, M> vertex = vertices[i];
      if (vertex.getVertexType() == VertexType.MIRROR) {
        if (isRecoverMirror(vertex, crashedPeerIndices)) {
          this.localMasterNum++;
          vertex.runner = this;
          vertex.upgradeToMaster(crashedPeerIndices, myIdx);
          vertex.setInEdges(new ArrayList<RefEdge<V, E, M>>()); 
          vertex.setup(conf);        
          if (selectiveActivate){
            vertex.enableSelectiveActivate();
            for (Entry<Integer, List<RefEdge>> entry : vertex.getReplicaFTInfo().entrySet()){
              Integer node = entry.getKey();              
              for (RefEdge e : entry.getValue()){
                Assert.assertEquals(true, e.isOutEdge());
                vertex.addNeighborMap((V)e.getVertexID(), node);
              }
            }
          }
          vIdToLId.put(vertex.getVertexID(), new Integer(i)); 
          newMasters.add(vertex);
          if (fullRecovery)
            changedMasters.put(vertex.getVertexID(), vertex);
        }
      }
    }
    
    for (int i = 0; i < peer.getNumPeers(); i++) {
      if (!migrationTask.contains(i))
        peer.send(peer.getPeerName(i), new GraphJobMessage(vIdToLId, myIdx));
    }
    long migrateS1 = System.currentTimeMillis();
    migrateP1 = migrateS1 - migrateP1;

    peer.sync();

    long migrateP2 = System.currentTimeMillis();
    migrateS1 = migrateP2 - migrateS1;


    LOG.debug("Mirror Upgrade Over");
    
    GraphJobMessage msg = null;
    Map<Writable, Integer> tmpMap = new HashMap<Writable, Integer>();
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isMigrateMessage(), true);

      Map<Writable, Integer> vertexIdMap = msg.getVertexIdMap();
      int peerIndex = msg.getPeerIndex().get();
      for (Entry<Writable, Integer> e : vertexIdMap.entrySet()) {
	//the encoding is only OK for machine number < 100
	Integer val = new Integer(e.getValue().intValue() * 100 + peerIndex);
	tmpMap.put(e.getKey(), val);
      }
    }

    for (int i = 0; i < this.localVertexNum; i++) {
      boolean changed = false;
      Vertex<V, E, M> vertex = vertices[i];
      if (vertex.getVertexType() == VertexType.MASTER) {
	Map <Integer, List> sendMap = new HashMap<Integer, List>();
	for (Integer crashedPeer : crashedPeerIndices) {
	  List<RefEdge> edgesToBeRecovered = vertex.getReplicaFTInfo().remove(crashedPeer);
	  if (edgesToBeRecovered == null)
	    continue;
	  changed = true;
	  for (RefEdge<V, E, M> edge : edgesToBeRecovered) {
	    if (tmpMap.containsKey(edge.getVertexID())) {
	      int lVid = tmpMap.get(edge.getVertexID()).intValue();
	      int peerIndex = lVid % 100;
	      lVid = lVid / 100;
	      edge.setlVID(lVid);
	      if (selectiveActivate) {
	        edge.clearState();
	        if (vertex.activateAllNeighbors){
	          edge.activate();
	          edge.commit();
	        } else {
	          List<V> activateList = vertex.getActivateLists(peer.getNumPeers())[crashedPeer];
                  for (int k = 0; k < activateList.size(); k++) {
                    if (edge.getVertexID().equals(activateList.get(k))) {
		      edge.activate();
		      edge.commit();
		      break;
		    }
                  }
		}
                vertex.addNeighborMap(edge.getVertexID(), peerIndex);
	      }
	      
	      // because we replenish ReplicaFTInfo by master itself, this means the edge pair information is wrong; so we can use rebirth recovery approach after this.
	      if (myIdx != peerIndex) {
		List<RefEdge> edgeList = (List<RefEdge>) sendMap.get(new Integer(peerIndex));
		if (edgeList == null) {
		  edgeList = new ArrayList<RefEdge>();
		  sendMap.put(new Integer(peerIndex), edgeList);
		}
		edgeList.add((RefEdge<V, E, M>)edge);
		if (fullRecovery){
		  List<RefEdge> refEdgesOnReplica = vertex.getReplicaFTInfo().get(peerIndex);
		  if (refEdgesOnReplica == null){
		    refEdgesOnReplica = new ArrayList<RefEdge>();
		    vertex.addReplicaFTInfo(peerIndex, refEdgesOnReplica);
		  }
		  refEdgesOnReplica.add(edge);
		}
	      } else {
		vertex.getOutEdges().add(edge);
		if (fullRecovery)
		  vertex.addSelfFTInfo(); // the outEdge in FTInfo maybe a shadow copy (when vertex is the surviving master), so we had better use this interface.
		RefEdge<V, E, M> newInEdge = new RefEdge<V, E, M>(vertex.getVertexID(), i, edge.getValue(), RefEdgeType.INEDGE);
		
		if (selectiveActivate && edge.isAwake()) {
		  newInEdge.activate();
		  newInEdge.commit();
		}
                
		vertices[lVid].getInEdges().add(newInEdge);
		
		newInEdge.setvEID(vertex.getOutEdges().size()-1);
		edge.setvEID(vertices[lVid].getInEdges().size() - 1);                  
	      }
	    }
	  }
	}
        vertex.correctLocationInfo(crashedPeerIndices, myIdx);

	for (Entry<Integer, List> e : sendMap.entrySet()) {
	  int peerIndex = e.getKey().intValue();
	  List<RefEdge> edgeList = (List<RefEdge>) e.getValue();
	  if (vertex.getReplicaLocalIDAtIndex(peerIndex) == -1) {
	    peer.send(peer.getPeerName(peerIndex), new GraphJobMessage(vertex, edgeList, myIdx, vertex.getLocalID(), selectiveActivate));
	    vertex.addReplicaLocalID(peerIndex, -2);
	  } else {
	    peer.send(peer.getPeerName(peerIndex), new GraphJobMessage(vertex.getReplicaLocalIDAtIndex(peerIndex), edgeList, selectiveActivate));
	  }
	}
        
        if (fullRecovery){
          replenishMirror(peer, migrationTask, vertex);
          
          if (changed == true){
            changedMasters.put(vertex.getVertexID(), vertex);
          }
        }
      }
    }

    migrateP2 = System.currentTimeMillis() - migrateP2;

    long migrateS2 = System.currentTimeMillis();
    
    peer.sync();

    migrateS2 = System.currentTimeMillis() - migrateS2;

    LOG.debug("Migrate phase#2 Send topology over");

    long migrateP3 = System.currentTimeMillis();
    
    msg = null;
    
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isMigrateReplicaMessage() || msg.isMigrateMoreEdgeMessage(), true);
      
      if (msg.isMigrateReplicaMessage()) {
        List<AbstractRefEdge<V, E, M>> edgeList = (List<AbstractRefEdge<V, E, M>>) msg.getRefEdges();
        Vertex<V, E, M> replica = (Vertex<V, E, M>) msg.getVertex();
        //TODO: add replica to array; add inEdge to new masters;
        
        vertices[this.localVertexNum] = replica;
        replica.setLocalID(this.localVertexNum);
        this.localVertexNum++;
        replica.setVertexType(VertexType.REPLICA);
        
        replica.commitState();

	if (selectiveActivate)
	  replica.enableSelectiveActivate();
	replica.runner = this;
        
        replica.setOutEdges(new ArrayList<AbstractRefEdge<V, E, M>>());
        if (edgeList != null){
          for (AbstractRefEdge<V, E, M> edge : edgeList) {
            RefEdge<V, E, M> refEdge = (RefEdge<V, E, M>) edge;
            RefEdge<V, E, M> newInEdge = new RefEdge<V, E, M>(replica.getVertexID(), replica.getLocalID(), refEdge.getValue(), RefEdgeType.INEDGE);
          
            refEdge.commit();
            if (selectiveActivate && refEdge.isAwake()) {
              newInEdge.activate();
              newInEdge.commit();
            }
          
            vertices[refEdge.getlVID()].getInEdges().add(newInEdge);
            replica.getOutEdges().add(refEdge);
          
            newInEdge.setvEID(replica.getOutEdges().size() - 1);
            refEdge.setvEID(vertices[refEdge.getlVID()].getInEdges().size() - 1);
          }
        }
        peer.send(peer.getPeerName(msg.getPeerIndex().get()), new GraphJobMessage(replica.getLocalID(), replica.getVertexID(), myIdx, msg.getMasterLID().get()));
      } else if (msg.isMigrateMoreEdgeMessage()){
        List<RefEdge<V, E, M>> edgeList = (List<RefEdge<V, E, M>>) msg.getRefEdges();
        int lid = msg.getLocalID().get();
        
        for (RefEdge<V, E, M> edge : edgeList) {
          vertices[lid].getOutEdges().add(edge);
          RefEdge<V, E, M> newInEdge = new RefEdge<V, E, M>(vertices[lid].getVertexID(), lid, edge.getValue(), RefEdgeType.INEDGE);
         
          edge.commit();
          if (selectiveActivate && edge.isAwake()) {
            newInEdge.activate();
            newInEdge.commit();
          }

          vertices[edge.getlVID()].getInEdges().add(newInEdge);
          newInEdge.setvEID(vertices[lid].getOutEdges().size() - 1);
          edge.setvEID(vertices[edge.getlVID()].getInEdges().size() - 1);

        }
        
      }
    }

    migrateP3 = System.currentTimeMillis() - migrateP3;

    long migrateS3 = System.currentTimeMillis();
    
    peer.sync();

    migrateS3 = System.currentTimeMillis() - migrateS3;

    LOG.debug("Migrate phase#3 assembly topology over");

    long migrateP4 = System.currentTimeMillis();
    
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isMigrateRegisterMessage(), true);
      
      int localID = msg.getLocalID().get();
      int mLID = msg.getMasterLID().get();
      vertices[mLID].addReplicaLocalID(msg.getPeerIndex().get(), localID);
    }
    
   
    for (Vertex<V, E, M> vertex : newMasters) {
      if (danglingOPT){
        if (vertex.getEdgeNum() == 0) {
          vertex.compute(null);
        }
      }
      vertex.awakedByNeighbors();
    }

    migrateP4 = System.currentTimeMillis() - migrateP4;

    LOG.debug("Migrate phase#4 activate self over");
    
    long migrateS4 = System.currentTimeMillis();
    
    if (fullRecovery)
      this.replicateMirrorInfo(peer, changedMasters.values());

    migrateS4 = System.currentTimeMillis() - migrateS4;
    

    migrateTot = System.currentTimeMillis() - migrateTot;

    LOG.info("Migrate time: "+migrateTot);
    LOG.info("Migrate breakdown: "+migrateP1+" "+migrateS1+" "+migrateP2+" "+migrateS2+" "+migrateP3+" "+migrateS3+" "+migrateP4+" "+migrateS4);
  }
  
  
  /**
   * TODO: make the recovery a atomic operation and support the failure during recovery
   * @param peer
   * @param targetPeerIndices
   * @throws IOException
   * @throws SyncException
   * @throws InterruptedException
   */
  private void recoverPeer(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      List<Integer> targetPeerIndices) throws IOException, SyncException,
      InterruptedException {
    LOG.info("Start to help others to recover");
    int subRecovers = conf.getInt(Constants.RECOVER_RPC_CLIENT_NUM, 2);
    try {
      peer.initilizeLyraRecoverSender(subRecovers);
    } catch (ClassNotFoundException e) {
      throw new IOException("init recovery sender failed " + e.toString());
    }

    long recoverStart = System.currentTimeMillis();
    CyclicBarrier barrier = new CyclicBarrier(subRecovers+1);
    for (int i = 0; i < subRecovers; i++){
      new Thread(new RecoverSubthread<V, E, M>(peer, this, i, vertices, localVertexNum, barrier, targetPeerIndices, subRecovers)).start();
    }
    try {
      barrier.await();
    } catch (BrokenBarrierException e) {
      throw new SyncException("barrier operation fails " + e.toString());
    }
    long sendOver = System.currentTimeMillis();
    LOG.info("Time to send recover Message: " + (sendOver-recoverStart));
    peer.lyraSync(Constants.BARRIER_NORMAL); // waiting for the new coming node finishing recover phase and
                                             // receive msg
    long fstBarrier = System.currentTimeMillis();
    LOG.info("Time in first barrier: " + (fstBarrier-sendOver));
    
    peer.lyraSync(Constants.BARRIER_NORMAL); // waiting for the new coming node
    LOG.info("Time in second barrier: " + (System.currentTimeMillis()-fstBarrier));
    
  }

  private void clearDummyState() {
    for (Vertex<V, E, M> vertex : vertices) {
      if (vertex == null)
        continue;
      vertex.clearNewState();
    }
    
    globalStat.getNextState().clear();
  }

  private void recoverFromPeer(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Recieve Message And Start To Recover");
    
    RefEdge.vertexSet = vertices;
    
    peer.registerRecoverMessaging(vertices);
    
    long start = System.currentTimeMillis();
    peer.lyraSync(Constants.BARRIER_NORMAL);
    long firstBarrier = System.currentTimeMillis();
    LOG.info("Time to receive recover message:" + (firstBarrier-start));
    
    //GraphJobMessage msg = null;
    LOG.debug("receive recoverMessage done");

    List<LyraRecoverMessage>[] recoverMsgsArray = peer.uglyCombineRecoverMessages();
    for(List<LyraRecoverMessage> recoverMsgs : recoverMsgsArray){
      for (LyraRecoverMessage msg : recoverMsgs) {
        if (msg.isRecoverStatMessage()) {
          this.iteration = msg.getIter();
          this.numberVertices = msg.getNumVertices();
          this.globalStat.setStat(msg.getOldStat());
        } else {
          LOG.error("Something has wrong with sending recover messages");
          throw new IOException("Something has wrong with sending recover messages");
        }
      }
    }
    
    long buildTime = System.currentTimeMillis();
    LOG.info("Time to build structure in memory: " + (buildTime-firstBarrier));
    
    int subRecovees = 4;
    CyclicBarrier barrier1 = new CyclicBarrier(subRecovees+1);
    CyclicBarrier barrier2 = new CyclicBarrier(subRecovees+1);
    CyclicBarrier barrier3 = new CyclicBarrier(subRecovees+1);
    
    int[] masterNums = new int[subRecovees];
    for (int i = 0; i < subRecovees; i++){
      masterNums[i] = 0;
    }
    int[] verticesNums = new int[subRecovees];
    for (int i = 0; i < subRecovees; i++){
      verticesNums[i] = 0;
    }
    
    for (int i = 0; i < subRecovees; i++){
      new Thread(new RecoveeBuildSubthread<V, E, M>(this, this.conf, vertices, i, subRecovees, barrier1, barrier2, barrier3, masterNums, verticesNums)).start();
    }
    try {
      barrier1.await();
      barrier2.await();
      barrier3.await();
    } catch (BrokenBarrierException e) {
      throw new SyncException("barrier operation fails " + e.toString());
    }
    LOG.debug("Pass All barriers");
    this.localMasterNum = 0;
    for (int i = 0; i < subRecovees; i++){
      this.localMasterNum += masterNums[i];
    }
    this.localVertexNum = 0;
    for (int i = 0; i < subRecovees; i++){
      this.localVertexNum += verticesNums[i];
    }
    LOG.debug("localVertexNum: " + this.localMasterNum);

        
    // re-init some state
    this.initRunner(peer);

    long timeRedo = System.currentTimeMillis();
    LOG.info("Time to redo: " + (timeRedo-buildTime));
    LOG.info("Recover all the state");
    peer.lyraSync(Constants.BARRIER_NORMAL);
    long secondBarrier = System.currentTimeMillis();
    LOG.info("Second Barrier: " + (secondBarrier-timeRedo));
    LOG.info("Start the following computation");

  }

  /**
   * Counts vertices globally by sending the count of vertices in the map to the
   * other peers.
   * @throws IOException 
   * @throws InterruptedException 
   * @throws SyncException 
   * @throws Exception 
   */
  private void countGlobal(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) throws IOException, SyncException, InterruptedException
      {

    MapWritable sizeMap = new MapWritable();
    sizeMap.put(new IntWritable(0), new IntWritable(this.localMasterNum));
    sizeMap.put(new IntWritable(1), new IntWritable(this.localReplicaNum));
    sizeMap.put(new IntWritable(2), new IntWritable(this.localReplicaAdd));
    //sizeMap.put(new IntWritable(3), new IntWritable(this.localMirror));  // no chance to get the exact number of mirror except visiting the replicas
    for (String peerName : peer.getAllPeerNames())
      peer.send(peerName, new GraphJobMessage(sizeMap));
    
    LOG.info("peer "+peer.getPeerIndex() + " replica and mirror NUM " 
        + localMasterNum + " " + localReplicaNum + " " 
        + this.localReplicaAdd + " " + this.localMirror);

    peer.sync();
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(true, msg.isMapMessage());
      sizeMap = msg.getMap();
      for (Entry<Writable, Writable> entry : sizeMap.entrySet()){
        IntWritable key = (IntWritable) entry.getKey();
        IntWritable value = (IntWritable) entry.getValue();
        switch(key.get()){
          case 0: numberVertices += value.get();break;
          case 1: numberReplicas += value.get();break;
          case 2: globalReplicaAdd += value.get();break;
          //case 3: numberMirror += value.get();break;
          default:
            LOG.error("the key-value pair in map is wrong");
            throw new IOException("the key-value pair in map is wrong");
        }
      }
    }
    this.globalActiveVertexCount = this.numberVertices;

    if (isMasterTask(peer))
      peer.getCounter(GraphJobCounter.INPUT_VERTICES).increment(numberVertices);

    LOG.info("Counting finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of vertices: " + numberVertices + " and # of replicas: " + numberReplicas +
             " and # of extra replicas: " + globalReplicaAdd);
  }

  /**
   * @return the number of vertices, globally accumulated.
   */
  public final long getNumberVertices() {
    return numberVertices;
  }

  /**
   * @return the current number of iterations.
   */
  public final long getNumberIterations() {
    return iteration;
  }

  /**
   * @return the defined number of maximum iterations, -1 if not defined.
   */
  public final int getMaxIteration() {
    return maxIteration;
  }

  /**
   * @return the defined partitioner instance.
   */
  public final Partitioner<V, M> getPartitioner() {
    return partitioner;
  }

  /**
   * Gets the last aggregated value at the given index. The index is dependend
   * on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final Writable getLastAggregatedValue(int index) {
    return aggregationRunner.getLastAggregatedValue(index);
  }

  /**
   * Gets the last aggregated number of vertices at the given index. The index
   * is dependend on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final IntWritable getNumLastAggregatedVertices(int index) {
    return aggregationRunner.getNumLastAggregatedVertices(index);
  }

  /**
   * @return the peer instance.
   */
  public final BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> getPeer() {
    return peer;
  }

  /**
   * Checks if this is a master task. The master task is the first peer in the
   * peer array.
   */
  public static boolean isMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName().equals(getMasterTask(peer));
  }

  /**
   * @return the name of the master peer, the name at the first index of the
   *         peers.
   */
  public static String getMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName(0);
  }

  public AbstractGlobalState getGlobalStat() {
    return globalStat;
  }

  public boolean isSelectiveActivate() {
    return selectiveActivate;
  }

  public boolean isDanglingOPT() {
    return danglingOPT;
  }

  /**
   * @return a new vertex instance
   */
  public static <V extends Writable, E extends Writable, M extends Writable> Vertex<V, E, M> newVertexInstance(
      Class<?> vertexClass, Configuration conf) {
    @SuppressWarnings("unchecked")
    Vertex<V, E, M> vertex = (Vertex<V, E, M>) ReflectionUtils.newInstance(
        vertexClass, conf);
    return vertex;
  }

}
