/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.IOException;
import java.util.List;
import java.util.Map.Entry;
import java.util.concurrent.CyclicBarrier;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.Writable;
import org.apache.hama.bsp.AbstractVertex.VertexType;

public class RecoveeBuildSubthread<V extends Writable, E extends Writable, M extends Writable>  implements Runnable {
  private static final Log LOG = LogFactory.getLog(RecoveeBuildSubthread.class);
  
  private final int id;
  private GraphJobRunner<V, E, M> runner;
  private Configuration conf;
  private CyclicBarrier barrier1;
  private CyclicBarrier barrier2;
  private CyclicBarrier barrier3;
  private int[] masterNums;
  private int[] verticesNums;
  
  private Vertex<V, E, M>[] vertices;
  private int handlers;

  public RecoveeBuildSubthread(
      GraphJobRunner<V, E, M> runner, Configuration conf,
      Vertex<V, E, M>[] vertices, int id, int handlers, CyclicBarrier barrier1, CyclicBarrier barrier2, CyclicBarrier barrier3, int[] masterNums, int[] verticesNum) {
    super();
    this.runner = runner;
    this.conf = conf;
    this.id = id;
    this.vertices = vertices;
    this.handlers = handlers;
    
    this.barrier1 = barrier1;
    this.barrier2 = barrier2;
    this.barrier3 = barrier3;
    this.masterNums = masterNums;
    this.verticesNums = verticesNum;
  }

  @SuppressWarnings({ "rawtypes", "unchecked" })
  public void run(){    
    int tasks = vertices.length / handlers;
    int begin = tasks * this.id;
    int end = tasks * (this.id+1);
    if (id == handlers-1){
      end = vertices.length;
    }

    boolean selectiveActivate = runner.isSelectiveActivate();
    boolean danglingOPT = runner.isDanglingOPT();

    for (int i = begin; i < end; i++) {
      Vertex<V, E, M> vertex = vertices[i];
      if (vertex == null)
        continue;
      if (selectiveActivate)
        vertex.enableSelectiveActivate();
      if (danglingOPT)
        vertex.enableDanglingOpt();
      vertex.runner = runner;
      if (vertex.isMaster()){
        masterNums[id]++;
        if (selectiveActivate){
          vertex.enableSelectiveActivate();
          for (Entry<Integer, List<RefEdge>> entry : vertex.getReplicaFTInfo().entrySet()){
            Integer node = entry.getKey();
            
            for (RefEdge e : entry.getValue()){
              Assert.assertEquals(true, e.isOutEdge());
              vertex.addNeighborMap((V)e.getVertexID(), node);
            }
          }
        }
        vertex.setup(conf); // call setup after merge
      }
      
      try {
        vertex.localActivateVertices();
      } catch (IOException e) {
        e.printStackTrace();
        LOG.error(e);
      }
    }
    LOG.debug("Finish activate at" + this.id);
    try {
      barrier1.await();
    } catch (Exception e) {
      e.printStackTrace();
      LOG.error(e);
    }

    for (int i = begin; i < end; i++) {
      Vertex<V, E, M> vertex = vertices[i];
      if (vertex != null){
        vertex.commitState();
        verticesNums[id]++;
      }
    }
    LOG.debug("Finish commitState at recovery node at" + this.id);
    try {
      barrier2.await();
    } catch (Exception e) {
      e.printStackTrace();
      LOG.error(e);
    }


    if (danglingOPT){
      for (int i = begin; i < end; i++) {
        Vertex<V, E, M> vertex = vertices[i];
        if (vertex == null)
          continue;
        if (vertex.getVertexType() == VertexType.MASTER && vertex.getEdgeNum() == 0){
          try {
            vertex.compute(null);
          } catch (IOException e) {
            e.printStackTrace();
            LOG.error(e);
          }
          vertex.commitState();
        }
      }
    }
    try {
      barrier3.await();
    } catch (Exception e) {
      e.printStackTrace();
      LOG.error(e);
    }

    LOG.info("subRecovee " + this.id + " has finished at" + this.id);
    return;
  }

}
