/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.util.List;
import java.util.concurrent.CyclicBarrier;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.io.Writable;
import org.apache.hama.bsp.AbstractVertex.VertexType;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.LyraRecoverMessage;
import org.apache.hama.bsp.LyraRecoverMessageBundle;

public class RecoverSubthread<V extends Writable, E extends Writable, M extends Writable>  implements Runnable {
  private static final Log LOG = LogFactory.getLog(RecoverSubthread.class);
  
  private final int id;
  private BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer;
  private GraphJobRunner<V, E, M> runner;
  private CyclicBarrier barrier;

  private Vertex<V, E, M>[] vertices;
  private int vertexNum; 
  @SuppressWarnings("rawtypes")
  private LyraRecoverMessageBundle[] subBundles;
  private List<Integer> targetPeerIndices;
  private int handlers;

  @SuppressWarnings("rawtypes")
  public RecoverSubthread(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer, GraphJobRunner<V, E, M> runner,
      int id, Vertex<V, E, M>[] vertices, int vertexNum, CyclicBarrier barrier, List<Integer> targetPeerIndices, int handlers) {
    super();
    this.peer = peer;
    this.runner = runner;
    this.id = id;
    this.vertices = vertices;
    this.vertexNum = vertexNum;
    this.barrier = barrier;
    this.targetPeerIndices = targetPeerIndices;
    this.handlers = handlers;

    subBundles = new LyraRecoverMessageBundle[peer.getNumPeers()];
    for (int i = 0; i < subBundles.length; i++)
      subBundles[i] = new LyraRecoverMessageBundle();
  }

  
  /**
   * 
   * @param replica vertex's type can be mirror, replica, but it should not be master
   * @param targetPeerIndices
   * @return whether this vertex will not participate in recover or not
   */
  private boolean isRecoverMirror(Vertex<V, E, M> replica, List<Integer> targetPeerIndices){
    if (replica.isMirrorReplica == false)
      return false;
    if (!targetPeerIndices.contains(new Integer(replica.getMasterPeerIndex())))
      return false;
    int mirrorIdx = -1;
    for (int i = 0; i < replica.getNumOfMirror(); i++){
      mirrorIdx = replica.getMirrorIndex(i);
      if (!targetPeerIndices.contains(new Integer(mirrorIdx))){
        break;
      }
    }
    Assert.assertEquals(true, (mirrorIdx != -1));
    if (mirrorIdx == this.peer.getPeerIndex())
      return true;
    else
      return false;    
  }

  public void run(){    
    int tasks = this.vertexNum / handlers;
    int begin = tasks * this.id;
    int end = tasks * (this.id+1);
    if (id == handlers-1){
      end = this.vertexNum;
    }
    for (int i = begin; i < end; i++){
      Vertex<V, E, M> vertex = vertices[i];
      if (vertex.getVertexType() == VertexType.MASTER){
        for (Integer index : targetPeerIndices) {
          int targetPeerIndex = index.intValue();
          String peerName = peer.getPeerName(targetPeerIndex);
          if (vertex.getReplicaLocalIDAtIndex(targetPeerIndex) != -1L) {
            boolean isMirror = false;
            for (int k = 0; k < vertex.getNumOfMirror(); k++) {
              if (vertex.getMirrorIndex(k) == targetPeerIndex) {
                isMirror = true;
                break;
              }
            }
            peer.transferRecoveryMessage(id, peerName, new LyraRecoverMessage(vertex, targetPeerIndex, isMirror));
          }
        }
      } else if (vertex.getVertexType() == VertexType.MIRROR){
        if (isRecoverMirror(vertex, targetPeerIndices)) {
          peer.transferRecoveryMessage(id, peer.getPeerName(vertex.getMasterPeerIndex()), new LyraRecoverMessage(vertex)); // recover master
          for (Integer index : targetPeerIndices) {
            int targetPeerIndex = index.intValue();
            String peerName = peer.getPeerName(targetPeerIndex);
            if (index !=  vertex.getMasterPeerIndex() && vertex.getReplicaLocalIDAtIndex(targetPeerIndex) != -1L) {
              boolean isMirror = false;
              for (int k = 0; k < vertex.getNumOfMirror(); k++) {
                if (vertex.getMirrorIndex(k) == targetPeerIndex) {
                  isMirror = true;
                  break;
                }
              }
               peer.transferRecoveryMessage(id,
                  peerName,
                  new LyraRecoverMessage(vertex, targetPeerIndex, isMirror));
            }
          }
        }
      }
    }

    if (this.id == 0){
      for (Integer i : targetPeerIndices) {
        peer.transferRecoveryMessage(id, peer.getPeerName(i.intValue()), new LyraRecoverMessage(runner.getGlobalStat().getState(),
            runner.getNumberIterations(), runner.getNumberVertices()));
      }
    }
    
    for (Integer newbie : targetPeerIndices){
      peer.clearRecoveryMessageBuffer(id, peer.getPeerName(newbie));
    }
    
    try {
      this.barrier.await();
    } catch (Exception e) {
      e.printStackTrace();
      LOG.error(e);
    }
    LOG.info("subRecover " + this.id + " has finished");
    return;
  }

}