/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hama.bsp.AbstractRefEdge;

/**
 * (BSP-g) The ref-edge class
 */
public final class RefEdge<V extends Writable, E extends Writable, M extends Writable> 
  extends AbstractRefEdge<V, E, M> {
  
  @SuppressWarnings("rawtypes")
  public static Vertex[] vertexSet;
  
  public static enum RefEdgeType {
    INEDGE, OUTEDGE
  }
  
  private RefEdgeType edgeType;
  
  protected V vid;
  protected int lVID; // the index number of a vertex in the vertexSet
  protected int vEID; // the index number of an edge in the edgeList
  
  //private Vertex<V, E, M> vertex; 
  private final E cost;
  private boolean awaked; // check, means the neighbor awaked the vertex in last iteration
  protected boolean active; // set, the neighbor awaked the vertex in this iteration
  //private boolean active; // set, the neighbor awaked the vertex in this iteration

  public RefEdge(V vid, int lVID, E cost, RefEdgeType type) {
    this.vid = vid;
    this.edgeType = type;
    this.lVID = lVID;
    
    if (cost == null || cost instanceof NullWritable) {
      this.cost = null;
    } else {
      this.cost = cost;
    }
    awaked = false;
    active = false;
  }
  
  public void activate() {
    this.active = true;
  }
  
  public boolean isAwake() {
    return awaked;
  }
  
  public void commit() {
    this.awaked = this.active;
    this.active = false;
  }
  
  public void clearState(){
    awaked = false;
    active = false;
  }
  
  @Override
  public V getVertexID() {
    return vid;
  }

  @SuppressWarnings("unchecked")
  @Override
  public Vertex<V, E, M> getVertex() {
    return vertexSet[lVID];
  }
  
  @SuppressWarnings("unchecked")
  @Override
  public AbstractRefEdge<V, E, M> getPair() {
    if (this.edgeType == RefEdgeType.INEDGE){
      return (AbstractRefEdge<V, E, M>) vertexSet[lVID].getOutEdges().get(this.vEID);
    } else {
      return (AbstractRefEdge<V, E, M>) vertexSet[lVID].getInEdges().get(this.vEID);
    }
  }
  
  public E getValue() {
    return cost;
  }
  
  public int getlVID() {
    return lVID;
  }

  public void setlVID(int lVID) {
    this.lVID = lVID;
  }

  public int getvEID() {
    return vEID;
  }

  public void setvEID(int vEID) {
    this.vEID = vEID;
  }

  public E getCost() {
    return cost;
  }

  public boolean isInEdge() {
    return this.edgeType == RefEdgeType.INEDGE;
  }
  
  public boolean isOutEdge() {
    return this.edgeType == RefEdgeType.OUTEDGE;
  }
  
  @SuppressWarnings("unchecked")
  @Override
  public boolean equals(Object o) {
    if (this == o)
      return true;
    if (o == null)
      return false;
    if (this.getClass() == o.getClass()) {
      if (this.getVertexID() == ((RefEdge<V, E, M>)o).getVertexID() && this.getValue() == ((RefEdge<V, E, M>)o).getValue())
        return true;
      else {
        return false;
      }
    } else {
      return false;
    }
  }
  
  @Override
  public String toString() {
    String state = " awaked:" + awaked + " active:" + this.active;
    return (this.isInEdge() ? "[In-Edge]" : "[Out-Edge]") + this.getVertexID() + " lID: " + this.getlVID() + ":" + this.getValue() + state;
  }
}
