/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.bsp.AbstractRefEdge;
import org.apache.hama.bsp.AbstractVertex;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.LyraSyncBundle;
import org.apache.hama.bsp.LyraSyncMessage;
import org.apache.hama.bsp.Partitioner;
import org.apache.hama.graph.RefEdge.RefEdgeType;
import org.junit.Assert;

public abstract class Vertex<V extends Writable, E extends Writable, M extends Writable>
    extends AbstractVertex<V, E, M> implements Writable {
  public static final Log LOG = LogFactory.getLog(Vertex.class);
  
  GraphJobRunner<?, ?, ?> runner;
  
  private List<Edge<V, E>> edges;
  private Map<V, Integer> neighborsPeerIndexMap;
  @SuppressWarnings("rawtypes")
  private Map<Integer, List<RefEdge>> replicaFTInfo = new HashMap<Integer, List<RefEdge>>();
  private boolean selectiveActivate = false;
  private boolean danglingOPT = false;
  
  // TODO: double state in in/out edges
  // in-edges, only used by master vertex
  private List<RefEdge<V, E, M>> inEdges = new ArrayList<RefEdge<V, E, M>>();

  // used by master
  private int numberReplicas = 0;
  private int[] replicaIDs; // use peerIndex as array index, -1 means no replica in that peer;
                            // Note: master will set its localID at [masterIndex], which is used for Rebirth Recovery, but the numberReplicas would not increase for this value

  private int[] mirrorReplicaIndex = null; // use mirrorIndex as array index, not peerIndex!!!!. 
  private boolean[] peerToMirrorCache = null; // this data structure is derived from mirrorReplicaIndex.

  private int numOfMirror = 0;
  
  // used by replica
  private int masterIndex = -1;
  private int localID = -1;
  public boolean isMirrorReplica = false;
  

  public Configuration getConf() {
    return runner.getPeer().getConfiguration();
  }

  //@Override
  public void setup(Configuration conf) {
    Assert.assertTrue(isMaster());
  }
  
  /**
   * The user-defined function
   */
  public void compute(Iterator<M> messages) throws IOException {}
  
  /**
   * @return the destination peer name of the destination of the given directed
   *         edge.
   */
  public String getDestinationPeerName(Edge<V, E> edge) {
    return getDestinationPeerName(edge.getDestinationVertexID());
  }

  /**
   * @return the destination peer name of the given vertex id, determined by the
   *         partitioner.
   */
  public String getDestinationPeerName(V vertexId) {
    return runner.getPeer().getPeerName(
        getPartitioner().getPartition(vertexId, value,
            runner.getPeer().getNumPeers()));
  }

  //@Override
  public void sendMessageToNeighbors(M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public void sendMessage(V destinationVertexID, M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public void sendMessage(Edge<V, E> e, M msg) throws IOException {
    throw new UnsupportedOperationException("sendMessage unsupported now!");
  }

  //@Override
  public long getSuperstepCount() {
    return runner.getNumberIterations();
  }

  // edges
  public void setEdges(List<Edge<V, E>> list) {
    this.edges = list;
  }

  public void addEdge(Edge<V, E> edge) {
    if (edges == null)
      this.edges = new ArrayList<Edge<V, E>>();
    this.edges.add(edge);
  }

  //@Override
  public List<Edge<V, E>> getEdges() {
    return edges;
  }
  
  public void clearEdges() {
    this.edges.clear();
    this.edges = null;
  }
  
  @SuppressWarnings("rawtypes")
  public void addReplicaFTInfo(int peerIndex, List<RefEdge> outEdge){
    Assert.assertNull(replicaFTInfo.put(peerIndex, outEdge));
  }
  
  @SuppressWarnings({ "unchecked", "rawtypes" })
  public void addSelfFTInfo(){
    // a ugly way to pass the type check
    List<RefEdge> tmp =  (List<RefEdge>) ((Object) this.outEdges);
    replicaFTInfo.put(this.masterIndex, tmp);
    // add my own local ID and not add the replicaNum
    if (replicaIDs != null)
      replicaIDs[this.masterIndex] = this.localID;
  }
  
  @SuppressWarnings("rawtypes")
  public Map<Integer, List<RefEdge>> getReplicaFTInfo() {
    return replicaFTInfo;
  }

  @SuppressWarnings("rawtypes")
  public void setReplicaFTInfo(Map<Integer, List<RefEdge>> replicaFTInfo) {
    this.replicaFTInfo = replicaFTInfo;
  }

  public void addNeighborMap(V vertexID, int peerIndex) {
    neighborsPeerIndexMap.put(vertexID, new Integer(peerIndex));
  }
  
  public int getNeighorPeerIndex(V vertexID) {
    return neighborsPeerIndexMap.get(vertexID).intValue();
  }
  
  public void enableSelectiveActivate() {
    selectiveActivate = true;
    neighborsPeerIndexMap = new HashMap<V, Integer>();
  }
  
  public void enableDanglingOpt(){
    danglingOPT = true;
  }

  //@Override
  public M getValue() { // only used by APP
    return value;
  }

  //@Override
  public void setValue(M value) { // only used by APP
    this.nValue = value;
  }
  
  //@Override
  public M getMessage() { // only used by APP
    return this.msg;
  }

  public int getMaxIteration() {
    return runner.getMaxIteration();
  }

  /**
   * Get the last aggregated value of the defined aggregator, null if nothing
   * was configured or not returned a result. You have to supply an index, the
   * index is defined by the order you set the aggregator classes in
   * {@link GraphJob#setAggregatorClass(Class...)}. Index is starting at zero,
   * so if you have a single aggregator you can retrieve it via
   * {@link #getLastAggregatedValue}(0).
   */
  @SuppressWarnings("unchecked")
  public M getLastAggregatedValue(int index) {
    return (M) runner.getLastAggregatedValue(index);
  }

  /**
   * Get the number of aggregated vertices in the last superstep. Or null if no
   * aggregator is available.You have to supply an index, the index is defined
   * by the order you set the aggregator classes in
   * {@link GraphJob#setAggregatorClass(Class...)}. Index is starting at zero,
   * so if you have a single aggregator you can retrieve it via
   * {@link #getNumLastAggregatedVertices}(0).
   */
  public IntWritable getNumLastAggregatedVertices(int index) {
    return runner.getNumLastAggregatedVertices(index);
  }

  public int getNumPeers() {
    return runner.getPeer().getNumPeers();
  }

  /**
   * Gives access to the BSP primitives and additional features by a peer.
   */
  public BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> getPeer() {
    return runner.getPeer();
  }

  /**
   * @return the configured partitioner instance to message vertices.
   */
  @SuppressWarnings("unchecked")
  public Partitioner<V, M> getPartitioner() {
    return (Partitioner<V, M>) runner.getPartitioner();
  }

  //@Override
  public long getNumVertices() {
    return runner.getNumberVertices();
  }

  //@Override
  public void voteToHalt() {
    nVotedToHalt = true;
  }


  public boolean isActive() {
    return (!votedToHalt) || (awaked);
  }
  
  public boolean isConverged() {
    return !nActivate;
  }

  /**
   * (BSP-g) default implementation
   */
  //@Override
  public boolean merge(Vertex<V, E, M> vertex) {
    Assert.assertEquals("can't merge vertex:" + this.getVertexID() 
        + " with vertex:" + vertex.getVertexID(), this, vertex);
    for (Edge<V, E> e: vertex.getEdges())
        this.addEdge(e);
    return true;
  }

  /**
   * (BSP-g) add peer index with the replica of vertex
   */
  
  public void addReplicaLocalID(int peerIndex, int localID) {
    if (this.numberReplicas == 0 && this.replicaIDs == null) { // init replica set
      this.replicaIDs = new int[getPeer().getNumPeers()];
      for (int i = 0; i < this.replicaIDs.length; i++)
        this.replicaIDs[i] = -1;
    }

    int oldLocalID = this.replicaIDs[peerIndex];
    this.replicaIDs[peerIndex] = localID;
    if (oldLocalID == -1) 
      this.numberReplicas++;
  }
  
  public int getNumOfLocalIDs() {
    return this.numberReplicas;
  }

  public int[] getLocalIDList() {
    return this.replicaIDs;
  }

  public void setLocalIDList(int num, int[] theList) {
    this.numberReplicas = num;
    this.replicaIDs = theList;
  }

  public void addMirrorIndex(int ind) {
    if (this.numOfMirror == 0) {
      this.mirrorReplicaIndex = new int[getPeer().getNumPeers()];
      for (int i = 0; i < this.mirrorReplicaIndex.length; i++)
	this.mirrorReplicaIndex[i] = -1;
    }
    
    this.mirrorReplicaIndex[this.numOfMirror] = ind;
    this.numOfMirror++;
  }

  public int getNumOfMirror() {
    return this.numOfMirror;
  }

  public int getMirrorIndex(int idx) {
    if (this.numOfMirror == 0)
      return -1;
    else
      return this.mirrorReplicaIndex[idx];
  }
  
  public int getReplicaLocalIDAtIndex(int peerIndex) {
    if (this.numberReplicas == 0)
      return -1;
    return this.replicaIDs[peerIndex];
  }

  
  public void setMasterPeerIndex(int peerIndex) {
    this.masterIndex = peerIndex;
  }

  public int getMasterPeerIndex() {
    return this.masterIndex;
  }
  
  public void setLocalID(int id) {
    this.localID = id;
  }

  public int getLocalID() {
    return this.localID;
  }

  public boolean vertexHasBeenAwaked() {
    return nAwaked;
  }

  
  public boolean judgeMirrorHost(int peer){
    boolean isMirrorHost = false;
    for (int j = 0; j < this.numOfMirror; j++){
      if (this.mirrorReplicaIndex[j] == peer){
        isMirrorHost = true;
        break;
      }
    }
    return isMirrorHost;
  }
  
  /**
   * if there is no peerToMirrorCache entry, it will create a cache. So compared to the judgeMirrorHost, this function has side effect
   * @param peer
   * @return
   */
  private boolean isMirrorHost(int peer){
    if (this.peerToMirrorCache == null){
      int num = 0;
      peerToMirrorCache = new boolean[getPeer().getNumPeers()];
      for (int i = 0; i < peerToMirrorCache.length; i++){
        peerToMirrorCache[i] = false;
        for (int j = 0; j < this.numOfMirror; j++){
          if (this.mirrorReplicaIndex[j] == i){
            peerToMirrorCache[i] = true;
            num++;
          }
        }
      }
      Assert.assertEquals(num, this.numOfMirror);
    }
    return peerToMirrorCache[peer];
    
  }
  
  public void setNextState(LyraSyncMessage<V, M> msg) {
    nVotedToHalt = msg.votedToHalt;
    nAwaked = msg.awaked;
    nValue = msg.value;
    nMsg = msg.msg;
    nActivate = msg.activated;
    nActivateAllNeighbors = msg.activateAllNeighbors;
    if (msg.activateLists != null){
      nActivateLists = msg.activateLists;
    } else if (msg.activateList != null){
      nActivateLists = this.getnActivateLists(getNumPeers());
      nActivateLists[this.getPeer().getPeerIndex()] = msg.activateList;
    }
  }
  
  
  public long remoteSync(LyraSyncBundle[] bundles) {
    /*
     * For Replica-base FT, All Mirror Replicas will have the same state with master
     */
    if (this.danglingOPT && this.getEdgeNum()== 0)
      return 0;
    if (numberReplicas > 0) {
      for (int i=0; i < this.replicaIDs.length; i++) {
        if (this.replicaIDs[i] != -1 && i != this.masterIndex) {
          if (isMirrorHost(i)){
            bundles[i].addMessage(new LyraSyncMessage<V,M>(new IntWritable(this.replicaIDs[i]),this, -1));
          } else {
            bundles[i].addMessage(new LyraSyncMessage<V,M>(new IntWritable(this.replicaIDs[i]),this, i));
          }
        }
      }
      return numberReplicas;
    }
    return 0;
  }
  
  
  /*
   * change global state based on nextState  
   */
  //@Override
  public void commitState() {
    votedToHalt = nVotedToHalt;
    // nVotedToHalt = true;
    // value = nValue;
    // msg = nMsg;
    activate = nActivate;
    activateAllNeighbors = nActivateAllNeighbors;
    activateLists = nActivateLists;

    // reset the next state
    // nVotedToHalt = false;
    nActivate = false;
    nActivateAllNeighbors = false;
    nActivateLists = null;

    value = nValue;
    msg = nMsg;
     
    awaked = nAwaked;
    
    nAwaked = false;

    if (awaked && this.vertexType == VertexType.MASTER){
      for (RefEdge<V, E, M> e : this.inEdges){
        e.commit();
      }
    }
    return;
  }
  
  public void clearNewState(){
    nAwaked = false;

    nValue = value;
    nMsg = msg;
 
    nActivate = false;
    nActivateAllNeighbors = false;
    nActivateLists = null;
  }
  
  /**
   * (BSP-g) sync vertex with replicas
   * since called after commitState, the state is the newest state
   * @return #local-message
   */
  public long localActivateVertices() throws IOException {
    if (this.nActivate) { // sync with activate
      if (this.nActivateAllNeighbors) {
        return this.awakeNeighbors();
      } else {
        if (!selectiveActivate)
          throw new UnsupportedOperationException("forget to enable 'hama.graph.selective.activate'");
        return this.awake(this.getnActivateLists(getNumPeers())[getPeer().getPeerIndex()]);
      }
    }
    return 0;
  }
    
  //@Override
  public void activate(Edge<V, E> e, M msg) throws IOException {
    activate(e.getDestinationVertexID(), msg);
  }
  
  //@Override
  public void activate(V destinationVertexID, M msg) throws IOException {
    if (!selectiveActivate)
      throw new UnsupportedOperationException("forget to enable 'hama.graph.selective.activate'");
    
    this.nActivate = true;
    int peerIndex = getNeighorPeerIndex(destinationVertexID);
    
    
    this.getnActivateLists(getNumPeers())[peerIndex].add(destinationVertexID);
    //this.activateLists[peerIndex].add(destinationVertexID);
    
    this.nMsg = msg;
  }
  
  //@Override
  public void activateNeighbors(M msg) throws IOException {
    this.nActivate = true;
    this.nActivateAllNeighbors = true;
    
    this.nMsg = msg;
  }
  
  /**
   * (BSP-g) set in-edges
   */
  public void setInEdges(List<RefEdge<V, E, M>> list) {
    this.inEdges = list;
  }

  /**
   * (BSP-g) add an in-edge
   */
  public void addInEdge(RefEdge<V, E, M> edge) {
    this.inEdges.add(edge);
  }
  
  /**
   * (BSP-g) get the in-edges
   * @return a list of in-edges of this vertex
   */
  public List<RefEdge<V, E, M>> getInEdges() {
    return this.inEdges;
  }
  
  private int edgeNumCache = -1;
  public int getEdgeNum(){
    if (edgeNumCache == -1){
      if (edges != null){
        edgeNumCache = edges.size();
      } else {
        int num = 0;
        if (replicaFTInfo.size() != 0){
          for (@SuppressWarnings("rawtypes") List<RefEdge> edgeOnNode : replicaFTInfo.values()){
            num += edgeOnNode.size();
          }
        }
        edgeNumCache = num;
      }
    }
    return edgeNumCache;
  }

  
  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((vertexID == null) ? 0 : vertexID.hashCode());
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Vertex<?, ?, ?> other = (Vertex<?, ?, ?>) obj;
    if (vertexID == null) {
      if (other.vertexID != null)
        return false;
    } else if (!vertexID.equals(other.vertexID))
      return false;
    return true;
  }

  protected abstract V getNewInstanceOfV();

  public void rollBack() {
    this.nValue = this.value;
    this.nMsg = this.msg;
    
    this.nActivate = this.activate;
    this.nActivateAllNeighbors = this.activateAllNeighbors;
    this.nAwaked = this.awaked;
    this.nVotedToHalt = this.votedToHalt;
    
    this.nActivateLists = this.activateLists;
  }
  
  
    
  public void write(DataOutput out) throws IOException {
    this.vertexID.write(out);
    if (this.nValue != null) {
      out.writeBoolean(true);
      this.nValue.write(out);
    } else {
      out.writeBoolean(false);
    }
    if (this.value != null) {
      out.writeBoolean(true);
      this.value.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    if (this.nMsg != null) {
      out.writeBoolean(true);
      this.nMsg.write(out);
    } else {
      out.writeBoolean(false);
    }
    if (this.msg != null) {
      out.writeBoolean(true);
      this.msg.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    out.writeBoolean(this.activate);
    out.writeBoolean(this.activateAllNeighbors);
    out.writeBoolean(this.awaked);
    out.writeBoolean(this.votedToHalt);
    
    out.writeBoolean(this.nActivate);
    out.writeBoolean(this.nActivateAllNeighbors);
    out.writeBoolean(this.nAwaked);
    out.writeBoolean(this.nVotedToHalt);
    
    out.writeInt(localID);

    out.writeBoolean(this.selectiveActivate);
    if (selectiveActivate) {
      if (activateLists == null) {
	out.writeInt(-1);
      } else {
	out.writeInt(activateLists.length);
	for (int i = 0; i < activateLists.length; i++) {
	  if (activateLists[i] == null)
	    out.writeInt(-1);
	  else {
	    out.writeInt(activateLists[i].size());
	    for (int j = 0; j < activateLists[i].size(); i++) {
	      activateLists[i].get(j).write(out);
	    }
	  }
	}
      }
      if (nActivateLists == null)
	out.writeInt(-1);
      else {
	out.writeInt(nActivateLists.length);
	for (int i = 0; i < nActivateLists.length; i++) {
	  if (nActivateLists[i] == null)
	    out.writeInt(-1);
	  else {
	    out.writeInt(nActivateLists[i].size());
	    for (int j = 0; j < nActivateLists[i].size(); j++) {
	      nActivateLists[i].get(j).write(out);
	    }
	  }
	}
      }
    }
  }

  public void writeNewStates(DataOutput out) throws IOException {
    this.vertexID.write(out);
    if (this.nValue != null) {
      out.writeBoolean(true);
      this.nValue.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    if (this.nMsg != null) {
      out.writeBoolean(true);
      this.nMsg.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    out.writeBoolean(this.nActivate);
    out.writeBoolean(this.nActivateAllNeighbors);
    out.writeBoolean(this.nAwaked);
    out.writeBoolean(this.nVotedToHalt);
    
    out.writeInt(localID);

    out.writeBoolean(this.selectiveActivate);
    if (selectiveActivate) {
      if (nActivateLists == null)
	out.writeInt(-1);
      else {
	out.writeInt(nActivateLists.length);
	for (int i = 0; i < nActivateLists.length; i++) {
	  if (nActivateLists[i] == null)
	    out.writeInt(-1);
	  else {
	    out.writeInt(nActivateLists[i].size());
	    for (int j = 0; j < nActivateLists[i].size(); j++) {
	      nActivateLists[i].get(j).write(out);
	    }
	  }
	}
      }
    }
  }

  public void writeOldStates(DataOutput out) throws IOException {
    this.vertexID.write(out);

    if (this.value != null) {
      out.writeBoolean(true);
      this.value.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    if (this.msg != null) {
      out.writeBoolean(true);
      this.msg.write(out);
    } else {
      out.writeBoolean(false);
    }
    
    out.writeBoolean(this.activate);
    out.writeBoolean(this.activateAllNeighbors);
    out.writeBoolean(this.awaked);
    out.writeBoolean(this.votedToHalt);
    
    out.writeInt(localID);

    out.writeBoolean(this.selectiveActivate);
    if (selectiveActivate) {
      if (activateLists == null) {
	out.writeInt(-1);
      } else {
	out.writeInt(activateLists.length);
	for (int i = 0; i < activateLists.length; i++) {
	  if (activateLists[i] == null)
	    out.writeInt(-1);
	  else {
	    out.writeInt(activateLists[i].size());
	    for (int j = 0; j < activateLists[i].size(); j++) {
	      activateLists[i].get(j).write(out);
	    }
	  }
	}
      }
    }
  }
  
  @SuppressWarnings("unchecked")
  public void readFields(DataInput in) throws IOException {
    this.vertexID.readFields(in);
    
    if (this.nValue == null)
      this.nValue = this.getNewInstanceOfM();
    if (this.value == null)
      this.value = this.getNewInstanceOfM();
    
    if (this.nMsg == null)
      this.nMsg = this.getNewInstanceOfM();
    if (this.msg == null)
      this.msg = this.getNewInstanceOfM();
    
    if (in.readBoolean())
      this.nValue.readFields(in);
    if (in.readBoolean())
      this.value.readFields(in);
    
    if (in.readBoolean())
      this.nMsg.readFields(in);
    if (in.readBoolean())
      this.msg.readFields(in);
    
    this.activate = in.readBoolean();
    this.activateAllNeighbors = in.readBoolean();
    this.awaked = in.readBoolean();
    this.votedToHalt = in.readBoolean();
    
    this.nActivate = in.readBoolean();
    this.nActivateAllNeighbors = in.readBoolean();
    this.nAwaked = in.readBoolean();
    this.nVotedToHalt = in.readBoolean();
    
    localID = in.readInt();

    this.selectiveActivate = in.readBoolean();
    if (selectiveActivate) {
      int listLen = in.readInt();
      if (listLen == -1) {
	activateLists = null;
      } else {
	activateLists = new List[listLen];
	for (int i = 0; i < listLen; i++) {
	  int len = in.readInt();
	  if (len == -1) 
	    activateLists[i] = null;
	  else {
	    activateLists[i] = new ArrayList<V>();
	    for (int j = 0; j < len; j++) {
	      //TODO: add new element here
	      activateLists[i].add(0, this.getNewInstanceOfV());
	      activateLists[i].get(0).readFields(in);
	    }
	  }
	}
      }

      listLen = in.readInt();
      if (listLen == -1) {
	nActivateLists = null;
      } else {
	nActivateLists = new List[listLen];
	for (int i = 0; i < listLen; i++) {
	  int len = in.readInt();
	  if (len == -1) 
	    nActivateLists[i] = null;
	  else {
	    nActivateLists[i] = new ArrayList<V>();
	    for (int j = 0; j < len; j++) {
	      //TODO: add new element here
	      nActivateLists[i].add(0, this.getNewInstanceOfV());
	      nActivateLists[i].get(0).readFields(in);
	    }
	  }
	}
      }
    }
  }

  @SuppressWarnings("unchecked")
  public void readNewStates(DataInput in) throws IOException {
    this.vertexID.readFields(in);
    
    if (this.nValue == null)
      this.nValue = this.getNewInstanceOfM();

    
    if (this.nMsg == null)
      this.nMsg = this.getNewInstanceOfM();

    
    if (in.readBoolean())
      this.nValue.readFields(in);
    
    if (in.readBoolean())
      this.nMsg.readFields(in);
    
    
    this.nActivate = in.readBoolean();
    this.nActivateAllNeighbors = in.readBoolean();
    this.nAwaked = in.readBoolean();
    this.nVotedToHalt = in.readBoolean();
    
    localID = in.readInt();

    this.selectiveActivate = in.readBoolean();
    if (selectiveActivate) {
      int listLen = in.readInt();
      if (listLen == -1) {
	nActivateLists = null;
      } else {
	nActivateLists = new List[listLen];
	for (int i = 0; i < listLen; i++) {
	  int len = in.readInt();
	  if (len == -1) 
	    nActivateLists[i] = null;
	  else {
	    nActivateLists[i] = new ArrayList<V>();
	    for (int j = 0; j < len; j++) {
	      //TODO: add new element here
	      nActivateLists[i].add(0, this.getNewInstanceOfV());
	      nActivateLists[i].get(0).readFields(in);
	    }
	  }
	}
      }
    }
  }

  @SuppressWarnings("unchecked")
  public void readOldStates(DataInput in) throws IOException {
    this.vertexID.readFields(in);
    
    if (this.value == null)
      this.value = this.getNewInstanceOfM();
    
    if (this.msg == null)
      this.msg = this.getNewInstanceOfM();
    
    if (in.readBoolean())
      this.value.readFields(in);
    
    if (in.readBoolean())
      this.msg.readFields(in);
    
    this.activate = in.readBoolean();
    this.activateAllNeighbors = in.readBoolean();
    this.awaked = in.readBoolean();
    this.votedToHalt = in.readBoolean();
    
    localID = in.readInt();

    this.selectiveActivate = in.readBoolean();
    if (selectiveActivate) {
      int listLen = in.readInt();
      if (listLen == -1) {
	activateLists = null;
      } else {
	activateLists = new List[listLen];
	for (int i = 0; i < listLen; i++) {
	  int len = in.readInt();
	  if (len == -1) 
	    activateLists[i] = null;
	  else {
	    activateLists[i] = new ArrayList<V>();
	    for (int j = 0; j < len; j++) {
	      //TODO: add new element here
	      activateLists[i].add(0, this.getNewInstanceOfV());
	      activateLists[i].get(0).readFields(in);
	    }
	  }
	}
      }
    }
  }


  
  public int[] getMirrorIndices() {
    return this.mirrorReplicaIndex;
  }
  
  public void setMirrorIndices(int[] mirrorIndices, int numOfMirror) {
    this.numOfMirror = numOfMirror;
    this.mirrorReplicaIndex = mirrorIndices;
  }

  public boolean stateEqual(Vertex<V, E, M> vertex) throws IOException {
    boolean result = true;
    if (vertex.nActivate != this.nActivate) {
      LOG.debug(this.getVertexID()+"nActivated wrong");
      result =  false;
    }
    if (vertex.nAwaked != this.nAwaked) {
      LOG.debug(this.getVertexID()+"nAwaked wrong");
      result = false;
    }
    if (vertex.nVotedToHalt != this.nVotedToHalt) {
      LOG.debug(this.getVertexID()+"nVotedToHalt wrong");
      result = false;
    }
    if (vertex.nActivateAllNeighbors != this.nActivateAllNeighbors) {
      LOG.debug(this.getVertexID()+"nActNeighbors wrong");
      result = false;
    }
    return result;
  }

  /*===================================================================== the below code is used for recovery, since we want to optimised the recovery, we break the abstract (-_-!)=====================================================*/
  public static Class<? extends Writable> V_CLASS;
  public static Class<? extends Writable> E_CLASS;
  public static Class<? extends Writable> M_CLASS;
  
  
  public void writeRecoverMasterMsg(DataOutput out) throws IOException{
    this.writeCommonStatesForRecover(out);
    this.writeActivateListsForRecover(out, true, -1);

    this.writeLocalIDs(out);

    this.writeMirrorIndices(out);

    writeMapHalper(out, this.getReplicaFTInfo());
    writeRefEdgeListHalper(out, this.getInEdges());
  }
  
  @SuppressWarnings("unchecked")
  public void readRecoverMasterMsg(DataInput in)  throws IOException{
    this.readCommonStatesForRecover(in);
    this.readActivateListsForRecover(in, true, -1);

    this.readLocalIDs(in);
    this.setLocalID(this.getLocalIDList()[this.masterIndex]);
    
    this.readMirrorIndices(in);
     
    readMapHelper(in, this.getReplicaFTInfo());
    List<AbstractRefEdge<V, E, M>> outEdge = (List<AbstractRefEdge<V, E, M>>) ((Object) this.getReplicaFTInfo().get(this.masterIndex));
    if (outEdge != null)
      this.setOutEdges(outEdge);
    
    this.setSpecialInEdges(readRefEdgeListHelp(in, RefEdgeType.INEDGE));

    this.setVertexType(VertexType.MASTER);
  }
  
  public void writeRecoverReplicaMsg(DataOutput out, int peerIndex, boolean isMirror) throws IOException{
    this.isMirrorReplica = isMirror;
    this.writeCommonStatesForRecover(out);
    this.writeActivateListsForRecover(out, isMirror, peerIndex);
    if (isMirror){
      this.writeLocalIDs(out);
      this.writeMirrorIndices(out);
      writeMapHalper(out, this.getReplicaFTInfo());
      writeRefEdgeListHalper(out, this.getInEdges());
    } else {
      out.writeInt(this.getReplicaLocalIDAtIndex(peerIndex));
      @SuppressWarnings("rawtypes")
      List<RefEdge> refEdges = this.getReplicaFTInfo().get(peerIndex);
      writeRefEdgeListHalper(out, refEdges);
    }
  }
  
  @SuppressWarnings("unchecked")
  public void readRecoverReplicaMsg(DataInput in,  int peerIndex, boolean isMirror) throws IOException{
    this.readCommonStatesForRecover(in);
    this.readActivateListsForRecover(in, isMirror, peerIndex);
    if (isMirror){
      this.setVertexType(VertexType.MIRROR);
      this.readLocalIDs(in);
      this.setLocalID(this.getReplicaLocalIDAtIndex(peerIndex));
      
      this.readMirrorIndices(in);
      
      readMapHelper(in, this.getReplicaFTInfo());
      List<AbstractRefEdge<V, E, M>> refEdges = (List<AbstractRefEdge<V, E, M>>) ((Object) this.getReplicaFTInfo().get(peerIndex));
      if (refEdges != null)
        this.setOutEdges(refEdges);
      
      this.setSpecialInEdges(readRefEdgeListHelp(in, RefEdgeType.INEDGE));
    } else {
      this.setVertexType(VertexType.REPLICA);
      this.setLocalID(in.readInt());
      List<AbstractRefEdge<V, E, M>> refEdges = (List<AbstractRefEdge<V, E, M>>) (Object)readRefEdgeListHelp(in, RefEdgeType.OUTEDGE);
      if (refEdges != null)
        this.setOutEdges(refEdges);
    }

  }
  
  public void writeMirrorInfo(DataOutput out) throws IOException
  {
    out.writeInt(this.masterIndex);
    this.writeLocalIDs(out);
    this.writeMirrorIndices(out);
    writeMapHalper(out, this.getReplicaFTInfo());
    writeRefEdgeListHalper(out, this.getInEdges());
  }
  
  public void readMirrorInfo(DataInput in) throws IOException
  {
    this.setVertexType(VertexType.MIRROR);
    this.isMirrorReplica = true;
    
    this.masterIndex = in.readInt();
    this.readLocalIDs(in);
    this.readMirrorIndices(in);
    this.replicaFTInfo.clear();
    readMapHelper(in, this.replicaFTInfo);
    this.setSpecialInEdges(readRefEdgeListHelp(in, RefEdgeType.INEDGE));
  }
  
  /* ------------------------subroutine for recover----------------------------------------- */
  
  /**
   * We will always write old state, since that state is consistent
   * @param out
   * @throws IOException
   */
  public void writeCommonStatesForRecover(DataOutput out) throws IOException {
    this.vertexID.write(out);
    out.writeInt(this.masterIndex);

    byte bitMap = 0;
    if (this.activate){
      bitMap |= 1;
    }
    if (this.activateAllNeighbors){
      bitMap |= 2;
    }
    if (this.awaked){
      bitMap |= 4;
    }
    if (this.votedToHalt){
      bitMap |= 8;
    }
    
    if (this.value != null){
      bitMap |= 16;
    }
    if (this.msg != null){
      bitMap |= 32;
    }
    out.writeByte(bitMap);
    
    if (this.value != null) {
      this.value.write(out);
    }
    if (this.msg != null) {
      this.msg.write(out);
    }    

  }
  
  /**
   * We will always read state to new state, since we will redo something.
   * @param in
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  public void readCommonStatesForRecover(DataInput in) throws IOException {
    if (this.vertexID == null){
      this.vertexID = (V) ReflectionUtils.newInstance(V_CLASS, null);
    }
    this.vertexID.readFields(in);
    this.masterIndex = in.readInt();
    
    
    byte bitMap = in.readByte();
    
    if ((bitMap & 1) == 1){
      this.nActivate = true;
    } else {
      this.nActivate = false;
    }
    if ((bitMap & 2) == 2){
      this.nActivateAllNeighbors = true;
    } else {
      this.nActivateAllNeighbors = false;
    }
    if ((bitMap & 4) == 4){
      this.nAwaked = true;
    } else {
      this.nAwaked = false;
    }
    if ((bitMap & 8) == 8){
      this.nVotedToHalt = true;
    } else {
      this.nVotedToHalt = false;
    }
    
    if (this.nValue == null)
      this.nValue = this.getNewInstanceOfM(); 
    if (this.nMsg == null)
      this.nMsg = this.getNewInstanceOfM();
    
    if ((bitMap & 16) == 16){
      this.nValue.readFields(in);
    }
    
    if ((bitMap & 32) == 32){
      this.nMsg.readFields(in);
    }
  }
  
  
  /**
   * 
   * @param out
   * @param isMirror
   * @param targetpeer if isMirror == true, targetPeer is dummy, and we will write the full activateLists for master and mirror
   * @throws IOException
   */
  public void writeActivateListsForRecover(DataOutput out, boolean isMirror, int targetPeer) throws IOException {
    if (activateLists == null) {
      out.writeInt(-1);
    } else {
      out.writeInt(activateLists.length);
      if (isMirror){
        for (int i = 0; i < activateLists.length; i++) {
          if (activateLists[i] == null)
            out.writeInt(-1);
          else {
            out.writeInt(activateLists[i].size());
            for (int j = 0; j < activateLists[i].size(); j++) {
              activateLists[i].get(j).write(out);
            }
          }
        }
      } else {
        if (activateLists[targetPeer] == null)
          out.writeInt(-1);
        else {
          out.writeInt(activateLists[targetPeer].size());
          for (int j = 0; j < activateLists[targetPeer].size(); j++) {
            activateLists[targetPeer].get(j).write(out);
          }
        }
      }
    }
    
  }
  
  /**
   * 
   * @param out
   * @param isMirror
   * @param targetpeer if isMirror == true, targetPeer is dummy, and we will write the full activateLists for master and mirror
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  public void readActivateListsForRecover(DataInput in, boolean isMirror, int target) throws IOException {
    int listLen = in.readInt();
    if (listLen == -1) {
      nActivateLists = null;
    } else {
      nActivateLists = new List[listLen];
      if (isMirror == true){
        for (int i = 0; i < listLen; i++) {
          int len = in.readInt();
          if (len == -1){
            nActivateLists[i] = null;
          } else {
            nActivateLists[i] = new ArrayList<V>();
            for (int j = 0; j < len; j++) {
              // TODO: add new element here
              nActivateLists[i].add(0, this.getNewInstanceOfV());
              nActivateLists[i].get(0).readFields(in);
            }
          }
        }
      } else {
        int len = in.readInt();
        if (len == -1){
          nActivateLists[target] = null;
        } else {
          nActivateLists[target] = new ArrayList<V>();
          for (int j = 0; j < len; j++) {
            // TODO: add new element here
            nActivateLists[target].add(0, this.getNewInstanceOfV());
            nActivateLists[target].get(0).readFields(in);
          }
        }
      }
    }
  }
  
  
  public void writeLocalIDs(DataOutput out) throws IOException {
    if (this.replicaIDs == null)
      out.writeBoolean(false);
    else {
      out.writeBoolean(true);
      out.writeInt(this.numberReplicas);
      out.writeInt(replicaIDs.length);
      for (int i = 0; i < replicaIDs.length; i++) {
        out.writeInt(this.replicaIDs[i]);
      }
    }
  }

  public void readLocalIDs(DataInput in) throws IOException {
    if (!in.readBoolean())
      this.replicaIDs = null;
    else {
      this.numberReplicas = in.readInt();
      int len = in.readInt();
      this.replicaIDs = new int[len];
      for (int i = 0; i < len; i++) {
        this.replicaIDs[i] = in.readInt();
      }
    }
  }
  
  public void writeMirrorIndices(DataOutput out) throws IOException {
    if (this.mirrorReplicaIndex == null)
      out.writeBoolean(false);
    else {
      out.writeBoolean(true);
      out.writeInt(this.numOfMirror);
      out.writeInt(this.mirrorReplicaIndex.length);
      for (int i = 0; i < mirrorReplicaIndex.length; i++) {
        out.writeInt(mirrorReplicaIndex[i]);
      }
    }
  }
  
  public void readMirrorIndices(DataInput in) throws IOException {
    if (!in.readBoolean())
      this.mirrorReplicaIndex = null;
    else {
      this.numOfMirror = in.readInt();
      int len = in.readInt();
      this.mirrorReplicaIndex = new int[len];
      for (int i = 0; i < len; i++) {
        mirrorReplicaIndex[i] = in.readInt();
      }
    }
  }
  
  @SuppressWarnings({ "rawtypes", "unchecked" })
  private List<RefEdge> readRefEdgeListHelp(DataInput in, RefEdge.RefEdgeType expectType) throws IOException{
    if (in.readBoolean()){
      int size = in.readInt();
      List<RefEdge> result = new ArrayList<RefEdge>(size);
      for (int i = 0; i < size; i++){
        Writable edgeVertexID = ReflectionUtils.newInstance(V_CLASS, null);
        edgeVertexID.readFields(in);
        int lVID = in.readInt();
        int lVEID = in.readInt();

        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(E_CLASS, null);
          edgeValue.readFields(in);
        }
        RefEdge tmp = new RefEdge(edgeVertexID, lVID, edgeValue, expectType);
        tmp.setvEID(lVEID);
        result.add(tmp);
      }
      return result;
    } else {
      return null;
    }
  }
  
  @SuppressWarnings("rawtypes")
  private void writeRefEdgeListHalper(DataOutput out, List<?> store) throws IOException{
    if (store == null){
      out.writeBoolean(false);
    } else {
      out.writeBoolean(true);
      out.writeInt(store.size());
      for (int i = 0; i < store.size(); i++){
        RefEdge edge = (RefEdge) store.get(i);
        edge.getVertexID().write(out);
        out.writeInt(edge.getlVID());
        out.writeInt(edge.getvEID());
        if (edge.getCost() != null) {
          out.writeBoolean(true);
          edge.getCost().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    }
  }
  
  @SuppressWarnings({ "rawtypes", "unchecked" })
  private void readMapHelper(DataInput in, Map<Integer, List<RefEdge>> buffer) throws IOException{
    int size = in.readInt();
    for (int j = 0; j < size; j++){
      int key = in.readInt();
      int lSize = in.readInt();      
      List<RefEdge> refEdges = new ArrayList<RefEdge>(lSize);
      for (int i = 0; i < lSize; i++){
        Writable edgeVertexID = ReflectionUtils.newInstance(V_CLASS, null);
        edgeVertexID.readFields(in);
        int lVID = in.readInt();
        int lVEID = in.readInt();

        Writable edgeValue = null;
        if (in.readBoolean()) {
          edgeValue = ReflectionUtils.newInstance(E_CLASS, null);
          edgeValue.readFields(in);
        }
        RefEdge tmp = new RefEdge(edgeVertexID, lVID, edgeValue, RefEdgeType.OUTEDGE);
        tmp.setvEID(lVEID);
        refEdges.add(tmp);
      }
      buffer.put(key, refEdges);
    }
  }
  
  @SuppressWarnings("rawtypes")
  private void writeMapHalper(DataOutput out, Map<Integer, List<RefEdge>> store) throws IOException{
    out.writeInt(store.size());
    for (Entry<Integer, List<RefEdge>> entry : store.entrySet()){
      out.writeInt(entry.getKey().intValue());
      List<RefEdge> outEdges = (List<RefEdge>) entry.getValue();
      out.writeInt(outEdges.size());
      for (int i = 0; i < outEdges.size(); i++){
        RefEdge<?, ?, ?> edge = (RefEdge<?, ?, ?>) outEdges.get(i);
        edge.getVertexID().write(out);
        out.writeInt(edge.getlVID());
        out.writeInt(edge.getvEID());
        if (edge.getCost() != null) {
          out.writeBoolean(true);
          edge.getCost().write(out);
        } else {
          out.writeBoolean(false);
        }
      }
    }
  }
  
  public void awakedByNeighbors() {
    if (selectiveActivate) {
      for (RefEdge<V, E, M> edge : this.getInEdges()) {
        if (edge.isAwake()) {
          this.awaked = true;
        }
      }
    } else {
      for (RefEdge<V, E, M> edge : this.getInEdges()) {
        Vertex<V, E, M> neighbor = edge.getVertex();
        if (neighbor.activate) {
          this.awaked = true;
          edge.activate();
          edge.commit();
        }
      }
    }
  }
  
  
  @SuppressWarnings("unchecked")
  public List<V>[] getActivateLists(int n) {
    if (activateLists == null){
      activateLists = new List[n];
      for (int i = 0; i < n; i++)
        activateLists[i] = new ArrayList<V>();
    }
    return activateLists;
  }

  /**
   * after this method, location info about crashed mirrors and replicas will be cleared
   * @param crashedPeerIndices
   * @param myHost
   */
  public void upgradeToMaster(List<Integer> crashedPeerIndices, int myHost) {
    this.vertexType = VertexType.MASTER;
    
    this.replicaIDs[this.masterIndex] = -1;  //clear master index, but Need not set myself, because it is already there
    this.masterIndex = myHost;
    this.numberReplicas--; // I becomes master
  }
  
  public void correctLocationInfo(List<Integer> crashedPeerIndices, int myHost){
    this.correctMirrorLocationInfo(crashedPeerIndices, myHost);
    this.correctReplicaLocationInfo(crashedPeerIndices);
  }
  
  private void correctMirrorLocationInfo(List<Integer> crashedPeerIndices, int myHost){
    int newNumOfMirror = 0;
    for (int i = 0; i < numOfMirror; i++){
      if (this.mirrorReplicaIndex[i] != myHost 
          && !crashedPeerIndices.contains(this.mirrorReplicaIndex[i])){ // skip myself (I may become master) and crashed mirrors
        this.mirrorReplicaIndex[newNumOfMirror] = this.mirrorReplicaIndex[i];
        newNumOfMirror++;
      }
    }
    this.numOfMirror = newNumOfMirror;
    
    this.peerToMirrorCache = null;
  }
  
  private void correctReplicaLocationInfo(List<Integer> crashedPeerIndices){
    for (int i : crashedPeerIndices){
      if (this.replicaIDs[i] != -1){
        this.replicaIDs[i] = -1;
        this.numberReplicas--;
      }
    }
  }

  @Override
  public String toString() {
    String tmp1 = "";
    if (this.replicaIDs != null){
      for (int i = 0; i < replicaIDs.length; i++){
        tmp1 += replicaIDs[i] + " ";
      }
    }
    String tmp2 = "";
    if (this.numOfMirror > 0){
      for (int i = 0; i < numOfMirror; i++){
        tmp2 += this.mirrorReplicaIndex[i] + " ";
      }
    }
    


    return getVertexID() + " (" + this.vertexType + ")" + " localID:" + this.localID
        + "  votedToHalt=" + votedToHalt + " awaked=" + awaked  + " value=" + value + " msg=" + msg + " activated=" + activate + " activateAllNeighbors=" + activateAllNeighbors
        + "  nVotedToHalt=" + nVotedToHalt + " nAwaked=" + nAwaked  + " nValue=" + nValue + " nMsg=" + nMsg + " nActivated=" + nActivate + " nActivateAllNeighbors=" + nActivateAllNeighbors
        + " // " + edges + " in{" + inEdges + "} lout{" + outEdges + "}" + "[LocalID:" + tmp1+"]" + " [MirrorIndex: " + tmp2+ "]";
  }
  

  @SuppressWarnings({ "unchecked", "rawtypes" })
  private void setSpecialInEdges(List<RefEdge> list) {
    // ugly pass the type check
    this.inEdges = (List<RefEdge<V, E, M>>)(Object)(list);
    
  }

}
